/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"

#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "apr_thread_cond.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include <assert.h>

#define MAGIC_GC 0xcaffe130

/* GC main ü.  */
struct boehm_gc 
{
  gc_collector_t collector;
};

/* GC type   ü.  */
struct boehm_gc_funcs
{
  walk_func_t walk;
  final_func_t final;
  destroy_func_t destroy;
  const char *description;
  int nr;		/* only used ifdef STATS */
  int mem;		/* only used ifdef STATS */
};

struct boehm_gc_mark 
{
  struct GC_ms_entry *mark_current;
  struct GC_ms_entry *mark_limit;
  GC_PTR original_object;
};

/* define alignment of 'void *' */
#ifndef ALIGNMENTOF_VOIDP
typedef struct { char c; void *p; } alignmentof_voidp_helper;
#define ALIGNMENTOF_VOIDP (offsetof (alignmentof_voidp_helper, p))
#endif

#define SIZEOF_DESC (((sizeof(struct boehm_gc_memdesc) + ALIGNMENTOF_VOIDP - 1) / ALIGNMENTOF_VOIDP) * ALIGNMENTOF_VOIDP)
#define SYSTEM_SIZE(s)		((s) + SIZEOF_DESC)
#define USER_SIZE(s)		((s) - SIZEOF_DESC)
#define ALIGN_FORWARD(p)	((void *) ((unsigned int) (p) + SIZEOF_DESC))
#define ALIGN_BACKWARD(p)	((void *) ((unsigned int) (p) - SIZEOF_DESC))

struct boehm_gc_memdesc 
{
  unsigned int magic;
  unsigned char memtype;
  size_t memsize;
};

#define	REFOBJHASHSZ	128
/* This is a bit homemade.  We need a 7-bit hash from the address here */
#define	REFOBJHASH(V)	((((unsigned int) (V) >> 2) ^ ((unsigned int) (V) >> 9)) % REFOBJHASHSZ)

struct boehm_gc_strong_ref_object
{
  const void*		mem;
  unsigned int		ref;
  struct boehm_gc_strong_ref_object *next;
};

struct boehm_gc_strong_ref_table 
{
  struct boehm_gc_strong_ref_object *hash[REFOBJHASHSZ];
};

struct boehm_gc_weak_ref_object 
{
  const void *mem;
  unsigned int ref;
  void ***all_refs;
  struct boehm_gc_weak_ref_object *next;
};

struct boehm_gc_weak_ref_table 
{
  struct boehm_gc_weak_ref_object *hash[REFOBJHASHSZ];
};

static struct boehm_gc_strong_ref_table strong_ref_objects;
static struct boehm_gc_weak_ref_table weak_ref_objects;

static struct boehm_gc bgc;
/** GC Type handling */
static struct boehm_gc_funcs bgc_funcs[MAX_EGC_ALLOC];
static int EGC_elang_proc;
static int EGC_elang_kind;
static ptr_t * EGC_elang_freelist;

apr_thread_t *finalmon, *gcmon;
static apr_thread_mutex_t *gcmon_lock;
static apr_thread_mutex_t *gcmonend_lock;
static apr_thread_mutex_t *finalmon_lock;
static apr_thread_cond_t *gcmon_cond;
static apr_thread_cond_t *gcmonend_cond;
static apr_thread_cond_t *finalmon_cond;

static volatile int final_running = -1; 
static volatile int gc_running = -1;
static volatile int gc_disabled = 0;

static const char *boehm_gc_get_collector_name (gc_collector_t *);
static void *boehm_gc_malloc (gc_collector_t *, size_t, gc_alloc_type_t);
static void *boehm_gc_realloc (gc_collector_t *, void*, size_t, gc_alloc_type_t);
static void boehm_gc_free (gc_collector_t *, void*);
static void boehm_gc_invoke_GC (gc_collector_t *, int);
static void boehm_gc_invoke_finalizer (gc_collector_t *);
static void boehm_gc_init (gc_collector_t *);
static void boehm_gc_enable (gc_collector_t *, apr_pool_t *);
static void boehm_gc_mark_address (gc_collector_t *, void *, const void *);
static unsigned int boehm_gc_get_object_size (gc_collector_t *, const void *);
static const char *boehm_gc_get_object_desc (gc_collector_t *, const void *);
static int boehm_gc_get_object_index (gc_collector_t *, const void *);
static void *boehm_gc_get_object_base (gc_collector_t *, void *);
static void boehm_gc_register_fixed_type_by_index (gc_collector_t *, gc_alloc_type_t, const char *);
static void boehm_gc_register_gc_type_by_index (gc_collector_t *, gc_alloc_type_t, walk_func_t, final_func_t, destroy_func_t, const char *);static eobjectref_t *boehm_gc_throwOOM (gc_collector_t *);
static void boehm_gc_enable_GC (gc_collector_t *);
static void boehm_gc_disable_GC (gc_collector_t *);
static unsigned int boehm_gc_get_heap_free (gc_collector_t *); 
static unsigned int boehm_gc_get_heap_limit (gc_collector_t *);
static unsigned int boehm_gc_get_heap_total (gc_collector_t *);
static int boehm_gc_add_ref (gc_collector_t *, const void *);
static int boehm_gc_rm_ref (gc_collector_t *, void *);
static int boehm_gc_add_weak_ref (gc_collector_t *, void *, void **);
static int boehm_gc_rm_weak_ref (gc_collector_t *, void *, void **);
static void boehm_gc_clear_weak_ref (gc_collector_t *, void *);
static struct GC_ms_entry *on_object_marking (GC_word *, struct GC_ms_entry *, struct GC_ms_entry *, GC_word);
static void maybe_finalize (void);

void EGC_elang_init (GC_mark_proc);
void *EGC_elang_malloc (size_t);

extern ptr_t GC_clear_stack (ptr_t);

static void
boehm_gc_signal_finalizer (void)
{
  apr_status_t rv;

  rv = apr_thread_mutex_lock (finalmon_lock);
  if (rv != APR_SUCCESS) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
		     "couldn't grab the finalmon mutex");
      return;
    }

  if (final_running == 0) 
    {
      final_running = 1;
      rv = apr_thread_cond_signal (finalmon_cond);
      if (rv != APR_SUCCESS)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
			 "couldn't signal the finalmon condition");
	  return;
	}
    }

  rv = apr_thread_mutex_unlock (finalmon_lock);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
		     "couldn't release the finalmon mutex");
      return;
    }
}

static void
boehm_gc_warnproc (char *msg, GC_word arg)
{
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, msg, arg);
}

static struct GC_ms_entry *
on_object_marking (addr, mark_stack_ptr, mark_stack_limit, env)
     GC_word *addr;
     struct GC_ms_entry *mark_stack_ptr;
     struct GC_ms_entry *mark_stack_limit;
     GC_word env ATTRIBUTE_UNUSED;
{
  walk_func_t walkf;
  struct boehm_gc_mark info_mark;
  struct boehm_gc_memdesc *desc = (struct boehm_gc_memdesc *) addr;
  int type = desc->memtype;
  
  /* Temporary hack. If the magic bytes are not there it means the descriptor has not
   * been initialized. The Boehm-Weiser GC should not call the function but it actually happens
   * we just silently ignore the problem. :(
   */
  if (desc->magic != MAGIC_GC)
    return mark_stack_ptr;
  
  info_mark.mark_current = mark_stack_ptr;
  info_mark.mark_limit = mark_stack_limit;
  info_mark.original_object = addr;

  walkf = bgc_funcs[type].walk;
  if (walkf)
    walkf (&bgc.collector, &info_mark, ALIGN_FORWARD (addr), desc->memsize); 

  return info_mark.mark_current;
}

void
EGC_elang_init (GC_mark_proc proc)
{
  EGC_elang_freelist = (ptr_t *) GC_new_free_list ();
  EGC_elang_proc = GC_new_proc (proc);
  EGC_elang_kind = GC_new_kind ((void **) EGC_elang_freelist, GC_MAKE_PROC (EGC_elang_proc, 0), FALSE, TRUE);
}

static const char *
boehm_gc_get_collector_name (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  return "boehm-gc";
}

static inline void
clear_and_adddesc (void *mem, struct boehm_gc_memdesc *desc)
{
  struct boehm_gc_memdesc *idx = (struct boehm_gc_memdesc *) mem;

  *idx = *desc;
  memset (ALIGN_FORWARD (idx), 0, desc->memsize);
}

static void
boehm_gc_finalize_object (void *ob, void *descriptor ATTRIBUTE_UNUSED)
{
  struct boehm_gc_memdesc *desc = (struct boehm_gc_memdesc *) ob;
  struct boehm_gc_funcs *f = &bgc_funcs[desc->memtype];

  assert (desc->magic == MAGIC_GC);

  if (f->final != EGC_OBJECT_NORMAL && f->final)
    f->final (&bgc.collector, ALIGN_FORWARD (ob));

  boehm_gc_clear_weak_ref (&bgc.collector, ALIGN_FORWARD (ob));

  if (f->destroy)
    f->destroy (&bgc.collector, ALIGN_FORWARD (ob));
}

/* We need a mechanism to release the lock and invoke finalizers.	*/
/* We don't really have an opportunity to do this on a rarely executed	*/
/* path on which the lock is not held.  Thus we check at a 		*/
/* rarely executed point at which it is safe to release the lock.	*/
/* We do this even where we could just call EGC_INVOKE_FINALIZERS,	*/
/* since it's probably cheaper and certainly more uniform.		*/
/* FIXME - Consider doing the same elsewhere?				*/
static void
maybe_finalize (void)
{
  static unsigned int last_finalized_no = 0;
  
  if (GC_gc_no == last_finalized_no) 
    return;
  if (!GC_is_initialized) 
    return;
  UNLOCK ();
  GC_INVOKE_FINALIZERS ();
  last_finalized_no = GC_gc_no;
  LOCK ();
}

#define GENERAL_MALLOC(lb,k) \
    (GC_PTR) GC_clear_stack (GC_generic_malloc_inner ((word) lb, k))

/* Allocate an object, clear it */
void *
EGC_elang_malloc (size_t lb)
{
  register ptr_t op;
  register ptr_t *opp;
  register word lw;
  DCL_LOCK_STATE;

  if(EXPECT (SMALL_OBJ (lb), 1)) 
    {
#ifdef MERGE_SIZES
      lw = GC_size_map[lb];
#else
      lw = ALIGNED_WORDS (lb);
#endif
      opp = &(EGC_elang_freelist[lw]);
      LOCK ();
      op = *opp;
      if(EXPECT(op == 0, 0)) 
	{
	  maybe_finalize ();
	  op = (ptr_t) GENERAL_MALLOC ((word) lb, EGC_elang_kind);
	  if (0 == op) 
	    {
	      UNLOCK ();
	      return (GC_oom_fn (lb));
	    }
#ifdef MERGE_SIZES
	  lw = GC_size_map[lb];	/* May have been uninitialized.	*/
#endif
        } 
      else 
	{
	  *opp = obj_link (op);
	  GC_words_allocd += lw;
        }
      UNLOCK ();
    } 
  else 
    {
      LOCK ();
      maybe_finalize ();
      op = (ptr_t) GENERAL_MALLOC ((word) lb, EGC_elang_kind);
      if (0 == op) 
	{
	  UNLOCK ();
	  return (GC_oom_fn (lb));
	}
      UNLOCK ();
    }
  return ((GC_PTR) op);
}

static void *
boehm_gc_malloc (gc, size, type)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     size_t size;
     gc_alloc_type_t type;
{
  void *ptr;
  struct boehm_gc_memdesc desc;

  assert (bgc_funcs[type].description != NULL);
  assert (size != 0);

  desc.memtype = type;
  desc.memsize = size;
  desc.magic = MAGIC_GC;

  if (bgc_funcs[type].final == EGC_OBJECT_FIXED)
    {
      if (type == EGC_ALLOC_THREADCTX)
	ptr = GC_malloc_uncollectable (SYSTEM_SIZE (size));
      else
	ptr = GC_malloc_atomic_uncollectable (SYSTEM_SIZE (size));
    }
  else
    ptr = EGC_elang_malloc (SYSTEM_SIZE (size));

  if (!ptr)
    return NULL;

  clear_and_adddesc (ptr, &desc);
  if (bgc_funcs[type].final != EGC_OBJECT_FIXED
      && (bgc_funcs[type].final != EGC_OBJECT_NORMAL
	  || bgc_funcs[type].destroy != NULL)) 
    {
      GC_REGISTER_FINALIZER_NO_ORDER (ptr, boehm_gc_finalize_object, 0, 0, 0);
    }

  return ALIGN_FORWARD (ptr);
}

static void *
boehm_gc_realloc (gc, ptr, size, type)
     gc_collector_t *gc;
     void *ptr;
     size_t size;
     gc_alloc_type_t type;
{
  void *new_ptr;
  struct boehm_gc_memdesc *desc;

  if (!ptr)
    return gc->ops->malloc (gc, size, type);
    
  new_ptr = GC_realloc (ALIGN_BACKWARD (ptr), (size_t) SYSTEM_SIZE (size));
  if (!new_ptr)
    return NULL;

  desc = (struct boehm_gc_memdesc *) new_ptr;    

  assert (desc->magic == MAGIC_GC);
  if (size > desc->memsize) 
    memset ((void *) ((unsigned int) new_ptr + SYSTEM_SIZE (desc->memsize)), 0, size - desc->memsize);
  desc->memtype = type;
  desc->memsize = size;

  return ALIGN_FORWARD (new_ptr);
}

static void
boehm_gc_free (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ptr;
{
  struct boehm_gc_memdesc *desc = (struct boehm_gc_memdesc *) ALIGN_BACKWARD (ptr);

  if (!ptr)
    return;

  assert (desc->magic == MAGIC_GC);
  GC_free (desc);
}

static void 
boehm_gc_invoke_GC (gc, mustgc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     int mustgc;
{
  apr_status_t rv;

  while (gc_running < 0)
    apr_thread_yield ();
  
  rv = apr_thread_mutex_lock (gcmon_lock);
  if (rv != APR_SUCCESS) 
    return;

  if (!gc_running)
    {
      gc_running = mustgc ? 2 : 1;
      if (!gc_disabled)
	apr_thread_cond_signal (gcmon_cond);
    }
  
  rv = apr_thread_mutex_lock (gcmonend_lock);
  if (rv != APR_SUCCESS) 
    return;
  rv = apr_thread_mutex_unlock (gcmon_lock);
  if (rv != APR_SUCCESS) 
    return;

  while (gc_running)
    {
      rv = apr_thread_cond_wait (gcmonend_cond, gcmonend_lock);
      if (rv != APR_SUCCESS) 
	return;
    }

  rv = apr_thread_mutex_unlock (gcmonend_lock);
  if (rv != APR_SUCCESS)
    return;
}

static void 
boehm_gc_invoke_finalizer (gc)
     gc_collector_t *gc;
{
  boehm_gc_invoke_GC (gc, 1);
  boehm_gc_signal_finalizer ();
}

static void 
boehm_gc_init (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  GC_all_interior_pointers = 0;
  GC_finalizer_notifier = boehm_gc_signal_finalizer;
  GC_java_finalization = 1;
  GC_finalize_on_demand = 1;
  GC_set_warn_proc (boehm_gc_warnproc);
  /* Ʒ Լ ϱ  GC_malloc   operation  ߻ؼ ȵȴ.  */
  GC_init ();
  GC_set_max_heap_size ((size_t) EGC_MAX_HEAPSIZE);

  if (GC_get_heap_size () < (size_t) EGC_MIN_HEAPSIZE)
    GC_expand_hp (EGC_MIN_HEAPSIZE - GC_get_heap_size ());

  EGC_elang_init (on_object_marking);
}

static void
finaliser_daemon (void *arg ATTRIBUTE_UNUSED)
{
  apr_status_t rv;

  rv = apr_thread_mutex_lock (finalmon_lock);

  while (1)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_INFO, "finaliser_daemon ");
      apr_sleep (46875);

      final_running = 0;
      while (final_running == 0)
	{
	  rv = apr_thread_cond_wait (finalmon_cond, finalmon_lock);
	}
      assert (final_running == 1);

      GC_invoke_finalizers ();
      
      /* Wake up anyone waiting for the finalizer to finish */
      rv = apr_thread_cond_broadcast (finalmon_cond);
    }

  rv = apr_thread_mutex_unlock (finalmon_lock);
}

static void
gc_daemon (void *arg ATTRIBUTE_UNUSED)
{
  apr_status_t rv;

  rv = apr_thread_mutex_lock (gcmon_lock);
  gc_running = 0;
  while (1)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_INFO, "gc_daemon ");
      apr_sleep (46875);

      while (gc_running == 0) 
	{
	  rv = apr_thread_cond_wait (gcmon_cond, gcmon_lock);
	}
      
      GC_gcollect ();
      
      /* Wake up anyone waiting for the finalizer to finish */
      rv = apr_thread_mutex_lock (gcmonend_lock);
      gc_running = 0;
      rv = apr_thread_cond_broadcast (gcmonend_cond);
      rv = apr_thread_mutex_unlock (gcmonend_lock);
    }
}

static void 
boehm_gc_enable (gc, p)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     apr_pool_t *p;
{
  apr_status_t rv;
  apr_threadattr_t *thread_attr;

  apr_threadattr_create (&thread_attr, p);
  /* 0 means PTHREAD_CREATE_JOINABLE */
  apr_threadattr_detach_set (thread_attr, 0);

  rv = apr_thread_create (&finalmon, thread_attr, (apr_thread_start_t) finaliser_daemon, NULL, p);
  if (rv != APR_SUCCESS) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ALERT,
		     "apr_thread_create: unable to create finaliser thread");
      return;
    }

  rv = apr_thread_create (&gcmon, thread_attr, (apr_thread_start_t) gc_daemon, NULL, p);
  if (rv != APR_SUCCESS) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ALERT,
		     "apr_thread_create: unable to create finaliser thread");
      return;
    }

  assert (finalmon && gcmon);
}

static void 
boehm_gc_mark_address (gc, gcinfo, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *gcinfo;
     const void *ptr;
{
  struct boehm_gc_mark *info_mark = (struct boehm_gc_mark *) gcinfo;

  info_mark->mark_current =
    GC_mark_and_push (ALIGN_BACKWARD (ptr), 
		      info_mark->mark_current, 
		      info_mark->mark_limit,
		      (GC_PTR *) info_mark->original_object);
}

static unsigned int 
boehm_gc_get_object_size (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     const void *ptr;
{
  struct boehm_gc_memdesc *desc;

  if (!ptr)
    return 0;

  desc = (struct boehm_gc_memdesc *) GC_base ((GC_PTR) ptr);
  if (!desc)
    return 0;

  assert (desc->magic == MAGIC_GC);
  return desc->memsize;
}

static const char *
boehm_gc_get_object_desc (gc, ptr)
     gc_collector_t *gc;
     const void *ptr;
{
  int idx = boehm_gc_get_object_index (gc, ptr);

  if (idx < 0)
    return NULL;

  return bgc_funcs[idx].description;
}

static int 
boehm_gc_get_object_index (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     const void *ptr;
{
  struct boehm_gc_memdesc *desc = (struct boehm_gc_memdesc *) GC_base ((void *) ptr);

  if (!desc)
    return -1;

  assert (desc->magic == MAGIC_GC);
  return desc->memtype;
}

static void *
boehm_gc_get_object_base (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ptr;
{
  void *base;

  if (!ptr)
    return NULL;
  
  base = GC_base (ptr);
  if (!base)
    return NULL;

  return ALIGN_FORWARD (base);
}

/**
   register an allocation type under a certain index
   NB: we could instead return a pointer to the record describing the
   allocation type.  This would give us more flexibility, but it wouldn't
   allow us to use compile-time constants.
*/
static void
register_type_by_index (gc_alloc_type_t type, walk_func_t walk, final_func_t final,
			destroy_func_t destroy,
			const char *desc)
{
  /* once only, please */
  assert (bgc_funcs[type].description == 0);
  /* don't exceed bounds */
  assert (type < sizeof (bgc_funcs) / sizeof (bgc_funcs[0]));

  bgc_funcs[type].walk = walk;
  bgc_funcs[type].final = final;
  bgc_funcs[type].destroy = destroy;
  bgc_funcs[type].description = desc;
}

static void 
boehm_gc_register_fixed_type_by_index (gc, type, desc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     gc_alloc_type_t type;
     const char *desc;
{
  register_type_by_index (type, NULL, EGC_OBJECT_FIXED, 0, desc);
}

static void 
boehm_gc_register_gc_type_by_index (gc, type, walk, final, destroy, desc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     gc_alloc_type_t type;
     walk_func_t walk;
     final_func_t final;
     destroy_func_t destroy;
     const char *desc;
{
  register_type_by_index (type, walk, final, destroy, desc);
}

static eobjectref_t *
boehm_gc_throwOOM (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "Throwing OutOfMemoryError is not implemented in Boehm-GC.");
  ENV_ABORT ();

  return NULL;
}

static void 
boehm_gc_enable_GC (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  apr_status_t rv;

  rv = apr_thread_mutex_lock (gcmon_lock);
  if (rv != APR_SUCCESS) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
		     "couldn't grab the gcmon mutex");
      return;
    }

  gc_disabled -= 1;
  if (!gc_disabled)
    {
      rv = apr_thread_cond_broadcast (gcmon_cond);
      if (rv != APR_SUCCESS)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
			 "couldn't broadcase the gcmon condition");
	  return;
	}
    }
  
  rv = apr_thread_mutex_unlock (gcmon_lock);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
		     "couldn't release the gcmon mutex");
      return;
    }
}

static void 
boehm_gc_disable_GC (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  apr_status_t rv;

  rv = apr_thread_mutex_lock (gcmon_lock);
  if (rv != APR_SUCCESS) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
		     "couldn't grab the gcmon mutex");
      return;
    }

  gc_disabled += 1;

  rv = apr_thread_mutex_unlock (gcmon_lock);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG,
		     "couldn't release the gcmon mutex");
      return;
    }
}

static unsigned int 
boehm_gc_get_heap_free (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  return GC_get_free_bytes();
}

static unsigned int 
boehm_gc_get_heap_limit (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  return 0;
}

static unsigned int 
boehm_gc_get_heap_total (gc)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
{
  return GC_get_heap_size();
}

/*
 * Add a persistent reference to an object.
 */
static int 
boehm_gc_add_ref (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     const void *ptr;
{
  unsigned int idx;
  struct boehm_gc_strong_ref_object *obj;

  idx = REFOBJHASH (ptr);
  for (obj = strong_ref_objects.hash[idx]; obj != 0; obj = obj->next) 
    {
      /* Found it - just increase reference */
      if (obj->mem == ptr) 
	{
	  obj->ref++;
	  return TRUE;
	}
    }

  /* Not found - create a new one */
  obj = (struct boehm_gc_strong_ref_object *) GC_malloc_uncollectable (sizeof(struct boehm_gc_strong_ref_object));
  if (!obj)
    return FALSE;
	
  obj->mem = ALIGN_BACKWARD (ptr);
  obj->ref = 1;
  obj->next = strong_ref_objects.hash[idx];
  strong_ref_objects.hash[idx] = obj;
  return TRUE;
}

static int 
boehm_gc_rm_ref (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ptr;
{
  unsigned int idx;
  struct boehm_gc_strong_ref_object **objp;
  struct boehm_gc_strong_ref_object *obj;

  idx = REFOBJHASH (ptr);
  ptr = ALIGN_BACKWARD (ptr);

  for (objp = &strong_ref_objects.hash[idx]; *objp != 0; objp = &obj->next)
    {
      obj = *objp;
      /* Found it - just decrease reference */
      if (obj->mem == ptr) 
	{
	  obj->ref--;
	  if (obj->ref == 0) 
	    {
	      *objp = obj->next;
	      GC_free (obj);
	    }
	  return TRUE;
	}
    }
  
  /* Not found!! */
  return FALSE;
}

static int 
boehm_gc_add_weak_ref (gc, ptr, refobj)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ptr, **refobj;
{
  int idx;
  struct boehm_gc_weak_ref_object *obj;

  idx = REFOBJHASH (ptr);
  for (obj = weak_ref_objects.hash[idx]; obj != 0; obj = obj->next) 
    {
      /* Found it - just register a new weak reference */
      if (obj->mem == ptr) 
	{
	  void ***newRefs;
	  obj->ref++;
	  
	  newRefs = (void ***) GC_malloc_uncollectable (sizeof (void ***) * obj->ref);
	  memcpy (newRefs, obj->all_refs, sizeof (void ***) * (obj->ref-1));
	  GC_free (obj->all_refs);
	  
	  obj->all_refs = newRefs;
	  obj->all_refs[obj->ref-1] = refobj;
	  return TRUE;
	}
    }

  /* Not found - create a new one */
  obj = (struct boehm_gc_weak_ref_object*) GC_malloc_uncollectable (sizeof (struct boehm_gc_weak_ref_object));
  if (!obj)
    return FALSE;

  obj->mem = ptr;
  obj->ref = 1;
  obj->all_refs = (void ***) GC_malloc_uncollectable (sizeof (void ***));
  obj->all_refs[0] = refobj;
  obj->next = weak_ref_objects.hash[idx];
  weak_ref_objects.hash[idx] = obj;

  return TRUE;
}

static int 
boehm_gc_rm_weak_ref (gc, ptr, refobj)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ptr, **refobj;
{
  unsigned int idx;
  struct boehm_gc_weak_ref_object **objp;
  struct boehm_gc_weak_ref_object *obj;
  unsigned int i;

  idx = REFOBJHASH (ptr);
  for (objp = &weak_ref_objects.hash[idx]; *objp != 0; objp = &obj->next) 
    {
      obj = *objp;
      /* Found it - just decrease reference */
      if (obj->mem == ptr)
	{
	  for (i = 0; i < obj->ref; i++)
	    {
	      if (obj->all_refs[i] == refobj)
		{
		  void ***newRefs;
		  
		  obj->ref--;
		  newRefs = (void ***) GC_malloc_uncollectable (sizeof (void ***) * obj->ref);
		  memcpy (newRefs, obj->all_refs, i * sizeof (void ***));
		  memcpy (&newRefs[i], &obj->all_refs[i+1], obj->ref * sizeof (void ***));
		  GC_free (obj->all_refs);
		  obj->all_refs = newRefs;
		  break;
		}
	    }
	  if (i == obj->ref)
	    return FALSE;
	  if (obj->ref == 0) 
	    {
	      *objp = obj->next;
	      GC_free (obj);
	    }
	  return TRUE;
	}
    }
  
  /* Not found!! */
  return FALSE;
}

/**
   This function clear all weak references to the specified object. 
   The references are then removed from the database.
   
   @param collector a garbage collector instance.
   @param mem a valid memory object.
*/
static void
boehm_gc_clear_weak_ref (gc, ptr)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ptr;
{
  unsigned int idx;
  struct boehm_gc_weak_ref_object **objp;
  struct boehm_gc_weak_ref_object *obj;
  unsigned int i;

  idx = REFOBJHASH (ptr);
  for (objp = &weak_ref_objects.hash[idx]; *objp != 0; objp = &obj->next)
    {
      obj = *objp;
      /* Found it - clear all references attached to it. */
      if (obj->mem == ptr)
	{
	  for (i = 0; i < obj->ref; i++)
	    *(obj->all_refs[i]) = NULL;
	  GC_free (obj->all_refs);
	  
	  *objp = obj->next;
	  GC_free (obj);
	  return;
	}
    }
}

static struct gc_interface_ops GC_ops =
  {
    boehm_gc_get_collector_name,/* reserved */
    NULL,			/* reserved */
    NULL,			/* reserved */
    boehm_gc_malloc,
    boehm_gc_realloc,
    boehm_gc_free,
    boehm_gc_invoke_GC,
    boehm_gc_invoke_finalizer,
    boehm_gc_init,
    boehm_gc_enable,
    boehm_gc_mark_address,
    boehm_gc_mark_address,
    boehm_gc_get_object_size,
    boehm_gc_get_object_desc,
    boehm_gc_get_object_index,
    boehm_gc_get_object_base,
    boehm_gc_register_fixed_type_by_index,
    boehm_gc_register_gc_type_by_index,
    boehm_gc_throwOOM,
    boehm_gc_enable_GC,
    boehm_gc_disable_GC,
    boehm_gc_get_heap_free,
    boehm_gc_get_heap_limit,
    boehm_gc_get_heap_total,
    boehm_gc_add_ref,
    boehm_gc_rm_ref,
    boehm_gc_add_weak_ref,
    boehm_gc_rm_weak_ref
  };

gc_collector_t *
initialize_boehm_gc (p)
     apr_pool_t *p;
{
  apr_status_t rv;
  bgc.collector.ops = &GC_ops;

  rv = apr_thread_mutex_create (&gcmon_lock, APR_THREAD_MUTEX_DEFAULT, p);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create gcmon lock");
      return NULL;
    }

  rv = apr_thread_mutex_create (&gcmonend_lock, APR_THREAD_MUTEX_DEFAULT, p);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create gcmonend lock");
      return NULL;
    }

  rv = apr_thread_mutex_create (&finalmon_lock, APR_THREAD_MUTEX_DEFAULT, p);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create finalmon lock");
      return NULL;
    }

  rv = apr_thread_cond_create (&gcmon_cond, p);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create gcmon cond");
      return NULL;
    }

  rv = apr_thread_cond_create (&gcmonend_cond, p);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create gcmonend cond");
      return NULL;
    }

  rv = apr_thread_cond_create (&finalmon_cond, p);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create finalmon cond");
      return NULL;
    }

  return (&bgc.collector);
}
