/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ENV_GARBAGE_COLLECTION_H
#define ENV_GARBAGE_COLLECTION_H

/**
   Default values for initial and maximum heap size and heap increment.
*/
#define	EGC_MIN_HEAPSIZE	(5 * 1024 * 1024)
#define	EGC_MAX_HEAPSIZE	(64 * 1024 * 1024)
#define	EGC_ALLOC_HEAPSIZE	(1024 * 1024)

typedef struct gc_collector gc_collector_t;
typedef void (*walk_func_t) (gc_collector_t *, void *, void *, unsigned int);
typedef void (*final_func_t) (gc_collector_t *, void *);
typedef void (*destroy_func_t) (gc_collector_t *, void *);

#define	EGC_OBJECT_NORMAL	((final_func_t) 0)
#define	EGC_OBJECT_FIXED	((final_func_t) 1)

/**
   Garbage collector interface.
*/
typedef enum 
  {
    /* allocation types for different kinds of elang objects */
    EGC_ALLOC_ELANG_STRING,
    EGC_ALLOC_NOWALK,
    EGC_ALLOC_NORMALOBJECT,
    EGC_ALLOC_PRIMARRAY,
    EGC_ALLOC_REFARRAY,
    EGC_ALLOC_FINALIZEOBJECT,

    /* allocation types used for java.lang.Class and its parts */
    EGC_ALLOC_CLASS,
    EGC_ALLOC_BYTECODE,
    EGC_ALLOC_EXCEPTIONTABLE,
    EGC_ALLOC_STATICDATA,
    EGC_ALLOC_CONSTANT,
    EGC_ALLOC_DISPATCHTABLE,
    EGC_ALLOC_METHOD,
    EGC_ALLOC_FIELD,
    EGC_ALLOC_INTERFACE,
    EGC_ALLOC_LINENRTABLE,
    EGC_ALLOC_LOCALVARTABLE,
    EGC_ALLOC_DECLAREDEXC,
    EGC_ALLOC_INTERFACE_TABLE,
    EGC_ALLOC_CLASSMISC,
    
    /* miscelanious allocation types */
    EGC_ALLOC_FIXED,
    EGC_ALLOC_UTF8CONST,
    EGC_ALLOC_LOCK,
    EGC_ALLOC_THREADCTX,
    EGC_ALLOC_STATIC_THREADDATA,
    EGC_ALLOC_REF,
    EGC_ALLOC_CODEANALYSE,
    EGC_ALLOC_CLASSPOOL,
    EGC_ALLOC_VERIFIER,
    EGC_ALLOC_NATIVELIB,
    MAX_EGC_ALLOC
  } gc_alloc_type_t;

/**
   Virtual Interface for GC.
*/
struct gc_interface_ops
{
  const char *(*get_collector_name) (gc_collector_t *);
  void *reserved_2;
  void *reserved_3;
  void *(*malloc) (gc_collector_t *, size_t, gc_alloc_type_t);
  void *(*realloc) (gc_collector_t *, void*, size_t, gc_alloc_type_t);
  void (*free) (gc_collector_t *, void*);
  
  void (*invoke) (gc_collector_t *, int);
  void (*invoke_finalizer) (gc_collector_t *);
  void (*init) (gc_collector_t *);
  void (*enable) (gc_collector_t *, apr_pool_t *);
  
  void (*mark_address) (gc_collector_t *, void *, const void *);
  void (*mark_object) (gc_collector_t *, void *, const void *);
  unsigned int (*get_object_size) (gc_collector_t *, const void *);
  const char *(*get_object_desc) (gc_collector_t *, const void *);
  int (*get_object_index) (gc_collector_t *, const void *);
  void *(*get_object_base) (gc_collector_t *, void *);
  
  void (*register_fixed_type_by_index) (gc_collector_t *, gc_alloc_type_t, const char *);
  void (*register_gc_type_by_index) (gc_collector_t *, gc_alloc_type_t, walk_func_t, final_func_t, destroy_func_t, const char *);
  eobjectref_t *(*throwOOM) (gc_collector_t *);
  
  void (*enable_GC) (gc_collector_t *);
  void (*disable_GC) (gc_collector_t *);
  
  unsigned int (*get_heap_free) (gc_collector_t *); 
  unsigned int (*get_heap_limit) (gc_collector_t *);
  unsigned int (*get_heap_total) (gc_collector_t *);
  
  int (*add_ref) (gc_collector_t *, const void *);
  int (*rm_ref) (gc_collector_t *, void *);
  int (*add_weak_ref) (gc_collector_t *, void *, void **);
  int (*rm_weak_ref) (gc_collector_t *, void *, void **);
};

#define EGC_COLLECTOR_OPERATION(NODE)	((NODE)->ops)

struct gc_collector
{
  struct gc_interface_ops *ops;
};

/**
   Convenience macros
*/
#define EGC_malloc(G, size, type)		((G)->ops->malloc) ((gc_collector_t*) (G), (size), (type))
#define EGC_realloc(G, addr, size, type)	((G)->ops->realloc) ((gc_collector_t*) (G), (addr), (size), (type))
#define EGC_free(G, addr)			((G)->ops->free) ((gc_collector_t*) (G), (addr))
#define EGC_invoke(G, mustgc)			((G)->ops->invoke) ((gc_collector_t*) (G), (mustgc))
#define EGC_invoke_finalizer(G)			((G)->ops->invoke_finalizer) ((gc_collector_t*) (G))
#define EGC_init(G)				((G)->ops->init) ((gc_collector_t*) (G))
#define EGC_enable(G, P)			((G)->ops->enable) ((gc_collector_t*) (G), (P))
#define EGC_throwOOM(G)				((G)->ops->throwOOM) ((gc_collector_t*) (G))
#define EGC_mark_address(G, gc_info, addr)	((G)->ops->mark_address) ((gc_collector_t*) (G), (gc_info), (addr))
#define EGC_add_ref(G, addr)			((G)->ops->add_ref) ((gc_collector_t*) (G), (addr))
#define EGC_rm_ref(G, addr)			((G)->ops->rm_ref) ((gc_collector_t*) (G), (addr))
#define EGC_add_weak_ref(G, addr, ref)		((G)->ops->add_weak_ref) ((gc_collector_t *) (G), (addr), (ref))
#define EGC_rm_weak_ref(G, addr, ref)		((G)->ops->rm_weak_ref) ((gc_collector_t *) (G), (addr), (ref))
static inline void
EGC_mark_object (gc_collector_t *g, void *gc_info, const void *addr)
{
  if (addr)
    (g)->ops->mark_object (g, gc_info, addr);
}
#define EGC_get_object_size(G, obj)		((G)->ops->get_object_size) ((gc_collector_t*) (G), (obj))
#define EGC_get_object_description(G, obj)	((G)->ops->get_object_description) ((gc_collector_t*) (G), (obj))
#define EGC_get_object_index(G, obj)		((G)->ops->get_object_index) ((gc_collector_t*) (G), (obj))
#define EGC_get_object_base(G, obj)		((G)->ops->get_object_base) ((gc_collector_t*) (G), (obj))
#define EGC_register_fixed_type_by_index(G, idx, desc) \
	((G)->ops->register_fixed_type_by_index) ((gc_collector_t*) (G), (idx), (desc))
#define EGC_register_gc_type_by_index(G, idx, walk, final, destroy, desc) \
	((G)->ops->register_gc_type_by_index) ((gc_collector_t*) (G), (idx), (walk), (final), (destroy), (desc))
#define EGC_enable_GC(G)			((G)->ops->enable_GC) ((gc_collector_t*) (G));
#define EGC_disable_GC(G)			((G)->ops->disable_GC) ((gc_collector_t*) (G));
#define EGC_get_heap_free(G)			((G)->ops->get_heap_free) ((gc_collector_t *) (G));
#define EGC_get_heap_limit(G)			((G)->ops->get_heap_limit) ((gc_collector_t *) (G));
#define EGC_get_heap_total(G)			((G)->ops->get_heap_total) ((gc_collector_t *) (G));
#define EGC_get_collector_name(G)		((G)->ops->get_collector_name) ((gc_collector_t *) (G));

/**
   Compatibility macros to access GC functions
*/
#define	egc_malloc(A,B)		EGC_malloc (VM_GC (ESERVER_VM (eserver)), A, B)
#define	egc_calloc(A,B,C)	EGC_malloc (VM_GC (ESERVER_VM (eserver)), (A) * (B), C)
#define	egc_realloc(A,B,C)	EGC_realloc (VM_GC (ESERVER_VM (eserver)), (A),(B), C)
#define	egc_free(A)		EGC_free (VM_GC (ESERVER_VM (eserver)), (A))
#define egc_add_ref(A)		EGC_add_ref (VM_GC (ESERVER_VM (eserver)), (A))
#define egc_rm_ref(A)		EGC_rm_ref (VM_GC (ESERVER_VM (eserver)), (A))

#define	egc_invoke_GC()		EGC_invoke (VM_GC (ESERVER_VM (eserver)), 1)
#define	egc_advise_GC()		EGC_invoke (VM_GC (ESERVER_VM (eserver)), 0)
#define	egc_invoke_finalizer()	EGC_invoke_finalizer (VM_GC (ESERVER_VM (eserver)))

#define egc_throwOOM()		EGC_throwOOM (VM_GC (ESERVER_VM (eserver)))

#define egc_enable_GC()		EGC_enable_GC (VM_GC (ESERVER_VM (eserver)))
#define egc_disable_GC()	EGC_disable_GC (VM_GC (ESERVER_VM (eserver)))

gc_collector_t * initialize_boehm_gc (apr_pool_t *);
extern gc_collector_t * initialize_gc (apr_pool_t *);

#endif
