/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

typedef struct schemes_t schemes_t;

/** Structure to store various schemes and their default ports */
struct schemes_t 
{
  /** The name of the scheme */
  const char *name;
  /** The default port for the scheme */
  unsigned short default_port;
};

/* Some WWW schemes and their default ports; this is basically /etc/services */
/* This will become global when the protocol abstraction comes */
/* As the schemes are searched by a linear search, */
/* they are sorted by their expected frequency */
static schemes_t schemes[] =
{
    {"e",	APR_URI_HTTP_DEFAULT_PORT},
    { NULL, 0xFFFF }     /* unknown port */
};

unsigned short euri_port_of_scheme (const char *);
unsigned short euri_default_port_for_scheme (const char *);

unsigned short euri_port_of_scheme (const char *scheme_str)
{
  schemes_t *scheme;
  
  if (scheme_str) 
    {
      for (scheme = schemes; scheme->name != NULL; ++scheme) 
	{
	  if (strcasecmp (scheme_str, scheme->name) == 0) 
	    {
	      return scheme->default_port;
            }
        }
    }
  return 0;
}

/** @deprecated @see euri_port_of_scheme */
unsigned short euri_default_port_for_scheme (const char *scheme_str)
{
  return euri_port_of_scheme (scheme_str);
}

/* Here is the hand-optimized parse_uri_components ().  There are some wild
 * tricks we could pull in assembly language that we don't pull here... like we
 * can do word-at-time scans for delimiter characters using the same technique
 * that fast memchr ()s use.  But that would be way non-portable. -djg
 */

/* We have a apr_table_t that we can index by character and it tells us if the
 * character is one of the interesting delimiters.  Note that we even get
 * compares for NUL for free -- it's just another delimiter.
 */

#define T_COLON           0x01        /* ':' */
#define T_SLASH           0x02        /* '/' */
#define T_QUESTION        0x04        /* '?' */
#define T_HASH            0x08        /* '#' */
#define T_NUL             0x80        /* '\0' */

/* this file is automatically generated by gen_uri_delims, do not edit */
static const unsigned char uri_delims[256] = 
  {
    T_NUL,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,T_HASH,0,0,0,0,
    0,0,0,0,0,0,0,T_SLASH,0,0,0,0,0,0,0,0,0,0,T_COLON,0,
    0,0,0,T_QUESTION,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  };

/* it works like this:
    if (uri_delims[ch] & NOTEND_foobar) {
        then we're not at a delimiter for foobar
    }
*/

/* Note that we optimize the scheme scanning here, we cheat and let the
 * compiler know that it doesn't have to do the & masking.
 */
#define NOTEND_SCHEME     (0xff)
#define NOTEND_HOSTINFO   (T_SLASH | T_QUESTION | T_HASH | T_NUL)
#define NOTEND_PATH       (T_QUESTION | T_HASH | T_NUL)

/* parse_uri_components ():
 * Parse a given URI, fill in all supplied fields of a uri_components
 * structure. This eliminates the necessity of extracting host, port,
 * path, query info repeatedly in the modules.
 * Side effects:
 *  - fills in fields of uri_components *uptr
 *  - none on any of the r->* fields
 */
int euri_parse (p, uri, uptr)
     apr_pool_t *p;
     const char *uri;
     euri_t *uptr;
{
  const char *s, *s1;
  const char *hostinfo;
  char *endstr;
  int port;
  int v6_offset1 = 0, v6_offset2 = 0;
  
  /* Initialize the structure. parse_uri () and parse_uri_components ()
   * can be called more than once per request.
   */
  memset (uptr, '\0', sizeof (*uptr));
  uptr->is_initialized = 1;
  
  /* We assume the processor has a branch predictor like most --
   * it assumes forward branches are untaken and backwards are taken.  That's
   * the reason for the gotos.  -djg
   */
  if (uri[0] == '/') 
    {
    deal_with_path:
      /* we expect uri to point to first character of path ... remember
       * that the path could be empty -- http://foobar?query for example
       */
      s = uri;
      while ((uri_delims[*(unsigned char *)s] & NOTEND_PATH) == 0) 
	{
	  ++s;
        }
      if (s != uri) 
	{
	  { /*  ڵ  -->   */
	    const char *bs = uri, *nse;
	    const char *ks = NULL, *ke = NULL;
	    int one_com = 0, two_com = 0;
	    
	    while (bs+1 <= s 
		   && !((*bs == '(' && *(bs+1) == '\'')
			|| (*bs == '(' && *(bs+1) == '"')))
	      {
		++bs;
	      }
	    if (bs != s)
	      {
		if (bs < s && bs[0] == '(' && bs[1] == '\'')
		  one_com = 1;
		else if (bs < s && bs[0] == '(' && bs[1] == '"')
		  two_com = 1;
		nse = bs;
		ks = bs = bs + 2;
		while (s >= bs -1 && !((*(bs-1) == '\'' && *bs == ')')
				       || (*(bs-1) == '"' && *bs == ')')))
		  {
		    bs++;
		  }
		if ((one_com && bs[-1] == '\'')
		    || (two_com && bs[-1] == '"'))
		  ke = bs - 1;
		
	      }
	    if (ks && ke && (ks != ke))
	      {
		int keysize;
		/*  迭 static  Ǿ ִµ,  Լ  primitive type  
		   ؼ Ǿ Ѵ.  */
		char *tmpkeystr;
		ENVKEY_CHAR sk[ENVKEY_SIZE];
		envkey_t key;

		tmpkeystr = apr_pstrmemdup (p, ks, ke - ks);
		keysize = apr_base64_decode_binary (sk, tmpkeystr);
		if (keysize != ENVKEY_SIZE)
		  return APR_EGENERAL;

		PUT_ENVKEY (sk, &key);
		PDUPLICATE_ENVKEY (p, &key, uptr->key);

		s = nse;
	      }
	  } /*  ڵ  -->   */

	  uptr->path = apr_pstrmemdup (p, uri, s - uri);
        }
      if (*s == 0) 
	{
	  return APR_SUCCESS;
        }
      if (*s == '?') 
	{
	  ++s;
	  s1 = strchr (s, '#');
	  if (s1) 
	    {
	      uptr->fragment = apr_pstrdup (p, s1 + 1);
	      uptr->query = apr_pstrmemdup (p, s, s1 - s);
            }
	  else 
	    {
	      uptr->query = apr_pstrdup (p, s);
            }
	  return APR_SUCCESS;
        }
      /* otherwise it's a fragment */
      uptr->fragment = apr_pstrdup (p, s + 1);
      return APR_SUCCESS;
    }
  
  /* find the scheme: */
  s = uri;
  while ((uri_delims[*(unsigned char *)s] & NOTEND_SCHEME) == 0) 
    {
      ++s;
    }
  /* scheme must be non-empty and followed by :// */
  if (s == uri || s[0] != ':' || s[1] != '/' || s[2] != '/') 
    {
      goto deal_with_path;        /* backwards predicted taken! */
    }
  
  uptr->scheme = apr_pstrmemdup (p, uri, s - uri);
  s += 3;
  hostinfo = s;
  while ((uri_delims[*(unsigned char *)s] & NOTEND_HOSTINFO) == 0) 
    {
      ++s;
    }
  uri = s;        /* whatever follows hostinfo is start of uri */
  uptr->hostinfo = apr_pstrmemdup (p, hostinfo, uri - hostinfo);
  
  /* If there's a username:password@host:port, the @ we want is the last @...
   * too bad there's no memrchr ()... For the C purists, note that hostinfo
   * is definately not the first character of the original uri so therefore
   * &hostinfo[-1] < &hostinfo[0] ... and this loop is valid C.
   */
  do 
    {
      --s;
    } 
  while (s >= hostinfo && *s != '@');
  if (s < hostinfo) 
    {
      /* again we want the common case to be fall through */
    deal_with_host:
      /* We expect hostinfo to point to the first character of
       * the hostname.  If there's a port it is the first colon,
       * except with IPv6.
       */
      if (*hostinfo == '[') 
	{
	  v6_offset1 = 1;
	  v6_offset2 = 2;
	  s = memchr (hostinfo, ']', uri - hostinfo);
	  if (s == NULL) 
	    {
	      return APR_EGENERAL;
	    }
	  if (*++s != ':') 
	    {
	      s = NULL; /* no port */
            }
        }
      else 
	{
	  s = memchr (hostinfo, ':', uri - hostinfo);
        }
      if (s == NULL) 
	{
	  /* we expect the common case to have no port */
	  uptr->hostname = apr_pstrmemdup (p,
					  hostinfo + v6_offset1,
					  uri - hostinfo - v6_offset2);
	  goto deal_with_path;
        }
      uptr->hostname = apr_pstrmemdup (p,
				      hostinfo + v6_offset1,
				      s - hostinfo - v6_offset2);
      ++s;
      uptr->port_str = apr_pstrmemdup (p, s, uri - s);
      if (uri != s) 
	{
	  port = strtol (uptr->port_str, &endstr, 10);
	  uptr->port = port;
	  if (*endstr == '\0') 
	    {
	      goto deal_with_path;
            }
	  /* Invalid characters after ':' found */
	  return APR_EGENERAL;
        }
      uptr->port = euri_port_of_scheme (uptr->scheme);
      goto deal_with_path;
    }
  
  /* first colon delimits username:password */
  s1 = memchr (hostinfo, ':', s - hostinfo);
  if (s1) 
    {
      uptr->user = apr_pstrmemdup (p, hostinfo, s1 - hostinfo);
      ++s1;
      uptr->password = apr_pstrmemdup (p, s1, s - s1);
    }
  else 
    {
      uptr->user = apr_pstrmemdup (p, hostinfo, s - hostinfo);
    }
  hostinfo = s + 1;
  goto deal_with_host;
}
