/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

void 
zend_llist_init (zend_llist *l, size_t size, llist_dtor_func_t dtor)
{
  l->head  = NULL;
  l->tail  = NULL;
  l->count = 0;
  l->size  = size;
  l->dtor  = dtor;
}

void 
zend_llist_add_element (zend_llist *l, void *element)
{
  zend_llist_element *tmp = env_malloc (sizeof (zend_llist_element)+l->size-1);
  
  tmp->prev = l->tail;
  tmp->next = NULL;
  if (l->tail) 
    {
      l->tail->next = tmp;
    } 
  else 
    {
      l->head = tmp;
    }
  l->tail = tmp;
  memcpy (tmp->data, element, l->size);
  
  ++l->count;
}

void
zend_llist_prepend_element (zend_llist *l, void *element)
{
  zend_llist_element *tmp = env_malloc (sizeof (zend_llist_element)+l->size-1);
  
  tmp->next = l->head;
  tmp->prev = NULL;
  if (l->head) 
    {
      l->head->prev = tmp;
    }
  else 
    {
      l->tail = tmp;
    }
  l->head = tmp;
  memcpy (tmp->data, element, l->size);
  
  ++l->count;
}

#define DEL_LLIST_ELEMENT(current, l) \
			if ((current)->prev) {\
				(current)->prev->next = (current)->next;\
			} else {\
				(l)->head = (current)->next;\
			}\
			if ((current)->next) {\
				(current)->next->prev = (current)->prev;\
			} else {\
				(l)->tail = (current)->prev;\
			}\
			if ((l)->dtor) {\
				(l)->dtor ((current)->data);\
				env_free ((current));\
			} \
			--(l)->count;

void
zend_llist_del_element (zend_llist *l, void *element, int (*compare)(void *element1, void *element2))
{
  zend_llist_element *current=l->head;
  zend_llist_element *next;
  
  while (current) 
    {
      next = current->next;
      if (compare (current->data, element)) 
	{
	  DEL_LLIST_ELEMENT (current, l);
	  break;
	}
      current = next;
    }
}

void 
zend_llist_destroy (zend_llist *l)
{
  zend_llist_element *current=l->head, *next;
  
  while (current) 
    {
      next = current->next;
      if (l->dtor) 
	{
	  l->dtor (current->data);
	}
      env_free (current);
      current = next;
    }
  
  l->count = 0;
}

void 
zend_llist_clean (zend_llist *l)
{
  zend_llist_destroy (l);
  l->head = l->tail = NULL;
}

void *
zend_llist_remove_tail (zend_llist *l)
{
  zend_llist_element *old_tail;
  void *data;
  
  if ((old_tail = l->tail)) 
    {
      if (l->tail->prev) 
	{
	  l->tail->prev->next = NULL;
	}
      
      data = old_tail->data;
      
      l->tail = l->tail->prev;
      if (l->dtor) 
	{
	  l->dtor (data);
	}
      env_free (old_tail);
      
      --l->count;
      
      return data;
    }
  
  return NULL;
}

void
zend_llist_copy (zend_llist *dst, zend_llist *src)
{
  zend_llist_element *ptr;
  
  zend_llist_init (dst, src->size, src->dtor);
  ptr = src->head;
  while (ptr) 
    {
      zend_llist_add_element (dst, ptr->data);
      ptr = ptr->next;
    }
}

void
zend_llist_apply_with_del (zend_llist *l, int (*func)(void *data))
{
  zend_llist_element *element, *next;
  
  element=l->head;
  while (element) 
    {
      next = element->next;
      if (func (element->data)) 
	{
	  DEL_LLIST_ELEMENT (element, l);
	}
      element = next;
    }
}

void
zend_llist_apply (zend_llist *l, llist_apply_func_t func)
{
  zend_llist_element *element;
  
  for (element=l->head; element; element=element->next) 
    {
      func (element->data);
    }
}

void
zend_llist_sort (zend_llist *l, llist_compare_func_t comp_func)
{
  size_t i;
  
  zend_llist_element **elements;
  zend_llist_element *element, **ptr;
  
  if (l->count <= 0) 
    {
      return;
    }
  
  elements = (zend_llist_element **) env_malloc (l->count * sizeof (zend_llist_element *));
  
  ptr = &elements[0];
  
  for (element=l->head; element; element=element->next) 
    {
      *ptr++ = element;
    }
  
  zend_qsort (elements, l->count, sizeof (zend_llist_element *), (compare_func_t) comp_func);
  
  l->head = elements[0];
  elements[0]->prev = NULL;
  
  for (i = 1; i < l->count; i++) 
    {
      elements[i]->prev = elements[i-1];
      elements[i-1]->next = elements[i];
    }
  elements[i-1]->next = NULL;
  l->tail = elements[i-1];
  env_free (elements);
}

void
zend_llist_apply_with_argument (zend_llist *l, llist_apply_with_arg_func_t func, void *arg)
{
  zend_llist_element *element;
  
  for (element=l->head; element; element=element->next) 
    {
      func (element->data, arg);
    }
}

void
zend_llist_apply_with_arguments (zend_llist *l, llist_apply_with_args_func_t func, int num_args, ...)
{
  zend_llist_element *element;
  va_list args;
  
  va_start (args, num_args);
  for (element=l->head; element; element=element->next) 
    {
      func (element->data, num_args, args);
    }
  va_end (args);
}

int 
zend_llist_count (zend_llist *l)
{
  return l->count;
}
    
void *
zend_llist_get_first_ex (zend_llist *l, zend_llist_position *pos)
{
  zend_llist_position *current = pos ? pos : &l->traverse_ptr;
  
  *current = l->head;
  if (*current) 
    {
      return (*current)->data;
    } 
  else 
    {
      return NULL;
    }
}

void *
zend_llist_get_last_ex (zend_llist *l, zend_llist_position *pos)
{
  zend_llist_position *current = pos ? pos : &l->traverse_ptr;
  
  *current = l->tail;
  if (*current) 
    {
      return (*current)->data;
    }
  else 
    {
      return NULL;
    }
}

void *
zend_llist_get_next_ex (zend_llist *l, zend_llist_position *pos)
{
  zend_llist_position *current = pos ? pos : &l->traverse_ptr;
  
  if (*current) 
    {
      *current = (*current)->next;
      if (*current) 
	{
	  return (*current)->data;
	}
    }

  return NULL;
}

void *
zend_llist_get_prev_ex (zend_llist *l, zend_llist_position *pos)
{
  zend_llist_position *current = pos ? pos : &l->traverse_ptr;
  
  if (*current) 
    {
      *current = (*current)->prev;
      if (*current) 
	{
	  return (*current)->data;
	}
    }

  return NULL;
}
