/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include "env2etx.h"

enum e_method_number_type
  {
    INVALID_METHOD_NUMBER = 0,
    INSTRUCTION_METHOD_NUMBER = 1,
    OPERATION_METHOD_NUMBER = 2
  };

static void readu1 (efile_t *, u1 *);
static void readu2 (efile_t *, u2 *);
static void readu4 (efile_t *, u4 *);
static void readm (efile_t *, void *, size_t, size_t);
static void seekm (efile_t *, size_t);
static int read_info (env_t, efile_t *);
static int read_constant_pool (env_t, efile_t *);
#ifdef ENABLE_ACCESS_HOSTS
static int read_access_hosts (apr_pool_t *, access_hosts_t *, efile_t *);
#endif
static int setup_environment (env_t, u2, access_hosts_t *, u2, u2);
static int add_code (env_t, emethod_t *, efile_t *, unsigned int);
static int add_sourcefile (env_t, unsigned int);
static int read_interfaces (env_t, efile_t *);
static void start_fields (env_t, u2);
static efield_t * add_field (env_t, u2, access_hosts_t *, u2, u2);
static void finish_fields (env_t);
static int read_fields (env_t, efile_t *);
struct emethod_type *parse_method_type (env_t, strconst_t *);
static void start_methods (env_t, u2);
static void finish_methods (env_t);
static emethod_t * add_method (env_t, u2, struct access_hosts *, u2, u2);
static int read_methods (env_t, efile_t *);
static int resolve_super (env_t);
static int add_linenumbers (emethod_t *, uint32, efile_t *);
static int add_interfaces (env_t, u2, env_t *);
static int resolve_interfaces (env_t);
static int check_envkey_integrity (efile_t *);
static int set_field_value (env_t, efield_t *, u2);

int sizeof_type_char (char, int);
int sizeof_type_item (const char**, int);
int count_args_in_type (const char *);
int read_environment (env_t, efile_t *);
int finish_environment (env_t);
int resolve_static_fields (env_t);
int resolve_object_fields (env_t);
int build_dispatch_table (env_t);
int resolve_field_type (env_t, efield_t *);
int resolve_field_types (env_t);
int putfield (eobjectref_t *, const char *, const void *);
int resolve_constants (env_t);
int process_env_clinit (env_t);
void efile_free (efile_t *);
efile_t * efile_read (const char *);
env_t  efile_load (efile_t *);
int  e_process_file_callback (void **, void *);
enum e_method_number_type get_method_number_type (int);
int e_process_load_method (erequest_t *);
int e_process_info_method (erequest_t *);
int e_process_invokevirtual_instruction (erequest_t *);
int e_process_invokespecial_instruction (erequest_t *);
int e_process_invokestatic_instruction (erequest_t *);
int e_process_invokeinterface_instruction (erequest_t *);
int e_execute_instruction_loader (erequest_t *);
int e_execute_operation_loader (erequest_t *);
int process_return_type (erequest_t *, evalue_t *);
void print_emethod (emethod_t *);
evalue_t *getfield (eobjectref_t *, const char *);
char *eobject_to_char_buffer (eobjectref_t *, char *, int);


/** EF κ 1 Ʈ C  ϰ, ġ ̵Ѵ.  */
static void
readu1 (ef, c)
     efile_t *ef;
     u1 *c;
{
  *c = EFILE_BUFCUR_ARRAY (ef, 0);
  EFILE_BUFCUR (ef) += 1;
}

static void 
readu2 (ef, c)
     efile_t *ef;
     u2 *c;
{
  *c = (EFILE_BUFCUR_ARRAY (ef, 0) << 8) | (EFILE_BUFCUR_ARRAY (ef, 1));
  EFILE_BUFCUR (ef) += 2;
}

/** EF κ 4 Ʈ C  ϰ, ġ ̵Ѵ.  */
static void 
readu4 (ef, c)
     efile_t *ef;
     u4 *c;
{
  *c = (EFILE_BUFCUR_ARRAY (ef, 0) << 24)
    | (EFILE_BUFCUR_ARRAY (ef, 1) << 16)
    | (EFILE_BUFCUR_ARRAY (ef, 2) << 8)
    | (EFILE_BUFCUR_ARRAY (ef, 3));
  EFILE_BUFCUR (ef) += 4;
}

/** struct environment_file  LEN*SIZE ŭ о, DEST 
   Ѵ.  */
static void
readm (ef, dest, len, size)
     efile_t *ef;
     void *dest;
     size_t len;
     size_t size;
{
  int tsize = len * size;

  memcpy (dest, EFILE_BUFCUR (ef), tsize);
  EFILE_BUFCUR (ef) += tsize;
}

/** struct environment_file  ġ ̵Ų.  */
static void
seekm (ef, len)
     efile_t *ef;
     size_t len;
{
  EFILE_BUFCUR (ef) += len;
}

/** Environment Ͽ ش  unique ID    ̼  ؼ
   Environment ü Ѵ.  */
static int
read_info (e, f)
     env_t e;
     efile_t *f;
{
  u1 writer_len, license_len;
  ENVKEY_CHAR s[ENVKEY_STRSIZE];

  /* Environment  Unique ID  , efile_read  EFILE_KEY (f)  о ұ
      ٽ  ʿ .    Ѵ.  */
  seekm (f, ENVKEY_SIZE);
  PDUPLICATE_ENVKEY (ENV_MPOOL (e), EFILE_KEY (f), ENV_KEY (e));
  GET_ENVKEY_STRING (EFILE_KEY (f), s);
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "Unique Key \"[%s]\"", s);

  /* Environment   ý Ͻý   θ ´.  */
  ENV_FILEPATH (e) = apr_pstrdup (ENV_MPOOL (e), EFILE_FILEPATH (f));

  /* Environment    д´.  */
  readu1 (f, &writer_len);
  ENV_WRITER (e) = (char *) apr_pcalloc (ENV_MPOOL (e), writer_len + 1);
  readm (f, ENV_WRITER (e), writer_len, sizeof (char));
  ENV_WRITER (e)[writer_len] = '\0';
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "Writer \"[%s] [%d]\"", ENV_WRITER (e), writer_len);

  /* Environment  ̼  д´.  */
  readu1 (f, &license_len);
  ENV_LICENSE (e) = (char *) apr_pcalloc (ENV_MPOOL (e), license_len + 1);
  readm (f, ENV_LICENSE (e), license_len, sizeof (char));
  ENV_LICENSE (e)[license_len] = '\0';
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "License \"[%s] [%d]\"", ENV_LICENSE (e), license_len);

  return 0;
}

/** Environment File κ Constant Pool  оδ.  */

static int
read_constant_pool (e, f)
     env_t e;
     efile_t *f;
{
  u1 tag;
  u2 poolsize, d2f, d2s, len;
  u4 d4, d4b;
  unsigned int i;
  econst_pool_t *cst = ENV_CONSTANT_POOL (e);

  /* Constant Pool  ũⰪ ´.  */
  readu2 (f, &poolsize);

  CONSTPOOL_SIZE (cst) = poolsize;
  CONSTPOOL_TAGS (cst) = (u1 *) apr_pcalloc (ENV_MPOOL (e), sizeof (u1) * poolsize);
  if (!CONSTPOOL_TAGS (cst))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Out of memory.\n");
      return -1;
    }

  CONSTPOOL_DATA (cst) = (int *) apr_pcalloc (ENV_MPOOL (e), sizeof (int) * poolsize);
  if (!CONSTPOOL_DATA (cst))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Out of memory.\n");
      return -1;
    }

  CONSTPOOL_TAGS_IDX (cst, 0) = CONSTANT_Unknown;
  CONSTPOOL_DATA_IDX (cst, 0) = 0;

  for (i=1; i<CONSTPOOL_SIZE (cst); i++)
    {
      readu1 (f, &tag);

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		    "\tConstant[0x%02x] tag 0x%x", i, tag);

      CONSTPOOL_TAGS_IDX (cst, i) = tag;
      switch (tag)
	{
	case CONSTANT_String:
	  {
	    readu2 (f, &d2f);
	    CONSTPOOL_DATA_IDX (cst, i) = d2f;
	    break;
	  }
	case CONSTANT_Fieldref:
	case CONSTANT_Methodref:
	case CONSTANT_InterfaceMethodref:
	case CONSTANT_NameAndType:
	case CONSTANT_Sha1AndName:
	case CONSTANT_Accessref:
	case CONSTANT_Environment:
	case CONSTANT_Host:
	  {
	    readu2 (f, &d2f);
	    readu2 (f, &d2s);
	    CONSTPOOL_DATA_IDX (cst, i) = (d2s << 16) | d2f;
	    break;
	  }
	case CONSTANT_Utf8:
	  {
	    readu2 (f, &len);
	    CONSTPOOL_DATA_IDX (cst, i) = (int) new_strconst (e, EFILE_BUFCUR (f), len);
	    seekm (f, len);
	    break;
	  }
	case CONSTANT_Sha1:
	  {
	    char *key = (char *) apr_pcalloc (ENV_MPOOL (e), ENVKEY_SIZE);
	    readm (f, key, ENVKEY_SIZE, sizeof (ENVKEY_CHAR));
	    CONSTPOOL_DATA_IDX (cst, i) = (int) key;
	    break;
	  }
	case CONSTANT_Sha1This:
	  /* CONSTANT_Sha1This  , Environment  Unique ID  Ͽ, 
	     CONSTANT_Sha1  ȯѴ. */
	  CONSTPOOL_TAGS_IDX (cst, i) = CONSTANT_Sha1;
	  CONSTPOOL_DATA_IDX (cst, i) = (int) ENV_KEY (e);
	  break;
	case CONSTANT_Integer:
	case CONSTANT_Float:
	  readu4 (f, &d4);
	  CONSTPOOL_DATA_IDX (cst, i) = d4;
	  break;
	case CONSTANT_Long:
	  readu4 (f, &d4);
	  readu4 (f, &d4b);
#if SIZEOF_VOIDP == 8
	  CONSTPOOL_DATA_IDX (cst, i) = WORDS_TO_LONG (d4, d4b);
	  i++;
	  CONSTPOOL_DATA_IDX (cst, i) = 0;
#else
	  CONSTPOOL_DATA_IDX (cst, i) = d4;
	  i++;
	  CONSTPOOL_DATA_IDX (cst, i) = d4b;
#endif /* SIZEOF_VOIDP == 8 */
	  CONSTPOOL_TAGS_IDX (cst, i) = CONSTANT_Unknown;
	  break;
	case CONSTANT_Double:
	  readu4 (f, &d4);
	  readu4 (f, &d4b);
	  
#if SIZEOF_VOIDP == 8
#if defined(DOUBLE_ORDER_OPPOSITE)
	  CONSTPOOL_DATA_IDX (cst, i) = WORDS_TO_LONG (d4b, d4);
#else
	  CONSTPOOL_DATA_IDX (cst, i) = WORDS_TO_LONG (d4, d4b);
#endif /* DOUBLE_ORDER_OPPOSITE */
	  i++;
	  CONSTPOOL_DATA_IDX (cst, i) = 0;
#else
#if defined(DOUBLE_ORDER_OPPOSITE)
	  CONSTPOOL_DATA_IDX (cst, i) = d4b;
	  i++;
	  CONSTPOOL_DATA_IDX (cst, i) = d4;
#else
	  CONSTPOOL_DATA_IDX (cst, i) = d4;
	  i++;
	  CONSTPOOL_DATA_IDX (cst, i) = d4b;
#endif /* DOUBLE_ORDER_OPPOSITE */
#endif /* SIZEOF_VOIDP == 8 */
	  CONSTPOOL_TAGS_IDX (cst, i) = CONSTANT_Unknown;
	  break;
	default:
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR,
			"Invalid constant type %d in class", tag);
	  break;
	}
    }

  return 0;
}

#ifdef ENABLE_ACCESS_HOSTS
/** Environment κ Access Hosts  д´.  */
static int
read_access_hosts (mp, hosts, f)
     apr_pool_t *mp;
     access_hosts_t *hosts;
     efile_t *f;
{
  unsigned int i;
  u2 len;
  
  readu2 (f, &len);

  ACCESS_HOSTS_SIZE (hosts) = len;
  ACCESS_HOSTS_DATA (hosts) = (int *) apr_pcalloc (mp, sizeof (int) * len);

  for (i=0; i<ACCESS_HOSTS_SIZE (hosts); i++)
    {
      u2 num;

      readu2 (f, &num);
      ACCESS_HOSTS_DATA_IDX (hosts, i) = num;
    }

  return 0;
}
#endif

/** Environment ̳ʸ д ߿ ȣǰ ȴ.  */
static int
setup_environment (e, access_flags, acchost, this_environment, super_environment)
     env_t e;
     u2 access_flags, this_environment, super_environment;
     access_hosts_t *acchost;
{
  int name_index, sha1andname_index;
  econst_pool_t *pool = ENV_CONSTANT_POOL (e);
  
  if (CONSTPOOL_TAGS_IDX (pool, this_environment) != CONSTANT_Environment)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		    "this environment constant pool index is bogus. index = %d", this_environment);
      return -1;
    }

  /* Access Rights  鼭, ̰ ȣ ȴ.  */
  ENV_ACCESS_FLAGS (e) = access_flags;
  ENV_ACCESS_HOSTS (e) = acchost;
  ENV_SET_STATE (e, ENV_STATE_1_FILE_PARSING);
  ENV_THIS_INDEX (e) = this_environment;
  ENV_SUPER_INDEX (e) = super_environment;
  sha1andname_index = ENVIRONMENT_SHA1ANDNAME (pool, this_environment);
  name_index = SHA1ANDNAME_NAME (pool, sha1andname_index);
  ENV_NAME (e) = (strconst_t *) CONSTPOOL_DATA_IDX (pool, name_index);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		"Loading Environment : %s", STRCONST_PTR (ENV_NAME (e)));

  return 0;
}

static int read_attributes (env_t, efile_t *, enum read_attribute_type, void *);

static int
add_code (e, m, f, len)
     env_t e;
     emethod_t *m;
     efile_t *f;
     unsigned int len ATTRIBUTE_UNUSED;
{
  u1 *code;
  u2 max_stack, max_locals, elen, i2;
  u4 code_length;
  struct mexception *exp = NULL;

  readu2 (f, &max_stack);
  readu2 (f, &max_locals);
  readu4 (f, &code_length);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "add_code for method_code_length\tMax stack = %d\tMax locals = %d\tCode length = %d",
		 max_stack, max_locals, code_length);
  
  if (code_length > 0)
    {
      code = (u1 *) apr_pcalloc (ENV_MPOOL (e), code_length);
      if (!code)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Out of Memory.");
	  return -1;
	}

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		    "allocating bytecode 0x%p", code);

      readm (f, code, code_length, sizeof (u1));
    }
  else
    code = NULL;

  readu2 (f, &elen);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "Exception table length = %d", elen);

  if (elen > 0)
    {
      int i;

      exp = apr_pcalloc (ENV_MPOOL (e), 
			 sizeof (struct mexception) 
			 + ((elen - 1) * sizeof (struct mexception_entry)));
      MEXCEPTION_LENGTH (exp) = elen;
      for (i = 0; i < elen; i++) 
	{
	  readu2 (f, &i2);
	  MEXCEPTION_START_PC (exp, i) = i2;
	  readu2 (f, &i2);
	  MEXCEPTION_END_PC (exp, i) = i2;
	  readu2 (f, &i2);
	  MEXCEPTION_HANDLER_PC (exp, i) = i2;
	  readu2 (f, &i2);
	  MEXCEPTION_CATCH_IDX (exp, i) = i2;
	  MEXCEPTION_CATCH_TYPE (exp, i) = NULL;
	}
    }

  METHOD_BCODE (m) = code;
  METHOD_BCODE_LENGTH (m) = code_length;
  METHOD_STACK_SIZE (m) = max_stack;
  METHOD_LOCAL_SIZE (m) = max_locals;
  METHOD_EXCEPTION_TABLE (m) = exp;

  return read_attributes (METHOD_ENV (m), f, READATTR_METHOD, m);
}

static int
add_sourcefile (e, idx)
     env_t e;
     unsigned int idx;
{
  econst_pool_t *pool;
  const char *sourcefile;
  const char *basename;

  pool = ENV_CONSTANT_POOL (e);
  sourcefile = STRCONST_PTR (((strconst_t *) CONSTPOOL_DATA_IDX (pool, idx)));
  basename = strrchr (sourcefile, DIR_SEPARATOR);
  if (!basename) 
    basename = sourcefile;
  else
    basename++;

  ENV_SOURCEFILE_NAME (e) = apr_pcalloc (ENV_MPOOL (e), strlen (basename) + 1);
  if (ENV_SOURCEFILE_NAME (e) != 0)
    strcpy (ENV_SOURCEFILE_NAME (e), basename);
  else
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Out of memory.");
      return -1;
    }

  return 0;
}

/**
 * Read in line numbers assocated with code.
 */
static int
add_linenumbers (emethod_t *m, uint32 len ATTRIBUTE_UNUSED, efile_t *f)
{
  int i;
  u2 nr, data;
  struct linenumbers *lines;
  
  /* no checkBufSize, done in caller (readAttributes) */
  
  readu2(f, &nr);
  
  lines = (struct linenumbers *) apr_pcalloc (ENV_MPOOL (METHOD_ENV (m)), 
					      (sizeof (struct linenumbers) + sizeof (struct linenumber_entry) * nr));
  if (!lines)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Out of memory.");
      return -1;
    }
  
  LINENUMBER_LENGTH (lines) = nr;
  for (i = 0; i < nr; i++) 
    {
      readu2 (f, &data);
      LINENUMBER_ENTRY_START_PC (lines, i) = data;
      readu2 (f, &data);
      LINENUMBER_ENTRY_LINE_NR (lines, i) = data;
    }
  
  /* Attach lines to method */
  METHOD_LINES (m) = lines;

  return 0;
}

/**
  Method   exception  д´.
 */
static int
add_exceptions (emethod_t *m, uint32 len ATTRIBUTE_UNUSED, efile_t *f)
{
  int i;
  u2 nr, *idx;

  readu2 (f , &nr);
  if (!nr)
    return 0;

  idx = (u2 *) apr_pcalloc (ENV_MPOOL (METHOD_ENV (m)), sizeof (u2) * nr);
  if (!idx)
    return -1;

  for (i=0; i<nr; i++) 
    readu2 (f, idx + i);

  METHOD_DECLARED_EXCEPTION_NUMBER (m) = nr;
  METHOD_DECLARED_EXCEPTION (m) = idx;

  return 0;
}

static int
set_field_value (e, f, idx)
     env_t e ATTRIBUTE_UNUSED;
     efield_t *f;
     u2 idx;
{
  FIELD_CONSTANT_INDEX (f) = idx;
  FIELD_ACCESS_FLAGS (f) |= AF_CONSTANT_VALUE;

  return 0;
}

/** Environment κ attribute  о δ.  */
static int
read_attributes (e, f, attrtype, target)
     env_t e;
     efile_t *f;
     enum read_attribute_type attrtype;
     void *target;
{
  int i;
  u2 cnt;

  readu2 (f, &cnt);
  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "read_attributes: attributes_count = %d", cnt);
  
  for (i=0; i<cnt; i++)
    {
      u2 idx;
      u4 len;

      readu2 (f, &idx);
      readu4 (f, &len);

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		    "read_attributes: IDX = %d, LEN = 0x%x", idx, len);

      if (CONSTPOOL_TAGS_IDX (ENV_CONSTANT_POOL (e), idx) == CONSTANT_Utf8)
	{
	  strconst_t *sc;

	  sc = (strconst_t *) CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (e), idx);
	  if (strconstcmp (new_strconst (e, "Code", 4), sc) && attrtype == READATTR_METHOD)
	    {
	      if (add_code (e, (emethod_t *) target, f, len))
		{
		  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "add_code: error.");
		  return -1;
		}
	    }
	  else if (strconstcmp (new_strconst (e, "SourceFile", 10), sc) 
		   && attrtype == READATTR_ENVIRONMENT)
	    {
	      readu2 (f, &idx);
	      if (add_sourcefile ((env_t) target, idx))
		return -1;
	    }
	  else if (strconstcmp (new_strconst (e, "LineNumberTable", 15), sc)
		   && (attrtype == READATTR_METHOD)) 
	    {
	      if (add_linenumbers ((emethod_t *) target, len, f))
		return -1;
	    }
	  else if (strconstcmp (new_strconst (e, "Exceptions", 10), sc)
		   && (attrtype == READATTR_METHOD))
	    {
	      if (add_exceptions ((emethod_t *) target, len, f))
		return -1;
	    }
	  else if (strconstcmp (new_strconst (e, "ConstantValue", 13), sc)
		   && (attrtype == READATTR_FIELD))
	    {
	      readu2 (f, &idx);
	      set_field_value (e, (efield_t *) target, idx);
	    }
	  else if (strconstcmp (new_strconst (e, "Synthetic", 9), sc)) 
	    ; /* nothing.  */
	  else if (strconstcmp (new_strconst (e, "Deprecated", 10), sc))
	    ; /* nothing.  */
	  else if (strconstcmp (new_strconst (e, "gnu.gce.gce-compiled", 20), sc))
	    ; /* nothing.  */
	  else
	    {
	      env_log_error (ENVLOG_MARK, ENVLOG_WARNING,
			     "read_attributes: don't know how to parse. = %s", STRCONST_PTR (sc));
	      seekm (f, len);
	    }
	}
      else
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_WARNING,
			"WARNING! Skipping broken(?) attribute (name is not a Utf8 constant).");
	}
    }

  return 0;
}

static int
add_interfaces (e, count, interfaces)
     env_t e, *interfaces;
     u2 count;
{
  ENV_INTERFACES (e) = interfaces;
  ENV_INTERFACE_NUM (e) = count;

  return 0;
}

/** Environment κ Interface  оδ.  */
static int
read_interfaces (e, f)
     env_t e;
     efile_t *f;
{
  int i;
  u2 count;
  env_t *interfaces;

  readu2 (f, &count);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "read_interfaces: interfaces_count = %d", count);

  /*  interface  ó ϹǷ, 0  ƴ   Ѵ.  */
  if (count == 0) 
    return 0;

  interfaces = (env_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (env_t) * count);
  if (!interfaces)
    return -1;

  for (i=0; i<count; i++)
    {
      u2 val;

      readu2 (f, &val);
      /* Will be converted from idx to env_t in process_env () */
      interfaces[i] = ((env_t) ((size_t) val));
    }

  add_interfaces (e, count, interfaces);

  return 0;
}

/** Environment  field  ؼϱ  ؾ ϴ ۾ ⼭
   Ѵ.     environment ϳ ϴ ü field  
     ֱ , ̸  Ѵ. */
static void
start_fields (e, count)
     env_t e;
     u2 count;
{
  if (count)
    ENV_FIELDS (e) = (efield_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct efield) * count);  
  else
    ENV_FIELDS (e) = NULL;

  ENV_FIELDS_TMPCOUNT (e) = count;
  ENV_FIELDS_COUNT (e) = 0;
}

/** struct environment   ؼ field   ,  Լ 
   ȣȴ.  */
static efield_t *
add_field (e, access_flags, acchost, name_index, type)
     env_t e;
     u2 access_flags, name_index, type;
     access_hosts_t *acchost;
{
  int index;
  char *sig;
  econst_pool_t *pool = ENV_CONSTANT_POOL (e);
  efield_t *tf;
  strconst_t *sigtype;

  if (CONSTPOOL_TAGS_IDX (pool, name_index) != CONSTANT_Utf8) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		    "add_field: no field name. idx = 0x%x", name_index);
      return (0);
    }

  /* ENV_FIELDS_TMPCOUNT  start_fields ()  ó ʱȭ Ǿ ִ
     ̴.  */
  ENV_FIELDS_TMPCOUNT (e)--;

  if (access_flags & AF_STATIC)
    index = ENV_STATIC_FIELDS_COUNT (e);
  else
    index = ENV_FIELDS_TMPCOUNT (e) + ENV_STATIC_FIELDS_COUNT (e);

  /* start_fields () Ҵ struct field    Ʒ  
     Ǿ ̴.

     +---------------+--------------------------------+
     | static fields | non-static fields              |
     +---------------+--------------------------------+

     Static field  , ʿ ġϰ  ̸, ׷  ,  
     ̰ ȴ.  ̷ ݴ ̰ Ǵ  finish_fields () Լ ٽ
      ° ٽ 迭 ̴.  */
  tf = &ENV_FIELDS_IDX (e, index);

  FIELD_NAME (tf) = (strconst_t *) CONSTPOOL_DATA_IDX (pool, name_index);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "Adding field : %s", STRCONST_PTR (FIELD_NAME (tf)));

  if (CONSTPOOL_TAGS_IDX (pool, type) != CONSTANT_Utf8)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		    "add_field: no signature name. idx = 0x%x", type);
      ENV_FIELDS_COUNT (e)++;
      return NULL;
    }

  FIELD_ACCESS_FLAGS (tf) = access_flags;
  FIELD_ACCESS_HOSTS (tf) = acchost;

  sigtype = (strconst_t *) CONSTPOOL_DATA_IDX (pool, type);
  sig = STRCONST_PTR (sigtype);

  FIELD_TYPE (tf) = get_etype (sig);
  if (!FIELD_TYPE (tf))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "add_field: ߸ type.");
      ENV_FIELDS_COUNT (e)++;
      return NULL;
    }

  /*  field type  ó ʴ´.  */
  if (sig[0] == 'L' || sig[0] == '[')
    {
      /* L  [  ۵Ǵ  Object Ref Ȥ Array Ref    ε, 
	   װ ó   , ӽ   ִ ͵ ϰ
	 ߿ ⼭    ٸ ó ̷  ̴.  */
      FIELD_TYPE_CONSTPOOL_IDX (tf) = type;
    }

  ENV_FIELDS_COUNT (e)++;
  if (access_flags & AF_STATIC)
    ENV_STATIC_FIELDS_COUNT (e)++;

  return tf;
}

static void
finish_fields (e)
     env_t e;
{
  efield_t tmp;
  efield_t *fld;
  int n;
  
  /* Reverse the instance fields, so they are in "proper" order. */
  fld = &ENV_INSTANCE_FIELDS (e);
  n = ENV_INSTANCE_FIELDS_COUNT (e);
  while (n > 1) 
    {
      tmp = fld[0];
      fld[0] = fld[n-1];
      fld[n-1] = tmp;
      fld++;
      n -= 2;
    }
}

/** Environment κ FIELD  д´.  */
static int
read_fields (e, f)
     env_t e;
     efile_t *f;
{
  int i;
  u2 fields_count;

  readu2 (f, &fields_count);
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "Fields Count : %d", fields_count);

  start_fields (e, fields_count);

  for (i=0; i<fields_count; i++)
    {
      u2 access_flags;
      u2 name_index;
      u2 type;
      efield_t *this_field;
      access_hosts_t *acchost = 
	(access_hosts_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct access_hosts));

      readu2 (f, &access_flags);
#ifdef ENABLE_ACCESS_HOSTS
      if (read_access_hosts (ENV_MPOOL (e), acchost, f))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_access_hosts: error.");
	  return -1;
	}
#endif
      readu2 (f, &name_index);
      readu2 (f, &type);

      this_field = add_field (e, access_flags, acchost, name_index, type);

      if (read_attributes (e, f, READATTR_FIELD, this_field))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_attributes: error.");
	  return -1;
	}
    }

  finish_fields (e);

  return 0;
}

/**
 * Calculate size (in words) of a signature item.
 */
int
sizeof_type_char (ch, want_wide_refs)
     char ch;
     int want_wide_refs;
{
  switch (ch)
    {
    case 'V':
      return 0;
    case 'I':
    case 'Z':
    case 'S':
    case 'B':
    case 'C':
    case 'F':
      return 1;
      break;
    case 'D':
    case 'J':
      return 2;
      break;
    case '[':
    case 'L':
      return want_wide_refs ? sizeof (void*) / sizeof (int32) : 1;
    }

  return -1;
}

/**
 * Calculate size (in words) of a signature item and move *strp so
 * that it points to the next element of the signature
 */
int
sizeof_type_item (const char** strp, int want_wide_refs)
{
  int count;
  const char* str;
  
  for (str = *strp; ; str++) 
    {
      count = sizeof_type_char (*str, want_wide_refs);
      if (count == -1) 
	{
	  switch (*str) 
	    {
	    case '(':
	      continue;
	    case 0:
	    case ')':
	      break;
	    default:
	      count = -1;	/* avoid compiler warning */
	      abort ();
	    }
	} 
      else 
	{
	  while (*str == '[')
	    ++str;
	  if (*str == 'L') 
	    {
	      while (*str != ';') 
		str++;
	    }
	}
      
      *strp = str + 1;
      return (count);
    }
}

/**
  Count the number of arguments in a signature.
*/
int
count_args_in_type (type)
     const char *type;
{
  int nargs = 0;
  
  while (sizeof_type_item (&type, 0) != -1) 
    ++nargs;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "count_args_in_type: nargs = %d", nargs);

  return nargs;
}

struct emethod_type *
parse_method_type (e, type_name)
     env_t e;
     strconst_t *type_name;
{
  int nargs, count;
  const char *sig_iter;
  struct emethod_type *type;

  nargs = count_args_in_type (STRCONST_PTR (type_name));
  type = (struct emethod_type *) apr_pcalloc (ENV_MPOOL (e), 
					      sizeof (struct emethod_type) + (nargs - 1) * sizeof (MTYPE_ARGV_IDX (type, 0)));

  MTYPE_NAME (type) = type_name;

  MTYPE_ARGC (type) = nargs;
  
  sig_iter = STRCONST_PTR (type_name) + 1;	/* skip `(' */
  for (count = 0; count < nargs; count++) 
    {
      MTYPE_ARGV_IDX (type, count) = sig_iter - STRCONST_PTR (type_name);
      sizeof_type_item (&sig_iter, 0);
    }
  ++sig_iter; /* skip `)' */

  MTYPE_RETUEN (type) = sig_iter - STRCONST_PTR (type_name);
  
  return type;
}

/** Environment  method  óϱ ռ,  ־ ϴ ͵  ,
   struct environment  װ ݿѴ.  */
static void
start_methods (e, count)
     env_t e;
     u2 count;
{
  if (count)
    ENV_METHODS (e) = (struct emethod *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct emethod) * count);
  else
    ENV_METHODS (e) = NULL;

  ENV_METHODS_COUNT (e) = 0;
}

/** Environment  method  ó ,  ־ ϴ ͵  ,
   struct environment  װ ݿѴ.  */
static void
finish_methods (e)
     env_t e ATTRIBUTE_UNUSED;
{
  /*     ʾҴ.  */
}

static emethod_t *
add_method (e, access_flags, acchost, name_index, type)
     env_t e;
     u2 access_flags, name_index, type;
     struct access_hosts *acchost;
{
  econst_pool_t *pool = ENV_CONSTANT_POOL (e);
  emethod_t *tm;
  strconst_t *name, *type_name;

  /* Name Index  Type  .  */
  if (CONSTPOOL_TAGS_IDX (pool, name_index) != CONSTANT_Utf8) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "add_method: no method name.");
      return (0);
    }

  if (CONSTPOOL_TAGS_IDX (pool, type) != CONSTANT_Utf8)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "add_method: no signature name.");
      ENV_METHODS_COUNT (e)++;
      return (0);
    }

  type_name = (strconst_t *) CONSTPOOL_DATA_IDX (pool, type);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		"Adding method #%d: %d %d (%x)", ENV_METHODS_COUNT (e), name_index, type, access_flags);

  tm = &ENV_METHODS_IDX (e, ENV_METHODS_COUNT (e));

  /* ش Method     ִ´.  */
  name = (strconst_t *) CONSTPOOL_DATA_IDX (pool, name_index);
  
  METHOD_INDEX (tm) = ENV_METHODS_COUNT (e)++;
  METHOD_NAME (tm) = name;
  METHOD_TYPE (tm) = parse_method_type (e, type_name);
  METHOD_ENV (tm) = e;
  METHOD_STACK_SIZE (tm) = 0;
  METHOD_LOCAL_SIZE (tm) = 0;
  METHOD_ACCESS_FLAGS (tm) = access_flags;
  METHOD_ACCESS_HOSTS (tm) = acchost;

  /*  method ̸ constructor  , access flag  Ѵ.  */
  if (strconstcmp (name, new_strconst (e, "<cinit>", 7)))
    METHOD_ACCESS_FLAGS (tm) |= AF_CONSTRUCTOR;

  return tm;
}

/** Environment κ method  о δ.  */
static int
read_methods (e, f)
     env_t e;
     efile_t *f;
{
  int i;
  u2 methods_count;

  readu2 (f, &methods_count);
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "Methods Count: %d", methods_count);

  start_methods (e, methods_count);

  for (i = 0; i < methods_count; i++) 
    {
      u2 access_flags;
      u2 name_index;
      u2 type;
      emethod_t *this_methods;
      access_hosts_t *acchost = 
	(struct access_hosts *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct access_hosts));

      readu2 (f, &access_flags);
#ifdef ENABLE_ACCESS_HOSTS
      if (read_access_hosts (ENV_MPOOL (e), acchost, f))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_access_hosts: error.");
	  return -1;
	}
#endif
      readu2 (f, &name_index);
      readu2 (f, &type);

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		    "read_methods: access flags = 0x%x, name index = 0x%x, type = 0x%x", access_flags, name_index, type);
      this_methods = add_method (e, access_flags, acchost, name_index, type);

      if (read_attributes (e, f, READATTR_METHOD, this_methods))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_attributes: error.");
	  return -1;
	}
    }

  finish_methods (e);

  return 0;
}

/** Environment ü   ؼ   ־  κ  , ⼭
   óѴ.  */
int
finish_environment (e)
     env_t e;
{
  ENV_SET_STATE (e, ENV_STATE_2_FINISHED_FILE_PARSING);

  return 0;
}

/** Environment  EF κ о, THIS  ϴ entry point ̴.
   Environment File Format  ؼ ``Environment File Format''
   丮 ϶.  */
int
read_environment (e, f)
     env_t e;
     efile_t *f;
{
  u4 magic;
  u2 minor_version, major_version;
  u2 access_flags, this_environment, super_environment;
  access_hosts_t *acchost;

  /* Magic Number  д´.  */
  readu4 (f, &magic);
  if (magic != ENVIRONMENT_MAGIC)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Bad magic number 0x%x", magic);
      return -1;
    }

  /* Minor/Major Version  д´.  */
  readu2 (f, &minor_version);
  readu2 (f, &major_version);

  if (!(major_version == MAJOR_VERSION_V1_0 && minor_version == MINOR_VERSION_V1_0))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_WARNING,
		     "Unrecognized version number %d/%d in classfile.", major_version, minor_version);
      return -1;
    }
  
  /* Environment  , ̼  ߿ UNIQUE ID  óѴ.  */
  if (read_info (e, f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_info: error.");
      return -1;
    }

  /* Constant Pool  struct environment  д´.  */
  if (read_constant_pool (e, f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_constant_pool: error.");
      return -1;
    }

  /* Access Flags  Hosts  Name  о environment  Ѵ.  */
  acchost = (access_hosts_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct access_hosts));
  
  readu2 (f, &access_flags);
#ifdef ENABLE_ACCESS_HOSTS
  if (read_access_hosts (ENV_MPOOL (e), acchost, f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_access_hosts: error.");
      return -1;
    }
#endif
  readu2 (f, &this_environment);
  readu2 (f, &super_environment);
  
  if (setup_environment (e, access_flags, acchost, this_environment, super_environment))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "setup_environment: error.");
      return -1;
    }

  if (read_interfaces (e, f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_interfaces: error.");
      return -1;
    }

  if (read_fields (e, f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_fields: error.");
      return -1;
    }

  if (read_methods (e, f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_methods: error.");
      return -1;
    }

  if (read_attributes (e, f, READATTR_ENVIRONMENT, e))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "read_attributes: error.");
      return -1;
    }

  return finish_environment (e);
}

/** emethod  Ѵ.  */
void
print_emethod (m)
     emethod_t *m;
{
  if (!m)
    {
      printf ("<null>");
      return;
    }

  printf ("<emethod ");
  printf ("index %d ", METHOD_INDEX (m));
  printf ("access_flags %d ", METHOD_ACCESS_FLAGS (m));
  printf ("name \"%s\" ", STRCONST_PTR (METHOD_NAME (m)));
  printf ("type \"%s\"\n", STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (m))));

  printf ("    stack_size %d ", METHOD_STACK_SIZE (m));
  printf ("local_size %d ", METHOD_LOCAL_SIZE (m));

  printf (">");
  fflush (stdout);
}


/**
   Object field  , ش  object ref ü  ؾ ϱ ,
   ̿  index ó    offset  ̷ Ѵ. 

   TODO: field  TYPE ũ⸦  offset  ϴ  ؾ ޸
   뷮 ҽų  ִ.
*/
int
resolve_object_fields (e)
     env_t e;
{
  efield_t *ifs;
  int nifs, offset;

  offset = sizeof (struct eobjectref);

  ifs = &ENV_INSTANCE_FIELDS (e);
  nifs = ENV_INSTANCE_FIELDS_COUNT (e);

  if (!nifs)
    return 0;

  ENV_FIELDS_SIZE_IN_BYTE (e) = 
    (int) ENV_INSTANCE_FIELDS_COUNT (e) * sizeof (struct evalue);

  /* Field     Object ref  dispatch table ٷ ؿ ȴ.  */
  for (; --nifs >= 0; ifs++) 
    {
      FIELD_OBJECT_DATA_OFFSET (ifs) = offset;
      offset += sizeof (struct evalue);
    }

  return 0;
}

/** THIS  ϴ static FIELD    static_data  Ҵϰ
    field    INDEX ȣ (static   )  Ѵ.  */
int
resolve_static_fields (e)
     env_t e;
{
  efield_t *fld = ENV_STATIC_FIELDS (e);
  int sfc = ENV_STATIC_FIELDS_COUNT (e);
  int index = 0;

  if (!sfc)
    return 0;
  else if (sfc < 0)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		    "resolve_static_fields: static field   ̴.");
      return -1;
    }

  ENV_STATIC_FIELDS_DATA (e) = 
    (evalue_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct evalue) * sfc);

  for (; --sfc >= 0; fld++) 
    {
      if (!(FIELD_ACCESS_FLAGS (fld) & AF_STATIC))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR,
			"resolve_static_fields: óϰ ϴ field  static  ƴϴ.");
	  return -1;
	}

      /*  field  ConstantValue  Ǿ  , տ FIELD_CONSTANT_INDEX 
	 Ǿ  ̴.  ̰ 켱 ó ־ Ѵ.  */
      if (FIELD_ACCESS_FLAGS (fld) & AF_CONSTANT_VALUE)
	{
	  econst_pool_t *cst = ENV_CONSTANT_POOL (e);
	  u2 constidx = FIELD_CONSTANT_INDEX (fld);

	  switch (CONSTPOOL_TAGS_IDX (cst, constidx))
	    {
	    case CONSTANT_Integer:
	      {
		switch (ETYPE_KIND (FIELD_TYPE (fld)))
		  {
		  case ETYPE_BOOLEAN:
		  case ETYPE_BYTE:
		    EVALUE_BYTE (ENV_STATIC_FIELDS_DATA (e) + index) 
		      = CONSTPOOL_INT (cst, constidx);
		    break;
		  case ETYPE_CHAR:
		  case ETYPE_SHORT:
		    EVALUE_SHORT (ENV_STATIC_FIELDS_DATA (e) + index) 
		      = CONSTPOOL_INT (cst, constidx);
		    break;
		  default:
		    EVALUE_INT (ENV_STATIC_FIELDS_DATA (e) + index) 
		      = CONSTPOOL_INT (cst, constidx);
		    break;
		  }
		break;
	      }
	    case CONSTANT_Float:
	      EVALUE_INT (ENV_STATIC_FIELDS_DATA (e) + index) 
		= CONSTPOOL_INT (cst, constidx);
	      break;
	    case CONSTANT_Long:
	    case CONSTANT_Double:
	      EVALUE_LONG (ENV_STATIC_FIELDS_DATA (e) + index) 
		= CONSTPOOL_LONG (cst, constidx);
	      break;
	    case CONSTANT_String:
	      resolve_string (e, constidx);
	      /* ... fall through ... */
	    case CONSTANT_EnvedString:
	      EVALUE_ADDRESS (ENV_STATIC_FIELDS_DATA (e) + index)
		= (eref_t) CONSTPOOL_DATA_IDX (cst, constidx);
	      break;
	    }
	}

      FIELD_STATIC_DATA_INDEX (fld) = index++;
    }

  return 0;
}

/** Dispatch table  ʿ  environment  method  ϱ ؼ
   ׻ ش method  ּҸ ϴ  ŷӱ , ش struct method
     pointer  struct dispatch_table  ־ ν,  ϱ
   ؼ ̴.  */
int
build_dispatch_table (e)
     env_t e;
{
  int i;
  struct dispatch_table *d;
  emethod_t *meth;

  meth = ENV_METHODS (e);
  i = ENV_METHODS_COUNT (e);
  for (; --i >= 0; meth++)
    {
      if (METHOD_IS_STATIC (meth) || METHOD_IS_PRIVATE (meth)
	  || strconstcmp (METHOD_NAME (meth), new_strconst (e, "<cinit>", 7)))
	{
	  /* METHOD_INDEX  read_environment.c  add_method () Լ  Ǹ,
	     ⿡ dispatch table   ʿ䰡  method  ؼ INDEX 
	     -1  Ѵ.  */
	  METHOD_INDEX (meth) = -1;
	  continue;
	}
    }

  d = (struct dispatch_table *) apr_pcalloc (ENV_MPOOL (e), 
					     sizeof (struct dispatch_table) + (ENV_METHODS_COUNT (e) * sizeof (void *)));
  if (!d)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "build_dispatch_table: out of memory.");
      return -1;
    }
  
  DISPATCH_ENV (d) = e;

  for (i=0; i<ENV_METHODS_COUNT (e); i++)
    DISPATCH_METHOD_IDX (d, i) = (void *) &ENV_METHODS_IDX (e, i);

  ENV_DISPATCH_TABLE (e) = d; 

  return 0;
}

/** Environment ü E  ϴ field FI  type  resolve ǵ Ѵ.  */
int
resolve_field_type (e, fi)
     env_t e ATTRIBUTE_UNUSED;
     efield_t *fi;
{
  etype_t *type = FIELD_TYPE (fi);

  if (ETYPE_IS_RESOLVED (type))
    return 0;

  /* TODO:   type  resolve Ǿٰ   ׷  찡
     Ŀ ߻  ִ.  */
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "resolve_field_type: resolve   type  մϴ.");

  return 0;
}

/** Environment ü E   ִ field  type   resolve ǵ
   Ѵ.   L   ġ argument   , ش  load Ѵ.  */
int
resolve_field_types (e)
     env_t e;
{
  int i;
  efield_t *fld;

  fld = ENV_FIELDS (e);

  for (i=0; i<ENV_FIELDS_COUNT (e); fld++, i++)
    if (resolve_field_type (e, fld))
      return -1;

  return 0;
}

/**
   OBJREF   NAME  field  ϴ Ȯ , ش efield_t  ȯѴ.

   !! ش field    Ҹ pointer   ֱ , efield size 
    Ǳ ؾ Ѵ.

   !!  objref + offset ּҸ Ű  ش   Ű ִ 
    ƴϴ.
*/
evalue_t *
getfield (objref, name)
     eobjectref_t *objref;
     const char *name;
{
  env_t e;
  efield_t *field;
  evalue_t *ret;

  if (!objref || !name)
    return NULL;

  e = EOBJECTREF_ENV (objref);
    
  field = lookup_field (e, new_strconst (e, name, strlen (name)), NON_STATIC_FIELD_FLAG);
  if (!field || (field && !FIELD_TYPE (field)))
    return NULL;

  ret = (evalue_t *) (((unsigned char *) (objref)) + (off_t) FIELD_OBJECT_DATA_OFFSET (field));

  /* ȯ objref + offset ּ̱ , caller  field type    
      Ѵ.  */
  return ret;
}

/**
   OBJREF  Ű environment  field NAME   VALUE  ִ´.
   ⼭ VALUE    ִ  ּ ν, ش ּҰ ִ Ʈ, 
   NAME  Ͽ  FIELD  SIZE ŭ ϰ ȴ.  
*/
int
putfield (objref, name, value)
     eobjectref_t *objref;
     const char *name;
     const void *value;
{
  evalue_t *t;
  efield_t *field;
  env_t e;

  if (!objref || !value || !name)
    return -1;

  e = EOBJECTREF_ENV (objref);
    
  field = lookup_field (e, new_strconst (e, name, strlen (name)), NON_STATIC_FIELD_FLAG);
  if (!field || (field && !FIELD_TYPE (field)))
    return -1;

  /* TODO:  FIELD_OBJECT_DATA_OFFSET  offset  struct evalue ü ũ⿡
        ִ.  ̴ ٶ  ƴϸ,   ʿϰ  Ѵ.
     ׷ Ʒ  å   õѴ.  

     EFIELD  ׻ type   ֱ , Field object  data    
     ִ   , ش ũ   ˼ִ.  ׷  Field object 
       ش field   ° Ҵϰ ش offset  ϰ ϸ   
     , 䵵 Ȯϰ    ̴.  */

  t = (evalue_t *) ((unsigned char *) (objref) + (off_t) FIELD_OBJECT_DATA_OFFSET (field));

  switch (ETYPE_KIND (FIELD_TYPE (field)))
    {
    case ETYPE_LONG:
    case ETYPE_DOUBLE:
      memcpy ((void *) &EVALUE_DOUBLE (t), value, ETYPE_SIZE (FIELD_TYPE (field)));
      break;
    case ETYPE_ARRAYREF:
    case ETYPE_OBJECTREF:
    case ETYPE_INT:
    case ETYPE_FLOAT:
      memcpy ((void *) &EVALUE_ADDRESS (t), value, ETYPE_SIZE (FIELD_TYPE (field)));
      break;
    case ETYPE_SHORT:
    case ETYPE_CHAR:
      memcpy ((void *) &EVALUE_SHORT (t), value, ETYPE_SIZE (FIELD_TYPE (field)));
      break;
    case ETYPE_BOOLEAN:
    case ETYPE_BYTE:
      memcpy ((void *) &EVALUE_BYTE (t), value, ETYPE_SIZE (FIELD_TYPE (field)));
      break;
    default:
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "putfield: ó   type.");
      return -1;
    }

  return 0;
}

eobjectref_t *
char_to_eobject (str, len)
     char *str;
     int len;
{
  int zero = 0;
  eobjectref_t *so = NULL;
  earrayref_t *array = NULL;
  etype_t *t;

  /* [C 迭 object (, arrayref)  Ѵ.  */
  array = new_arrayref (lookup_etype ((char *) "C"), (size_t) len);
  if (!array)
    return NULL;

  memcpy (EARRAYREF_DATA (array), str, len);

  /* 켱 String Type   , String Objectref  Ѵ.  */
  t = get_etype ((char *) ENV_STRING_TYPE);
  if (!t)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		     "char_to_eobject: String Type  ã  ϴ.");
      return NULL;
    }

  so = new_objectref (ETYPE_U_ENV (t));
  if (!so)
    return NULL;

  /* String->value  String->count ʵ忡   Ѵ.  */
  if (putfield (so, "data", (const void *) &array))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		     "char_to_eobject: string object  `value' field  ã ");
      return NULL;
    }
  if (putfield (so, "count", (const void *) &len))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		     "char_to_eobject: string object  `const' field  ã ");
      return NULL;
    }
  if (putfield (so, "boffset", (const void *) &zero))
    return NULL;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "char_to_eobject: Ϸ -> string \"%s\" length %d", str, len);

  return so;
}

/**
    ڿ String Environment  Ѵ.  
*/
eobjectref_t *
strconst_to_eobject (str)
     strconst_t *str;
{
  return char_to_eobject (STRCONST_PTR (str), STRCONST_LENGTH (str));
}

/**
   String Object  C ڿ Ѵ.
*/
char *
eobject_to_char_buffer (so, cs, len)
     eobjectref_t *so;
     char *cs;
     int len;
{
  char *chrs;
  eint_t offset, count;
  earrayref_t *array;
  evalue_t *ret;

  if (len <= 0)
    return NULL;

  if (!so)
    {
      cs[0] = '\0';
      return NULL;
    }

  ret = getfield (so, "data");
  array = (earrayref_t *) EVALUE_ADDRESS (ret);
  ret = getfield (so, "boffset");
  offset = (eint_t) EVALUE_INT (ret);
  ret = getfield (so, "count");
  count = (eint_t) EVALUE_INT (ret);

  env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
		 "eobject_to_char_buffer:  getfield () Լ ȯ field type   ʴ´.");

  if (!array || offset < 0 || count < 0)
    {
      cs[0] = '\0';
      return NULL;
    }

  chrs = ((unsigned char *) EARRAYREF_DATA (array)) + offset;

  len--;
  if (len > count)
    len = count;

  cs[len] = '\0';
  while (--len >= 0)
    *cs++ = (char) *chrs++;
  
  return cs;
}

/** Convert a CONSTANT_String entry in the constant poool from utf8 to env.core.string */
eobjectref_t *
resolve_string (e, idx)
     env_t e;
     int idx;
{
  strconst_t *str;
  econst_pool_t *pool = ENV_CONSTANT_POOL (e);
  eobjectref_t *so = NULL;
  
  switch (CONSTPOOL_TAGS_IDX (pool, idx))
    {
    case CONSTANT_String:
      /* CONSTANT_String  ,  ڿ   ִ CONSTANT_Utf8   
	 index ȣ  ִ.  */
      str = (strconst_t *) CONSTPOOL_DATA_IDX (pool, CONSTPOOL_DATA_IDX (pool, idx));
      so = strconst_to_eobject (str);
      if (!so)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "strconst_to_eobject: error!!!");
	  return NULL;
	}
      
      CONSTPOOL_DATA_IDX (pool, idx) = (int) so;
      CONSTPOOL_TAGS_IDX (pool, idx) = CONSTANT_EnvedString;
      break;
    case CONSTANT_EnvedString:
      so = (eobjectref_t *) CONSTPOOL_DATA_IDX (pool, idx);
      break;
    default:
      break;
    }

  return so;
}

/** THIS  CONSTANT POOL  ʱȭ Ѵ.  ʱȭ Ѵٴ  CONSTANT_String  
   CONSTANT_Environment   load ǰų ȯǾ ϴµ, װ  Լ
   Ѵٴ ̴. 
   
     , 0  ׷   -1  ȯѴ.  */
int
resolve_constants (e)
     env_t e;
{
  unsigned int idx;
  strconst_t *str;
  econst_pool_t *pool;

  pool = ENV_CONSTANT_POOL (e);
  for (idx = 0; idx < CONSTPOOL_SIZE (pool); idx++) 
    {
      switch (CONSTPOOL_TAGS_IDX (pool, idx))
	{
	case CONSTANT_String:
	  {
	    eobjectref_t *so;

	    /* CONSTANT_String  ,  ڿ   ִ CONSTANT_Utf8   
	       index ȣ  ִ.  */
	    str = (strconst_t *) CONSTPOOL_DATA_IDX (pool, CONSTPOOL_DATA_IDX (pool, idx));
	    so = strconst_to_eobject (str);
	    if (!so)
	      {
		env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
			       "resolve_constants: out of memory!!!");
		return -1;
	      }

	    CONSTPOOL_DATA_IDX (pool, idx) = (int) so;
	    CONSTPOOL_TAGS_IDX (pool, idx) = CONSTANT_EnvedString;
	    break;
	  }
	case CONSTANT_Environment:
	  /* TODO:  κп  ٸ environment    ִ  迡  Ʈ
	      ʿ ̴.  */
	  if (!get_envstruct_with_idx (idx, e))
	    {
	      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
			     "resolve_constants: get_env .\n");
	      return -1;
	    }
	  break;
	}
    }

  return 0;
}

/** E  ʱȭ Լ <clinit>  ȣϿ, environment  LOAD Ǿ ,
   ؾ ൿ   ֵ Ѵ.  

   @return  <clinit>  Ǵ   Լ Ǿٸ 1  ȯϰ,
            ִٸ -1  ȯѴ.  ׷ ʴٸ 0  ȯѴ.
*/
int
process_env_clinit (e)
     env_t e;
{
  emethod_t *m;

  /*  ٸ μ󿡼  environment   ٽ ʱȭϰų, 
     Ŀ ؼ <clinit>  Ű ߿  ٽ  Լ ̸
      ,     ϱ   ؾ Ѵ.

     TODO:  ϳ μ <clinit>  óϴ  ٸ μ󿡼
     ÿ <clinit>  óϰ    óؾ Ѵ.   ׷
     ȵǾ ִ.  */
  if (ENV_STATE (e) == ENV_STATE_9_PROCESS_CLINIT)
    return 1;

  ENV_SET_STATE (e, ENV_STATE_9_PROCESS_CLINIT);

  m = find_method (e, new_strconst (e, "<clinit>", 8), new_strconst (e, "()V", 3));
  if (!m)
    return -1;

  unlock_envstruct (e);

  if (call_method_a (m, m, NULL, NULL, NULL))
    return -1;

  lock_envstruct (e);

  return 0;
}

/**  env_t E  super  εѴ.  */
static int
resolve_super (e)
     env_t e;
{
  if (ENV_SUPER_INDEX (e) < 0)
    return -1;

  if (!ENV_SUPER_INDEX (e))
    return 0;

  ENV_SUPER (e) = get_envstruct_with_idx (ENV_SUPER_INDEX (e), e);

  return 0;
}

static int
resolve_interfaces (e)
     env_t e;
{
  int i, j, k, len = ENV_INTERFACE_NUM (e), totalilen;
  env_t *newifaces, s = ENV_SUPER (e);
  etype_t *objtype = lookup_etype ((char *) ENV_OBJECT_TYPE);

  /*  objtype  NULL  ,  Object eclass  loading  ̴
     ̷    ʴ´.  */
  if (!objtype)
    return 0;

  /* Load all the implemented interfaces.   E  interface  
     Super Ŭ ݵ env.core.Object ̿ Ѵ.  ׷   
     eclass ̴.  */
  if (s && s != ETYPE_U_ENV (objtype))
    {
      if (ENV_IS_INTERFACE (e))
	{
	  /* postException(einfo, JAVA_LANG(VerifyError));  */
	  return -1;
	}
      len += ENV_TOTAL_INTERFACE_NUM (ENV_SUPER (e));
    }
  
  for (i=0; i<ENV_INTERFACE_NUM (e); i++)
    {
      int in = (int) ENV_INTERFACE_IDX (e, i);

      ENV_INTERFACE_IDX (e, i) = get_envstruct_with_idx (in, e);
      if (!ENV_INTERFACE_IDX (e, i))
	return -1;
      len += ENV_TOTAL_INTERFACE_NUM (ENV_INTERFACE_IDX (e, i));
    }

  totalilen = len;

  /* We rebuild a list of *all* interfaces this class can use */
  if (ENV_INTERFACE_NUM (e) != totalilen)
    {
      newifaces = (env_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (env_t) * totalilen);
      if (!newifaces)
	return -1;
      for (i=0; i<ENV_INTERFACE_NUM (e); i++)
	newifaces[i] = ENV_INTERFACE_IDX (e, i);
      s = ENV_SUPER (e);
      if (s && s != ETYPE_U_ENV (objtype))
	{
	  for (j=0; j<ENV_TOTAL_INTERFACE_NUM (s); j++, i++)
	    newifaces[i] = ENV_INTERFACE_IDX (s, i);
	}
      for (k=0; k<ENV_INTERFACE_NUM (e); k++) 
	{
	  s = ENV_INTERFACE_IDX (e, k);
	  for (j=0; j<ENV_TOTAL_INTERFACE_NUM (s); j++, i++)
	    newifaces[i] = ENV_INTERFACE_IDX (s, j);
	}
      ENV_INTERFACES (e) = newifaces;
    }

  /* don't set total_interface_len before interfaces to avoid
     having walkClass attempting to walk interfaces.  */
  ENV_TOTAL_INTERFACE_NUM (e) = totalilen;

  return 0;
}

/** Environment  LOAD  ,  environment  Ȱϰ óϱ ؼ
   ʿ غ ۾  ٸ environment   ʿ  óؾ
   Ѵ.  ó  0 , н -1  ȯѴ.  

   @param e	óϰ ϴ environment ü.
   @param complete
	          0  , ENV_STATE_5_PROCESS_METHODS ܰ
                ó Ѵ.   1  , ENV_STATE_COMPLETE  ɶ
		ó Ѵ.
*/
int
process_env (e, complete)
     env_t e;
     int complete;
{
  int rv;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "process_env: <%s> state = %d, complete %d", 
		 STRCONST_PTR (ENV_NAME (e)), ENV_STATE (e), complete);

  if (ENV_STATE (e) == ENV_STATE_COMPLETE)
    return 0;

  lock_envstruct (e);

  if (ENV_STATE (e) < ENV_STATE_4_PROCESS_FIELDS)
    {
      ENV_SET_STATE (e, ENV_STATE_4_PROCESS_FIELDS);

      if (resolve_field_types (e))
	goto process_failed;
      
      /* Static  ƴ field  óѴ.  ̷ field  , Object ref  
	  ش Object ref     , ̿  index ó  
	  offset  ̷ Ѵ.  */
      if (resolve_object_fields (e))
	goto process_failed;
      
      /* THIS   static field    Ҵϰ,  field   INDEX 
	 Ѵ.  */
      if (resolve_static_fields (e))
	goto process_failed;
    }

  if (ENV_STATE (e) < ENV_STATE_5_PROCESS_METHODS)
    {
      ENV_SET_STATE (e, ENV_STATE_5_PROCESS_METHODS);

      /*   environment  interface  ƴ , method     
	 ؼ, Dispatch Table  Ѵ.  */
      if (build_dispatch_table (e))
	goto process_failed;

      if (resolve_interfaces (e))
	goto process_failed;
    }

  if (complete && ENV_STATE (e) < ENV_STATE_6_PROCESS_CONSTANTS)
    {
      ENV_SET_STATE (e, ENV_STATE_6_PROCESS_CONSTANTS);

      /* Environment  ԵǾ ִ CONSTANT POOL  ʱȭǾ ϴ ͵
	 ʱȭѴ.  ʱȭǾ ϴ Ϳ CONSTANT_String  CONSTANT_Environment 
	  Ѵ.  */
      if (resolve_constants (e))
	goto process_failed;
    }

  if (complete && ENV_STATE (e) < ENV_STATE_7_PROCESS_SUPER)
    {
      ENV_SET_STATE (e, ENV_STATE_7_PROCESS_SUPER);

      if (resolve_super (e))
	goto process_failed;

      /*  ߴٸ ENV_STATE_8_USABLE  Ѵ.  
	 TODO: ġ ´ ɱ?  */
      ENV_SET_STATE (e, ENV_STATE_8_USABLE);
    }

  if (complete && ENV_STATE (e) < ENV_STATE_9_PROCESS_CLINIT)
    {
      /* Environment  ε Ǹ鼭 ؾ ϴ ʱȭڸ Ѵ.   ȯ 1  ,
	 ٷ  ,  COMPLETE   ʴ´.  */
      if ((rv = process_env_clinit (e)) == 1)
	goto process_succeed;

      /*  ߴٸ ENV_STATE_COMPLETE  Ѵ.  
	 TODO: ġ ´ ɱ?  */
      ENV_SET_STATE (e, ENV_STATE_COMPLETE);
    }

 process_succeed:
  unlock_envstruct (e);

  return 0;

  /* ERROR ߻: ó н  Ѿ´.  */
 process_failed:
  env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		 "process_env: <%s> ó  ߻", STRCONST_PTR (ENV_NAME (e)));

  ENV_SET_STATE (e, ENV_STATE_FAILED);
  unlock_envstruct (e);

  return -1;
}

/** efile_t ü 鼭 Ҵߴ ޸𸮸 Ѵ.  */
void
efile_free (f)
     efile_t *f;
{
  env_free (EFILE_FILEPATH (f));
  FREE_ENVKEY (EFILE_KEY (f));
  env_free (EFILE_BUFBASE (f));
  env_free (f);
}

/*  unique key  file buffer  Ἲ ˻Ѵ.  ٸ ,  ̻ ؼ ȵȴ.  */
static int
check_envkey_integrity (f)
     efile_t *f;
{
  int size;
  char *tbb, tmpbase64buf[256];
  ENVKEY_CHAR s2[ENVKEY_SIZE];
  envkey_t tmpkey;
  
  tbb = tmpbase64buf;
  apr_sha1_base64 (EFILE_BUFBASE (f), EFILE_BUFLEN (f), tbb);
  if (!strncmp (tbb, "{SHA}", 5))
    tbb += 5;
  size = apr_base64_decode_binary (s2, tbb);
  PUT_ENVKEY (s2, &tmpkey);
  if (size != ENVKEY_SIZE
      || (size == ENVKEY_SIZE && compare_envkey (&tmpkey, EFILE_KEY (f))))
    return -1;

  return 0;
}

/** PATH   environment file  о efile_t ü ϼѴ.  */
efile_t *
efile_read (path)
     const char *path;
{
  efile_t *f;
  FILE *fp;
  envkey_t key;
  size_t read_bytes;
  ENVKEY_CHAR s[ENVKEY_SIZE];

  /* ݵ PATH   κ Ͽ Ѵ.  */
  if (!path || (path && path[0] != DIR_SEPARATOR))
    return NULL;

  if (!(fp = fopen (path, "rb")))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "efile_read: fopen error.\n");
      return NULL;
    }

  f = (efile_t *) env_malloc (sizeof (efile_t));
  memset (f, '\0', sizeof (efile_t));

  if (fseek (fp, 0L, SEEK_END))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "efile_read: fseek error.\n");
      env_free (f);
      fclose (fp);
      return NULL;
    }

  EFILE_FILEPATH (f) = env_strdup (path);
  EFILE_TYPE (f) = EFILE_BYTEARRAY;
  EFILE_BUFLEN (f) = (int) ftell (fp);
  EFILE_BUFBASE (f) = EFILE_BUFCUR (f) = (unsigned char *) env_malloc (EFILE_BUFLEN (f));

  fseek (fp, 0L, SEEK_SET);
  read_bytes = fread (EFILE_BUFBASE (f), 1, EFILE_BUFLEN (f), fp);
  if (read_bytes != EFILE_BUFLEN (f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "efile_read: fread error.\n");
      env_free (EFILE_BUFBASE (f));
      env_free (f);
      fclose (fp);      
      return NULL;
    }

  if (EFILE_BUFLEN (f) < EFILE_UNIQUE_ID_STARTING_POINT + ENVKEY_SIZE)
    {
      env_free (EFILE_BUFBASE (f));
      env_free (f);
      fclose (fp);
      return NULL;
    }

  /* Unique KEY  а, ENVKEY_SIZE ũ ŭ 0  Ѵ.  */
  memcpy (s, EFILE_BUFBASE (f) + EFILE_UNIQUE_ID_STARTING_POINT, ENVKEY_SIZE);
  PUT_ENVKEY (s, &key);
  DUPLICATE_ENVKEY (&key, EFILE_KEY (f));
  memset (EFILE_BUFBASE (f) + EFILE_UNIQUE_ID_STARTING_POINT, '\0', ENVKEY_SIZE);

  if (check_envkey_integrity (f))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_INFO, "check_key_integrity:  ü uid   uid  ٸϴ.");
      efile_free (f);
      fclose (fp);

      return NULL;      
    }

  fclose (fp);

  return f;
}

/** F   ִ   E ̳ʸ load Ѵ.  Լ ϴ  
    .

     1. Environment Binary Format    parsing Ѵ.
     2. Parsing  ϷǸ, ش ̳ʸ  ׿ ˸° óѴ.   environment
        status   ر ϵ   Ѵ.  

    Ȯϰ  óǾٸ 0  ׷ ʴٸ -1  ȯѴ.  */
env_t 
efile_load (f)
     efile_t *f;
{
  int rv;
  env_t e;

  /* 켱 Global Environment Infomation   load ϰ ϴ environment  
     ϴ ȮѴ.   , ش environment pointer  ״ ϰ ׷
      ,  load Ѵ.  */
  e = get_envstruct (EFILE_KEY (f));

  if (!ENV_MPOOL (e))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "get_envstruct: memory pool  NULL.");
      return NULL;
    }
    
  /*  get_envstruct () Լ   env    environment 
     ü , ENV_STATE (env)  ENV_STATE_0_STRUCT_INITIALIZED  ̴.  */
  if (ENV_STATE (e) == ENV_STATE_0_STRUCT_INITIALIZED)
    {
      lock_envstruct (e);
      rv = read_environment (e, f);
      unlock_envstruct (e);
      if (rv)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
			 "read_efile_format: efile  д   ߻.");
	  return NULL;
	}
     
      /* Load  environment process ó.  */
      process_env (e, 0);
    }

  /* Binary   ó , ȣ ־ ϴ  ȣѴ.  */

  return e;
}

/**
   PATH  ϴ eclass  о load Ѵ.
*/
env_t
efile_readnload (path)
     const char *path;
{
  efile_t *f;
  env_t e;

  /* PATH  ϴ environment Ͽ   ä, ش ü
     ȯѴ.  */
  if (!(f = efile_read (path)))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
		     "efile_read: %s  µ, Ͽϴ.", path);
      return NULL;
    }

  /* F   ؼ  load Ѵ.  */
  e = efile_load (f);
  if (!e)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "efile_load: ߸ E .");
      efile_free (f);
      return NULL;
    }

  efile_free (f);

  return e;
}

/** ȯ TRUE    hash table  travel ϰ, FALSE   ߰ ǹǷ
    hash table entry   ؼ ׻ ȯ 1 ̿ Ѵ.  */
int 
e_process_file_callback (slotarg, info)
     void **slotarg;
     void *info ATTRIBUTE_UNUSED;
{
  char *path;
  //erequest_t *r = (erequest_t *) info;
  rfc1867_uploaded_files_t *entry, **slot = (rfc1867_uploaded_files_t **) slotarg;

  if (!slot)
    return 1;

  entry = *slot;
  if (!entry)
    return 1;

  /*  slot ̳ entry  NULL  ,  ȯ ̴.  */

  path = env_strndup (RFC1867_BUF (entry), RFC1867_LENGTH (entry));

  if (!efile_readnload (path))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "efile_readnload: return value is NULL.");
      env_free (path);
      return 0;
    }

  env_free (path);

  return 1;
}

/**  request  method number   ش з Ѵ.  */
enum e_method_number_type
get_method_number_type (mn)
     int mn;
{
  if (mn == ESERVER_METHOD_ARRAY (eserver, E_M_GETFIELD)
      || mn == ESERVER_METHOD_ARRAY (eserver, E_M_PUTFIELD)
      || mn == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKEVIRTUAL)
      || mn == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKESPECIAL)
      || mn == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKESTATIC)
      || mn == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKEINTERFACE))
    return INSTRUCTION_METHOD_NUMBER;
  else if (mn == ESERVER_METHOD_ARRAY (eserver, E_M_LOAD)
	   || mn == ESERVER_METHOD_ARRAY (eserver, E_M_INFO))
    return OPERATION_METHOD_NUMBER;

  return INVALID_METHOD_NUMBER;
}

/** LOAD method ó Լ.  */
int
e_process_load_method (er)
     erequest_t *er;
{
  erequest_info_t *ri = REQUEST_INFO (er);
  enum e_method_number_type mnt = get_method_number_type (RCONTENT_METHOD_NUMBER (ri));

  /* Step 1.  κп ش er->server_context->r->filename  load Ǿ ִ Ȯ 
     load Ǿ   , load Ѵ.  */
  if (mnt != INVALID_METHOD_NUMBER)
    {
      request_rec *r = CONTEXT_REQUEST (REQUEST_SERVER_CONTEXT (er));

      RCONTENT_ENV (ri) = efile_readnload ((char *) r->filename);
      if (!RCONTENT_ENV (ri))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "efile_readnload: return value is NULL.");
	  return -1;
	}
    }

  /* !!!DeleteMe!!! LOAD method ó ʿ  Ʒ .

     1. Load    - Post Ÿ ,   
        RCONTENT_RFC1867_UPLOADED_FILES () Hash table    ̴.  ̰ ϳ
	ҷ load Ű ȴ.  */
  /*
  if (RCONTENT_RFC1867_UPLOADED_FILES (ri))
    vm_htab_traverse (RCONTENT_RFC1867_UPLOADED_FILES (ri), e_process_file_callback, er);
  */

  return 0;
}

int
e_process_info_method (er)
     erequest_t *er;
{
  erequest_info_t *ri = REQUEST_INFO (er);

  /* 켱 ش û environment  load Ǿ ִ Ȯϰ, ȵǾ  ,
     load Ѵ.  */
  if (e_process_load_method (er))
    return -1;

  /* TODO: Ʒ ap_rvputs () ƾ eresponse_info_t ü   ִ API 
     ؾ  ̴.  */
  ap_rvputs (CONTEXT_REQUEST (REQUEST_SERVER_CONTEXT (er)),
	     env_to_etx (REQUEST_POOL (er), RCONTENT_ENV (ri)), NULL);

  return 0;
}

/**
   RET    return ڿ Ѵ.

   @param er  û
   @param ret Return  .  ׻ NULL  ƴϴ.
*/
int
process_return_type (er, ret)
     erequest_t *er;
     evalue_t *ret;
{
  etype_t *type = EVALUE_TYPE (ret);

  /* TODO:  ȯǾ return   Ѱ ϰ ִ.  ̿  
     ̷  ̴.  */

  if (!type)
    {
      ap_rvputs (CONTEXT_REQUEST (REQUEST_SERVER_CONTEXT (er)),
		 "(return (void))", NULL);
      return 0;
    }

  switch (ETYPE_KIND (type))
    {
    case ETYPE_BOOLEAN:
    case ETYPE_CHAR:
    case ETYPE_FLOAT:
    case ETYPE_DOUBLE:
    case ETYPE_BYTE:
    case ETYPE_SHORT:
    case ETYPE_INT:
    case ETYPE_LONG:
    case ETYPE_OBJECTREF:
    case ETYPE_ARRAYREF:
      {
	env_log_error (ENVLOG_MARK, ENVLOG_ERR, "process_return_type: ϼ ʼ.");
	break;
      }
    case ETYPE_VOID:
    case ETYPE_STRING:
    case ETYPE_KEY:
    default:
      /* ⿡   .  */
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "process_return_type: error.");
      break;
    }

  return -1;
}
/**
   ER  ϴ   Ư eclass  method  ȣѴ.

   @param er	EREQUEST ü
   @return	  Ǿٸ 0 , ׷ ʴٸ -1  ȯѴ.
*/
int
e_process_invokevirtual_instruction (er)
    erequest_t *er;
{
  env_t e;
  int i, namelen, typelen;
  char *namestr, *typestr;
  eargs_t *key, *name, *type;
  evalue_t ret = EVALUE_NULL, *argv = NULL;
  emethod_t *m;

  key  = lookup_erequest_args (er, "__KEY");
  name = lookup_erequest_args (er, "__NAME");
  type = lookup_erequest_args (er, "__TYPE");

  if (!key || !name || !type)
    return -1;

  namestr = EVALUE_STRING (EARGS_VALUE (name));
  typestr = EVALUE_STRING (EARGS_VALUE (type));
  namelen = strlen (namestr);
  typelen = strlen (typestr);

  /* KEY  Ȯ ʴٸ envstruct ü ã  ̴.  

     TODO:  lookup_envstruct () Լ ϰ NULL  , 丮  Ͽ
     ãƾ ұ?  */
  e = lookup_envstruct (EVALUE_ADDRESS (EARGS_VALUE (key)));
  if (!e)
    return -1;

  m = find_method (e, new_strconst (e, namestr, namelen), new_strconst (e, typestr, typelen));
  if (!m)
    return -1;

  /* static method  INVOKEVIRTUAL instruction  ȣϴ  ߸ ̴.  */
  if (METHOD_ACCESS_FLAGS (m) & AF_STATIC)
    return -1;

  /* ARGV .  */
  if (RCONTENT_CALL_ARGC (REQUEST_INFO (er)) > 0)
    {
      argv = (evalue_t *) apr_pcalloc (REQUEST_POOL (er), 
				       RCONTENT_CALL_ARGC (REQUEST_INFO (er)) * sizeof (struct evalue));
      for (i=0; i<RCONTENT_CALL_ARGC (REQUEST_INFO (er)); i++)
	argv[i] = *(RCONTENT_CALL_ARGV_IDX (REQUEST_INFO (er), i));
    }

  /* ARGV[0]  objetref Ȥ arrayref ΰ?  */
  if (argv && EVALUE_TYPE (&(argv[0])))
    {
      if (ETYPE_KIND (EVALUE_TYPE (&(argv[0]))) != ETYPE_OBJECTREF
	  && ETYPE_KIND (EVALUE_TYPE (&(argv[0]))) != ETYPE_ARRAYREF)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
			 "INVOKEVIRTUAL: ù° ڰ objectref Ȥ arrayref  ƴմϴ.");
	  return -1;
	}
    }

  if (call_method_a (m, m, NULL, &ret, argv))
    return -1;

  /* TODO: ret  return type  method signature  return type  ٸ
     츦  óؾ Ѵ.  */
  process_return_type (er, &ret);

  return 0;
}

int
e_process_invokespecial_instruction (er)
    erequest_t *er ATTRIBUTE_UNUSED;
{
  return -1;
}

int
e_process_invokestatic_instruction (er)
    erequest_t *er;
{
  env_t e;
  int i, namelen, typelen;
  char *namestr, *typestr;
  eargs_t *key, *name, *type;
  evalue_t ret = EVALUE_NULL, *argv = NULL;
  emethod_t *m;

  key  = lookup_erequest_args (er, "__KEY");
  name = lookup_erequest_args (er, "__NAME");
  type = lookup_erequest_args (er, "__TYPE");

  if (!key || !name || !type)
    return -1;

  namestr = EVALUE_STRING (EARGS_VALUE (name));
  typestr = EVALUE_STRING (EARGS_VALUE (type));
  namelen = strlen (namestr);
  typelen = strlen (typestr);

  /* KEY  Ȯ ʴٸ envstruct ü ã  ̴.  

     TODO:  get_envstruct () Լ ϰ NULL  , 丮  Ͽ
     ãƾ ұ?  */
  e = lookup_envstruct (EVALUE_ADDRESS (EARGS_VALUE (key)));
  if (!e)
    return -1;

  m = find_method (e, new_strconst (e, namestr, namelen), new_strconst (e, typestr, typelen));
  if (!m)
    return -1;

  /* static method  INVOKEVIRTUAL instruction  ȣϴ  ߸ ̴.  */
  if (!(METHOD_ACCESS_FLAGS (m) & AF_STATIC))
    return -1;

  /* ARGV .  */
  if (RCONTENT_CALL_ARGC (REQUEST_INFO (er)) > 0)
    {
      argv = (evalue_t *) apr_pcalloc (REQUEST_POOL (er), 
				       RCONTENT_CALL_ARGC (REQUEST_INFO (er)) * sizeof (struct evalue));
      for (i=0; i<RCONTENT_CALL_ARGC (REQUEST_INFO (er)); i++)
	argv[i] = *(RCONTENT_CALL_ARGV_IDX (REQUEST_INFO (er), i));
    }

  if (call_method_a (m, m, NULL, &ret, argv))
    return -1;

  /* TODO: ret  return type  method signature  return type  ٸ
     츦  óؾ Ѵ.  */
  process_return_type (er, &ret);

  return 0;
}

int
e_process_invokeinterface_instruction (er)
    erequest_t *er ATTRIBUTE_UNUSED;
{
  return -1;
}

/**
   Instruction method ó.  
*/
int
e_execute_instruction_loader (er)
     erequest_t *er;
{
  if (RCONTENT_METHOD_NUMBER (REQUEST_INFO (er)) == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKEVIRTUAL))
    e_process_invokevirtual_instruction (er);
  else if (RCONTENT_METHOD_NUMBER (REQUEST_INFO (er)) == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKESPECIAL))
    e_process_invokespecial_instruction (er);
  else if (RCONTENT_METHOD_NUMBER (REQUEST_INFO (er)) == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKESTATIC))
    e_process_invokestatic_instruction (er);
  else if (RCONTENT_METHOD_NUMBER (REQUEST_INFO (er)) == ESERVER_METHOD_ARRAY (eserver, E_M_INVOKEINTERFACE))
    e_process_invokeinterface_instruction (er);
  
  return 0;
}

/**
   Operation method ó.  
*/
int
e_execute_operation_loader (er)
     erequest_t *er;
{
  if (RCONTENT_METHOD_NUMBER (REQUEST_INFO (er)) == ESERVER_METHOD_ARRAY (eserver, E_M_LOAD))
    e_process_load_method (er);
  else if (RCONTENT_METHOD_NUMBER (REQUEST_INFO (er)) == ESERVER_METHOD_ARRAY (eserver, E_M_INFO))
    e_process_info_method (er);

  return 0;
}

/**
   POST data  ؼ ޵   ؼ, ش query  ؼϿ 
   ùٸ ó Ѵ.  */
int 
e_execute_loader (er)
     erequest_t *er;
{
  erequest_info_t *ri = REQUEST_INFO (er);
  enum e_method_number_type mnt = get_method_number_type (RCONTENT_METHOD_NUMBER (ri));

  /* Step 2.  ش environment   û Ʒ з  óѴ.  */
  switch (mnt)
    {
    case INSTRUCTION_METHOD_NUMBER:
      e_execute_instruction_loader (er);
      break;
    case OPERATION_METHOD_NUMBER:
      e_execute_operation_loader (er);
      break;
    default:
      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, " ʴ method Դϴ.");
      break;
    }

  return 0;
}
