/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include <setjmp.h>

#if !defined (P_tmpdir)
#define P_tmpdir ""
#endif

enum treat_data_type
  {
    PARSE_GET = 0,
    PARSE_POST,
    PARSE_COOKIE
  };

extern etree lex_number (apr_pool_t *, const char *, unsigned int);

static int eargs_hash_eq (const void *, const void *);
static void eargs_hash_del (void *);
static int read_post (eserver_context_t *, char *, uint);
static void env_read_standard_form_data (erequest_t *);
static void env_read_unknown_form_data (erequest_t *);
static void run_default_post_reader (erequest_t *);
static void run_post_reader (erequest_t *);

evalue_t * new_evalue_core (apr_pool_t *, enum evalue_type, va_list);
int is_reserved_argname (char *);
envkey_t * parse_base64_key (erequest_t *, char *, int);
evalue_t * handle_reserved_argname (erequest_t *, char *, enum evalue_type, va_list);
eargs_t * new_erequest_args (erequest_t *, char *, enum evalue_type, ...);
void env_standard_post_handler (erequest_t *, const char *, void *);
const char* multipart_get_tmp_dir (void);
int multipart_open_tmp_fd (const char *, const char *, char **);
FILE * multipart_open_tmp_file (const char *, const char *, char **);
int multipart_hash_eq (const void *, const void *);
void multipart_hash_del (void *);
void multipart_new_uploaded_files (vm_htab_t, char *);
void rfc1867_post_handler (erequest_t *, const char *, void *);
void handle_post (erequest_t *, void *);
void handle_get (erequest_t *, void *);
void default_treat_data (erequest_t *, enum treat_data_type);
void process_post_data (erequest_t *);
void process_interpretref (erequest_t *);
int convert_etree_to_evalue (erequest_t *, char, etree, evalue_t *);

/* EVALUE  type  TYPE  EVALUE ü  ȯѴ.  */
evalue_t *
new_evalue_core (p, type, ap)
     apr_pool_t *p;
     enum evalue_type type;
     va_list ap;
{
  char *estr;
  int eboolean, ebyte, echar, eshort, eint, len;
  long long elong;
  double efloat, edouble;
  void *eobj, *earray;
  evalue_t *val;

  val = (evalue_t *) apr_pcalloc (p, sizeof (struct evalue));
  EVALUE_TYPE (val) = get_etype_with_kind (type);
  EVALUE_IS_REMOTE (val) = 0;
  EVALUE_REFCOUNT (val) = 0;

  switch (type)
    {
    case ETYPE_BYTE:
      ebyte = va_arg (ap, int);
      EVALUE_BYTE (val) = (ebyte_t) (ebyte & 0xff);
      break;
    case ETYPE_SHORT:
      eshort = va_arg (ap, int);
      EVALUE_SHORT (val) = (eshort_t) (eshort & 0xffff);
      break;
    case ETYPE_INT:
      eint = va_arg (ap, int);
      EVALUE_INT (val) = (eint_t) eint;
      break;
    case ETYPE_LONG:
      elong = va_arg (ap, elong_t);
      EVALUE_LONG (val) = (elong_t) elong;
      break;
    case ETYPE_FLOAT:
      efloat = va_arg (ap, edouble_t);
      EVALUE_FLOAT (val) = (efloat_t) efloat;
      break;
    case ETYPE_DOUBLE:
      edouble = va_arg (ap, edouble_t);
      EVALUE_DOUBLE (val) = (edouble_t) edouble;
      break;
    case ETYPE_CHAR:
      echar = va_arg (ap, int);
      EVALUE_CHAR (val) = (echar_t) (echar & 0xffff);
      break;
    case ETYPE_BOOLEAN:
      eboolean = va_arg (ap, int);
      EVALUE_BOOLEAN (val) = (eboolean_t) (eboolean & 0xff);
      break;
    case ETYPE_OBJECTREF:
      eobj = va_arg (ap, eref_t);
      EVALUE_OBJECT (val) = (eref_t) eobj;
      break;
    case ETYPE_ARRAYREF:
      earray = va_arg (ap, eref_t);
      EVALUE_ARRAY (val) = (eref_t) earray;
      break;
    case ETYPE_STRING:
      estr = va_arg (ap, char *);
      len = va_arg (ap, int);
      if (estr && len >= 0)
	EVALUE_STRING (val) = apr_pstrndup (p, (char *) estr, len);
      break;
    default:
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, " ʴ EVALUE type.");
      break;
    }

  return val;
}

/* TYPE   EVALUE  Ͽ ȯѴ.  */
evalue_t *
new_evalue (apr_pool_t *p, enum evalue_type type, ...)
{
  evalue_t *val;
  va_list ap;

  va_start (ap, type);
  val = new_evalue_core (p, type, ap);
  va_end (ap);

  return val;
}

/* Hash Table  Ǵ ϼ ׽Ʈ Լν, ENTRY  
   Hash Table  ִ SLOT  ̰, X_ARG  vm_htab_find_slot_with_hash 
   ڷ  ̴.  */
static int
eargs_hash_eq (entry, x_arg)
     const void *entry, *x_arg;
{
  eargs_t *sa;
  char *sb;
  
  sa = (eargs_t *) entry;
  sb = (char *) x_arg;

  if (EARGS_VARIABLE (sa) && sb 
      && !strcmp (EARGS_VARIABLE (sa), sb))
    return 1;

  return 0;
}

/* Hash Table   Slot  ϵǾ ִ entry  delete   Ǵ Լ̴.
   vm_htab_create () Լ ȣ , Լ Ͱ  ϵȴ.  */
static void
eargs_hash_del (entry)
     void *entry ATTRIBUTE_UNUSED;
{
  //eargs_t *s = (eargs_t *) entry;
}

/*  ARGNAME   argument ̸ , TRUE  ׷ ʴٸ FALSE  ȯѴ.  */
int
is_reserved_argname (argname)
     char *argname;
{
  if (!argname
      || (argname && strlen (argname) < 3)
      || (argname && !(argname[0] == '_' && argname[1] == '_')))
    return 0;

  /*  ̸   Ѵ.  */
  if (!strncmp (argname, "__KEY", 5)
      || !strncmp (argname, "__NAME", 6)
      || !strncmp (argname, "__TYPE", 6))
    return 1;

  /* __<number><type>    Ѵ.  ⿡ argname[2] 
      츸 ó Ѵ.  */
  if (!ISDIGIT (argname[2]) || strlen (argname) < 4)
    return 0;

  return 1;
}

/* BASE64  encoding Ǿ ִ KEY ڿ ޾, ENVKEY  ȯϿ ȯѴ.
   Լ  ȯǴ ENVKEY  ޸ Ҵ  ݵ  ־ ϰڴ.  */
envkey_t *
parse_base64_key (er, str, len)
     erequest_t *er;
     char *str;
     int len;
{
  int keysize;
  char *tmp;
  ENVKEY_CHAR s[ENVKEY_SIZE];
  envkey_t key, *sk;
  
  tmp = (char *) apr_pstrndup (REQUEST_POOL (er), str, len);
  keysize = apr_base64_decode_binary (s, tmp);
  if (keysize != ENVKEY_SIZE)
    return NULL;
  
  /*  κп base64  encoding   ENVKEY  ȯϿ Ѵ.  */
  PUT_ENVKEY (s, &key);
  PDUPLICATE_ENVKEY (REQUEST_POOL (er), &key, sk);

  return sk;
}

/**
   ETREE  EVALUE  Ѵ.

   @param er	Request 
   @param type	 Ÿ
   @param val	 ϼ  EVALUE
   @param t	  ִ ETREE
*/
int
convert_etree_to_evalue (er, type, t, val)
     erequest_t *er ATTRIBUTE_UNUSED;
     char type ATTRIBUTE_UNUSED;
     etree t ATTRIBUTE_UNUSED;
     evalue_t *val ATTRIBUTE_UNUSED;
{
  return 0;
}

/* ARGNAME  __KEY, __NAME, __TYPE, __<number><type> ̶  ϰ ó Ѵ.
   ̷    ִ  is_reserved_argname () Լ θ Ǵ ֱ
   ̴.   TYPE  ݵ ETYPE_STRING ̿ Ѵ.

   TODO : Ʒ Ǵ ޸ Ҵ  request pool  ȯؾ Ѵ.  */
evalue_t *
handle_reserved_argname (er, argname, type, ap)
     erequest_t *er;
     char *argname;
     enum evalue_type type;
     va_list ap;
{
  char *estr, *p;
  int num = -1, len, is_pref = 0;
  evalue_t *val;

  if (type != ETYPE_STRING)
    return NULL;

  /* Ʒ ENV_MALLOC  Request Pool  ȯؾ Ѵ.  */
  val = (evalue_t *) apr_pcalloc (REQUEST_POOL (er), sizeof (struct evalue));
  EVALUE_IS_REMOTE (val) = 0;
  EVALUE_REFCOUNT (val) = 0;

  estr = va_arg (ap, char *);
  len = va_arg (ap, int);

  if (!strncmp (argname, "__KEY", 5))
    {
      EVALUE_TYPE (val) = get_etype_with_kind (ETYPE_KEY);
      EVALUE_ADDRESS (val) = (void *) parse_base64_key (er, estr, len);

      return val;
    }
  else if (!strncmp (argname, "__NAME", 6)
	   || !strncmp (argname, "__TYPE", 6))
    {
      EVALUE_TYPE (val) = get_etype_with_kind (ETYPE_STRING);
      if (estr && len >= 0)
	EVALUE_STRING (val) = apr_pstrndup (REQUEST_POOL (er), (char *) estr, len);
      else
	EVALUE_STRING (val) = NULL;

      /* __TYPE  , ARGC  ARGV    ֵ ش  Ѵ.  */
      if (!strncmp (argname, "__TYPE", 6) && EVALUE_STRING (val))
	{
	  int count = count_args_in_type (EVALUE_STRING (val));

	  RCONTENT_CALL_ARGC (REQUEST_INFO (er)) = count;
	  RCONTENT_CALL_ARGV (REQUEST_INFO (er)) = apr_pcalloc (REQUEST_POOL (er), count * sizeof (struct evalue));
	}

      return val;
    }

  /*  __<number><type>   κ Ʒ óϿ Ѵ.  */
  p = argname + 2;	/* __  skip.  */
  num = atoi (p);
  while (ISDIGIT (*p))	/* <number>  skip.  */
    p++;

  /*  [T<number>    , Ǯ ArrayRef  ̴.  ̰  ؼ 
     ټ ־ Ѵ.  */
  if (*p == '[' && *(p+1) == 'T' && ISDIGIT (*(p+2)))
    {
      /* P+3   ڿ ϸ,  ũ 256  Ѱܼ ȵȴ.  */
      char *t = p + 3;
      while (t != '\0')
	{
	  if (!ISDIGIT (*t))
	    return NULL;
	  t++;
	}
      is_pref = 1;
      EVALUE_TYPE (val) = get_etype (apr_pstrndup (REQUEST_POOL (er), p, 2));
    }
  else
    EVALUE_TYPE (val) = lookup_etype (p);

  if (!EVALUE_TYPE (val))
    return NULL;

  /* TODO:   쿡  ó  ̷ ʰ ִ.  */
  switch (*p)
    {
    case 'T':	/* ڿ   type */
      {
	EVALUE_STRING (val) = apr_pstrndup (REQUEST_POOL (er), estr, len);
	break;
      }
    case '[':	/* Arrayref */
      {
	if (is_pref)
	  {
	    int idx = atoi (p+2);
	    struct interpretref *node;
	    
	    EVALUE_STRING (val) = apr_pstrndup (REQUEST_POOL (er), estr, len);

	    node = apr_pcalloc (REQUEST_POOL (er), sizeof (struct interpretref));
	    INTERPRETREF_ARG_INDEX (node) = num;
	    INTERPRETREF_INDEX (node) = idx;
	    INTERPRETREF_VALUE (node) = val;
	    
	    /* Node ߰.  */
	    INTERPRETREF_NEXT (node) = RCONTENT_INTERPRETREF_LIST (REQUEST_INFO (er));
	    RCONTENT_INTERPRETREF_LIST (REQUEST_INFO (er)) = INTERPRETREF_NEXT (node);

	    break;
	  }
	/* is_pref  0  ,  .  */
      }
    case 'B':	/* Byte */
    case 'S':	/* Short */
    case 'I':	/* Integer */
    case 'J':	/* Long */
    case 'F':	/* Float */
    case 'D':	/* Double */
    case 'C':	/* Char */
    case 'Z':	/* Boolean */
    case 'L':	/* Objectref */
      convert_etree_to_evalue (er, *p, parse_number (REQUEST_POOL (er), estr, len), val);
      break;

    default:
      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "handle_reserved_argname:  ʴ TYPE Դϴ.");
      break;
    }

  /* ARGV  Ѵ.   is_pref  Ǿ ִٸ RCONTENT_CALL_ARGV_IDX 
      ʴ´.  ̰ ߿ process_interpretref () Լ ȴ.  */
  if (RCONTENT_CALL_ARGC (REQUEST_INFO (er)) == -1
      || num >= RCONTENT_CALL_ARGC (REQUEST_INFO (er)))
    return NULL;
  else if (!is_pref)
    RCONTENT_CALL_ARGV_IDX (REQUEST_INFO (er), num) = val;

  return val;
}

/**
   EARGS   VAR   eargs_t ü ȯѴ.

   @param er	EREQUEST ü
   @param var	ãϴ  ̸
   @return	new_erequest_args () Լ   eargs_t ü
*/
eargs_t *
lookup_erequest_args (erequest_t *er, const char *var)
{
  eargs_t **slot;
  vm_htab_t ht;
  vm_hashval_t hash;

  if (!er || !var)
    return NULL;

  ht = RCONTENT_ARGS (REQUEST_INFO (er));
  hash = vm_htab_hash_string (var);
  slot = (eargs_t **) vm_htab_find_slot_with_hash (ht, var, hash, NO_INSERT);

  if (slot)
    return *slot;

  return NULL;
}

/*  ̰ VAR_LEN  VAR   ̰ VAL_LEN  VAL  Ͽ, HT hash
   table  Ѵ.     Ѵٸ ο  üѴ.

   뿹)
     new_erequest_args (ht, "KEY", ETYPE_STRING, "djslkjdsklew9302=");
  */
eargs_t *
new_erequest_args (erequest_t *er, char *var, enum evalue_type type, ...)
{
  va_list ap;
  eargs_t **slot, *entry;
  vm_htab_t ht;
  vm_hashval_t hash;

  if (!er || !var)
    return NULL;	/*  */

  ht = RCONTENT_ARGS (REQUEST_INFO (er));
  hash = vm_htab_hash_string (var);

  slot = (eargs_t **) vm_htab_find_slot_with_hash (ht, var, hash, INSERT);
  entry = *slot;

  if (entry)
    /* ̹  */
    return entry;

  /* ⿡ ߴٴ  argument  տ ϵ   ̴.  */
  va_start (ap, type);
  entry = (eargs_t *) apr_pcalloc (REQUEST_POOL (er), sizeof (struct environment_args));
  EARGS_VARIABLE (entry) = apr_pstrdup (REQUEST_POOL (er), var);

  /*  argument ̸  Ư  ־Ѵ.  */
  if (is_reserved_argname (EARGS_VARIABLE (entry)))
    EARGS_VALUE (entry) = handle_reserved_argname (er, EARGS_VARIABLE (entry), type, ap);
  else
    EARGS_VALUE (entry) = new_evalue_core (REQUEST_POOL (er), type, ap);
  va_end (ap);

  *slot = entry;

  return entry;	/*  */
}

/* REQUEST_REC  brigade κ POST Ÿ COUNT_BYTES ŭ о BUF 
   Ѵ.  */
static int
read_post (ctx, buf, count_bytes)
     eserver_context_t *ctx;
     char *buf;
     uint count_bytes;
{
  apr_size_t len, tlen=0;
  request_rec *r;
  apr_bucket_brigade *brigade;

  r = CONTEXT_REQUEST (ctx);
  brigade = CONTEXT_BRIGADE (ctx);
  len = count_bytes;

  /* This loop is needed because ap_get_brigade () can return us partial data
     which would cause premature termination of request read. Therefor we
     need to make sure that if data is avaliable we fill the buffer completely.  */
  while (ap_get_brigade (r->input_filters, brigade, AP_MODE_READBYTES, APR_BLOCK_READ, len) == APR_SUCCESS) 
    {
      apr_brigade_flatten (brigade, buf, &len);
      apr_brigade_cleanup (brigade);
      tlen += len;
      if (tlen == count_bytes || !len)
	break;
      
      buf += len;
      len = count_bytes - tlen;
    }
  
  return tlen;
}

static void
env_read_standard_form_data (er)
     erequest_t *er;
{
  int read_bytes;
  int allocated_bytes = REQUEST_POST_BLOCK_SIZE;
  erequest_info_t *i = REQUEST_INFO (er);

  /* ⼭ Ҵ ޸𸮴 handle_post () Լ  memory leak  
     ʾƵ ȴ.  */
  RCONTENT_POST_DATA (i) = env_malloc (allocated_bytes);
  for (;;)
    {
      read_bytes = read_post (REQUEST_SERVER_CONTEXT (er), RCONTENT_POST_DATA (i) + RCONTENT_READ_POST_BYTES (i),
			      REQUEST_POST_BLOCK_SIZE);
      if (read_bytes <= 0)
	break;
      RCONTENT_READ_POST_BYTES (i) += read_bytes;
      if (read_bytes < REQUEST_POST_BLOCK_SIZE)
	break;
      if (RCONTENT_READ_POST_BYTES (i) + REQUEST_POST_BLOCK_SIZE >= allocated_bytes)
	{
	  allocated_bytes = RCONTENT_READ_POST_BYTES (i) + REQUEST_POST_BLOCK_SIZE;
	  RCONTENT_POST_DATA (i) = env_realloc (RCONTENT_POST_DATA (i), allocated_bytes);
	}
    }

  /* terminating NULL */
  RCONTENT_POST_DATA_ARRAY (i, RCONTENT_READ_POST_BYTES (i)) = '\0';
  RCONTENT_POST_DATA_LENGTH (i) = RCONTENT_READ_POST_BYTES (i);
}

/* CT  Content-Type ̰, ARG  ⼭ ؼǴ ڵ ԵǴ Hash Table
   ̴.  */
void
env_standard_post_handler (er, ct, ht)
     erequest_t *er;
     const char *ct ATTRIBUTE_UNUSED;
     void *ht ATTRIBUTE_UNUSED;
{
  int val_len;
  char *var, *val, *strtok_buf = NULL;
  erequest_info_t *i = REQUEST_INFO (er);

  if (!RCONTENT_POST_DATA (i))
    return;

  var = strtok_r (RCONTENT_POST_DATA (i), "&", &strtok_buf);
  while (var)
    {
      val = strchr (var, '=');
      if (val)
	{
	  *val++ = '\0';
	  url_decode (var, strlen (var));
	  val_len = url_decode (val, strlen (val));
	  /* variable  value  Hash table  Ѵ.  */
	  new_erequest_args (er, var, ETYPE_STRING, val, val_len);
	}
      var = strtok_r (NULL, "&", &strtok_buf);
    }
}

/***************************************************************************

  RFC1867

 ***************************************************************************/
/* Following code is based on apache_multipart_buffer.c from libapreq-0.33 package.  */

#define FILLUNIT (1024 * 5)

#define SAFE_RETURN { \
	if (lbuf) env_free (lbuf); \
	if (abuf) env_free (abuf); \
	if (array_index) env_free (array_index); \
	zend_llist_destroy (&header); \
	if (mbuff->boundary_next) env_free (mbuff->boundary_next); \
	if (mbuff->boundary) env_free (mbuff->boundary); \
	if (mbuff->buffer) env_free (mbuff->buffer); \
	if (mbuff) env_free (mbuff); \
	return; }

/* The longest property name we use in an uploaded file array */
#define MAX_SIZE_OF_INDEX sizeof ("[tmp_name]")

/* Errors */
#define UPLOAD_ERROR_OK   0  /* File upload succesful */
#define UPLOAD_ERROR_A    1  /* Uploaded file exceeded upload_max_filesize */
#define UPLOAD_ERROR_B    2  /* Uploaded file exceeded MAX_FILE_SIZE */
#define UPLOAD_ERROR_C    3  /* Partially uploaded */
#define UPLOAD_ERROR_D    4  /* No file uploaded */
#define UPLOAD_ERROR_E    6  /* Missing /tmp or similar directory */

typedef struct 
{
  /* read buffer */
  char *buffer;
  char *buf_begin;
  int bufsize;
  int bytes_in_buffer;
  
  /* boundary info */
  char *boundary;
  char *boundary_next;
  int boundary_next_len;

} multipart_buffer;

typedef struct 
{
  char *key;
  char *value;
} mime_header_entry;

/* create new multipart_buffer structure */
static multipart_buffer *
multipart_buffer_new (char *boundary, int boundary_len)
{
  int minsize = boundary_len + 6;
  multipart_buffer *self = (multipart_buffer *) env_calloc (1, sizeof (multipart_buffer));

  if (minsize < FILLUNIT) 
    minsize = FILLUNIT;
  
  self->buffer = (char *) env_calloc (1, minsize + 1);
  self->bufsize = minsize;
  
  self->boundary = (char *) env_calloc (1, boundary_len + 3); 
  sprintf (self->boundary, "--%s", boundary);
  
  self->boundary_next = (char *) env_calloc (1, boundary_len + 4);
  sprintf (self->boundary_next, "\n--%s", boundary);
  self->boundary_next_len = boundary_len + 3;
  
  self->buf_begin = self->buffer;
  self->bytes_in_buffer = 0;
  
  return self;
}

/*
  fill up the buffer with client data.
  returns number of bytes added to buffer.
*/
static int 
multipart_fill_buffer (erequest_t *er, multipart_buffer *self)
{
  int bytes_to_read, total_read = 0, actual_read = 0;
  
  /* shift the existing data if necessary */
  if (self->bytes_in_buffer > 0 && self->buf_begin != self->buffer) 
    memmove (self->buffer, self->buf_begin, self->bytes_in_buffer);
  
  self->buf_begin = self->buffer;
  
  /* calculate the free space in the buffer */
  bytes_to_read = self->bufsize - self->bytes_in_buffer;
  
  /* read the required number of bytes */
  while (bytes_to_read > 0) 
    {
      char *buf = self->buffer + self->bytes_in_buffer;
      
      actual_read = read_post (REQUEST_SERVER_CONTEXT (er), buf, bytes_to_read);

      /* update the buffer length */
      if (actual_read > 0) 
	{
	  erequest_info_t *i = REQUEST_INFO (er);

	  self->bytes_in_buffer += actual_read;
	  RCONTENT_READ_POST_BYTES (i) += actual_read;
	  total_read += actual_read;
	  bytes_to_read -= actual_read;
	} 
      else
	break;
    }
  
  return total_read;
}

/* eof if we are out of bytes, or if we hit the final boundary */
static int 
multipart_buffer_eof (erequest_t *er, multipart_buffer *self)
{
  if ((self->bytes_in_buffer == 0 && multipart_fill_buffer (er, self) < 1))
    return 1;
  else
    return 0;
}

/* Free header entry */
static void 
env_free_hdr_entry (mime_header_entry *h)
{
  if (h->key)
    env_free (h->key);

  if (h->value)
    env_free (h->value);
}

/*
  gets the next CRLF terminated line from the input buffer.
  if it doesn't find a CRLF, and the buffer isn't completely full, returns
  NULL; otherwise, returns the beginning of the null-terminated line,
  minus the CRLF.

  note that we really just look for LF terminated lines. this works
  around a bug in internet explorer for the macintosh which sends mime
  boundaries that are only LF terminated when you use an image submit
  button in a multipart/form-data form.
 */
static char *
multipart_next_line (multipart_buffer *self)
{
  /* look for LF in the data */
  char* line = self->buf_begin;
  char* ptr = memchr (self->buf_begin, '\n', self->bytes_in_buffer);
  
  if (ptr) 
    {	/* LF found */
      /* terminate the string, remove CRLF */
      if ((ptr - line) > 0 && *(ptr-1) == '\r') 
	{
	  *(ptr-1) = 0;
	} 
      else 
	{
	  *ptr = 0;
	}
      
      /* bump the pointer */
      self->buf_begin = ptr + 1;
      self->bytes_in_buffer -= (self->buf_begin - line);
    } 
  else 
    {	/* no LF found */
      /* buffer isn't completely full, fail */
      if (self->bytes_in_buffer < self->bufsize) 
	{
	  return NULL;
	}
      /* return entire buffer as a partial line */
      line[self->bufsize] = 0;
      self->buf_begin = ptr;
      self->bytes_in_buffer = 0;
    }
  
  return line;
}

/* returns the next CRLF terminated line from the client */
static char *
multipart_get_line (erequest_t *er, multipart_buffer *self)
{
  char* ptr = multipart_next_line (self);
  
  if (!ptr) {
    multipart_fill_buffer (er, self);
    ptr = multipart_next_line (self);
  }
  
  return ptr;
}

/* finds a boundary */
static int 
multipart_find_boundary (erequest_t *er, multipart_buffer *self, char *boundary)
{
  char *line;
  
  /* loop thru lines */
  while ( (line = multipart_get_line (er, self)) )
    {
      /* finished if we found the boundary */
      if (!strcmp (line, boundary)) 
	{
	  return 1;
	}
    }
  
  /* didn't find the boundary */
  return 0;
}

static char *
multipart_mime_get_hdr_value (zend_llist header, char *key)
{
  mime_header_entry *entry;
  
  if (key == NULL) 
    {
      return NULL;
    }
  
  entry = zend_llist_get_first (&header);
  while (entry) 
    {
      if (!strcasecmp (entry->key, key)) 
	{
	  return entry->value;
	}
      entry = zend_llist_get_next (&header);
    }
  
  return NULL;
}

static char *
multipart_ap_getword (char **line, char stop)
{
  char *pos = *line, quote;
  char *res;
  
  while (*pos && *pos != stop) 
    {
      if ((quote = *pos) == '"' || quote == '\'') 
	{
	  ++pos;
	  while (*pos && *pos != quote) 
	    {
	      if (*pos == '\\' && pos[1] && pos[1] == quote) 
		{
		  pos += 2;
		} 
	      else 
		{
		  ++pos;
		}
	    }
	  if (*pos) 
	    {
	      ++pos;
	    }
	}
      else
	++pos;
      
    }
  if (*pos == '\0') 
    {
      res = env_strdup (*line);
      *line += strlen (*line);
      return res;
    }
  
  res = env_strndup (*line, pos - *line);
  
  while (*pos == stop) 
    {
      ++pos;
    }
  
  *line = pos;
  return res;
}

/* parse headers */
static int 
multipart_buffer_headers (erequest_t *er, multipart_buffer *self, zend_llist *header)
{
  char *line;
  mime_header_entry prev_entry, entry;
  int prev_len, cur_len;
  
  /* didn't find boundary, abort */
  if (!multipart_find_boundary (er, self, self->boundary)) 
    return 0;
  
  /* get lines of text, or CRLF_CRLF */
  
  while ( (line = multipart_get_line (er, self)) && strlen (line) > 0 )
    {
      /* add header to table */
      char *key = line;
      char *value = NULL;
      
      /* space in the beginning means same header */
      if (!ISSPACE (line[0])) 
	{
	  value = strchr (line, ':');
	}
      
      if (value) 
	{
	  *value = 0;
	  do 
	    { 
	      value++; 
	    } while (ISSPACE (*value));
	  
	  entry.value = env_strdup (value);
	  entry.key = env_strdup (key);
	  
	}
      else if (zend_llist_count (header)) 
	{ /* If no ':' on the line, add to previous line */
	  
	  prev_len = strlen (prev_entry.value);
	  cur_len = strlen (line);
	  
	  entry.value = env_malloc (prev_len + cur_len + 1);
	  memcpy (entry.value, prev_entry.value, prev_len);
	  memcpy (entry.value + prev_len, line, cur_len);
	  entry.value[cur_len + prev_len] = '\0';
	  
	  entry.key = env_strdup (prev_entry.key);
	  
	  zend_llist_remove_tail (header);
	} 
      else 
	{
	  continue;
	}
      
      zend_llist_add_element (header, &entry);
      prev_entry = entry;
    }
  
  return 1;
}

static char *
multipart_substring_conf (char *start, int len, char quote)
{
  char *result = env_malloc (len + 2);
  char *resp = result;
  int i;
  
  for (i = 0; i < len; ++i) 
    {
      if (start[i] == '\\' && (start[i + 1] == '\\' || (quote && start[i + 1] == quote))) 
	{
	  *resp++ = start[++i];
	}
      else 
	{
	  *resp++ = start[i];
	}
    }
  
  *resp++ = '\0';

  return result;
}

static char *
multipart_ap_getword_conf (char **line)
{
  char *str = *line, *strend, *res, quote;
  
  while (*str && ISSPACE (*str)) 
    {
      ++str;
    }
  
  if (!*str) 
    {
      *line = str;
      return env_strdup ("");
    }

  if ((quote = *str) == '"' || quote == '\'') 
    {
      strend = str + 1;
    look_for_quote:
      while (*strend && *strend != quote) 
	{
	  if (*strend == '\\' && strend[1] && strend[1] == quote) 
	    {
	      strend += 2;
	    }
	  else 
	    {
	      ++strend;
	    }
	}
      if (*strend && *strend == quote) 
	{
	  char p = *(strend + 1);
	  if (p != '\r' && p != '\n' && p != '\0') 
	    {
	      strend++;
	      goto look_for_quote;
	    }
	}
      
      res = multipart_substring_conf (str + 1, strend - str - 1, quote);

      if (*strend == quote) 
	{
	  ++strend;
	}

    }
  else 
    {
      strend = str;
      while (*strend && !ISSPACE (*strend)) 
	{
	  ++strend;
	}
      res = multipart_substring_conf (str, strend - str, 0);
    }
  
  while (*strend && ISSPACE (*strend)) 
    {
      ++strend;
    }
  
  *line = strend;

  return res;
}

/*
  search for a string in a fixed-length byte string.
  if partial is true, partial matches are allowed at the end of the buffer.
  returns NULL if not found, or a pointer to the start of the first match.
*/
static void *
multipart_ap_memstr (char *haystack, int haystacklen, char *needle, int needlen, int partial)
{
  int len = haystacklen;
  char *ptr = haystack;
  
  /* iterate through first character matches */
  while ((ptr = memchr (ptr, needle[0], len))) 
    {
      /* calculate length after match */
      len = haystacklen - (ptr - (char *) haystack);
      
      /* done if matches up to capacity of buffer */
      if (memcmp (needle, ptr, needlen < len ? needlen : len) == 0 && (partial || len >= needlen)) 
	break;

      /* next character */
      ptr++; len--;
    }
  
  return ptr;
}

/* read until a boundary condition */
static int 
multipart_buffer_read (erequest_t *er, multipart_buffer *self, char *buf, int bytes)
{
  int len, max;
  char *bound;
  
  /* fill buffer if needed */
  if (bytes > self->bytes_in_buffer) 
    {
      multipart_fill_buffer (er, self);
    }

  /* look for a potential boundary match, only read data up to that point */
  if ((bound = multipart_ap_memstr (self->buf_begin, self->bytes_in_buffer, self->boundary_next, self->boundary_next_len, 1))) 
    max = bound - self->buf_begin;
  else 
    max = self->bytes_in_buffer;
  
  /* maximum number of bytes we are reading */
  len = max < bytes-1 ? max : bytes-1;
  
  /* if we read any data... */
  if (len > 0) 
    {
      /* copy the data */
      memcpy (buf, self->buf_begin, len);
      buf[len] = 0;
      
      if (bound && len > 0 && buf[len-1] == '\r') 
	buf[--len] = 0;
      
      /* update the buffer */
      self->bytes_in_buffer -= len;
      self->buf_begin += len;
    }
  
  return len;
}

/*
  XXX: this is horrible memory-usage-wise, but we only expect
  to do this on small pieces of form data.
*/
static char *
multipart_buffer_read_body (erequest_t *er, multipart_buffer *self)
{
  char buf[FILLUNIT], *out = NULL;
  int total_bytes = 0, read_bytes = 0;
  
  while ((read_bytes = multipart_buffer_read (er, self, buf, sizeof (buf)))) 
    {
      out = env_realloc (out, total_bytes + read_bytes + 1);
      memcpy (out + total_bytes, buf, read_bytes);
      total_bytes += read_bytes;
    }
  
  if (out)
    out[total_bytes] = '\0';
  
  return out;
}

/*
 *  Determine where to place temporary files.
 */
const char* 
multipart_get_tmp_dir (void)
{
  /* Cache the chosen temporary directory. */
  static char* temporary_directory;
  
  /* Did we determine the temporary directory already? */
  if (temporary_directory) 
    return temporary_directory;

#ifdef ENV_WIN32
  /* We can't count on the environment variables TEMP or TMP,
   * and so must make the Win32 API call to get the default
   * directory for temporary files.  Note this call checks
   * the environment values TMP and TEMP (in order) first.
   */
  {
    char sTemp[MAX_PATH];
    DWORD n = GetTempPath (sizeof (sTemp), sTemp);
    assert (0 < n);  /* should *never* fail! */
    temporary_directory = strdup (sTemp);
    return temporary_directory;
  }
#else
  /* On Unix use the (usual) TMPDIR environment variable. */
  {
    char* s = getenv ("TMPDIR");
    if (s) {
      temporary_directory = strdup (s);
      return temporary_directory;
    }
  }
#ifdef P_tmpdir
  /* Use the standard default temporary directory. */
  if (P_tmpdir) {
    temporary_directory = (char *) P_tmpdir;
    return temporary_directory;
  }
#endif
  /* Shouldn't ever (!) end up here ... last ditch default. */
  temporary_directory = (char *) "/tmp";
  return temporary_directory;
#endif
}

/* Loosely based on a tempnam () implementation by UCLA */

/*
 * Copyright (c) 1988, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

static int 
multipart_do_open_tmp_file (const char *path, const char *pfx, char **opened_path_p)
{
  char *trailing_slash;
  char *opened_path;
  int fd = -1;
#ifndef HAVE_MKSTEMP
  int open_flags = O_CREAT | O_TRUNC | O_RDWR
#ifdef ENV_WIN32
                   | _O_BINARY
#endif
    ;
#endif
#ifdef NETWARE
  char *file_path = NULL;
#endif
  
  if (!path)
    return -1;
  
  if (!(opened_path = env_malloc (MAXPATHLEN)))
    return -1;

  if (IS_SLASH (path[strlen (path)-1]))
    trailing_slash = (char *) "";
  else 
    trailing_slash = (char *) "/";

  (void)snprintf (opened_path, MAXPATHLEN, "%s%s%sXXXXXX", path, trailing_slash, pfx);
  
#ifdef ENV_WIN32
  if (GetTempFileName (path, pfx, 0, opened_path)) 
    {
      /* Some versions of windows set the temp file to be read-only,
       * which means that opening it will fail... */
      VCWD_CHMOD (opened_path, 0600);
      fd = VCWD_OPEN_MODE (opened_path, open_flags, 0600);
    }
#elif defined (NETWARE)
  /* Using standard mktemp () implementation for NetWare */
  file_path = mktemp (opened_path);
  if (file_path) 
    {
      fd = VCWD_OPEN (file_path, open_flags);
    }
#elif defined (HAVE_MKSTEMP)
  fd = mkstemp (opened_path);
#else
  if (mktemp (opened_path)) 
    {
      fd = VCWD_OPEN (opened_path, open_flags);
    }
#endif
  if (fd == -1 || !opened_path_p)
    env_free (opened_path);
  else 
    *opened_path_p = opened_path;

  return fd;
}


/* Unlike tempnam (), the supplied dir argument takes precedence
   over the TMPDIR environment variable
   This function should do its best to return a file pointer to a newly created
   unique file, on every platform.  */
int 
multipart_open_tmp_fd (dir, pfx, opened_path_p)
     const char *dir, *pfx;
     char **opened_path_p;
{
  int fd;
  
  if (!pfx)
    pfx = "tmp.";

  if (opened_path_p)
    *opened_path_p = NULL;

  /* Try the directory given as parameter. */
  fd = multipart_do_open_tmp_file (dir, pfx, opened_path_p);
  if (fd == -1) 
    {
      /* Use default temporary directory. */
      fd = multipart_do_open_tmp_file (multipart_get_tmp_dir (), pfx, opened_path_p);
    }

  return fd;
}

FILE *
multipart_open_tmp_file (dir, pfx, opened_path_p)
     const char *dir, *pfx;
     char **opened_path_p;
{
  FILE *fp;
  int fd = multipart_open_tmp_fd (dir, pfx, opened_path_p);
  
  if (fd == -1)
    return NULL;
  
  fp = fdopen (fd, "r+b");
  if (fp == NULL)
    close (fd);
  
  return fp;
}

/* Hash Table  Ǵ ϼ ׽Ʈ Լν, ENTRY  
   Hash Table  ִ SLOT  ̰, X_ARG  vm_htab_find_slot_with_hash 
   ڷ  ̴.  */
int
multipart_hash_eq (entry, x_arg)
     const void *entry, *x_arg;
{
  rfc1867_uploaded_files_t *sa;
  char *sb;
  
  sa = (rfc1867_uploaded_files_t *) entry;
  sb = (char *) x_arg;

  if (sa->buf && sb && !strcmp (sa->buf, sb))
    return 1;

  return 0;
}

/* Hash Table   Slot  ϵǾ ִ entry  delete   Ǵ Լ̴.
   vm_htab_create () Լ ȣ , Լ Ͱ  ϵȴ.  */
void
multipart_hash_del (entry)
     void *entry;
{
  rfc1867_uploaded_files_t *s = (rfc1867_uploaded_files_t *) entry;

  env_free (s->buf);
  env_free (s);
}

/* HT  ̸ FILENAME  Ѵ. */
void
multipart_new_uploaded_files (ht, filename)
     vm_htab_t ht;
     char *filename;
{
  vm_hashval_t hash = vm_htab_hash_string (filename);
  rfc1867_uploaded_files_t **slot, *entry;
  
  slot = (rfc1867_uploaded_files_t **) vm_htab_find_slot_with_hash (ht, filename, hash, INSERT);
  entry = *slot;

  if (entry)
    return;

  entry = (rfc1867_uploaded_files_t *) env_malloc (sizeof (struct rfc1867_uploaded_files_constant));
  entry->hash = hash;
  entry->buf = env_strdup (filename);
  entry->len = strlen (filename);

  *slot = entry;
  
  return;
}

void
rfc1867_post_handler (er, ct, ht)
     erequest_t *er;
     const char *ct;
     void *ht ATTRIBUTE_UNUSED;
{
  char *boundary, *boundary_end = NULL;
  char *lbuf = NULL, *abuf = NULL, *s = NULL;
  char *array_index = NULL, *temp_filename = NULL;
  int boundary_len = 0, max_file_size = 0, skip_upload= 0, total_bytes = 0, cancel_upload = 0;
  multipart_buffer *mbuff;
  zend_llist header;
  FILE *fp;
  vm_htab_t uploaded_files;
  erequest_info_t *i = REQUEST_INFO (er);

  /* Get the boundary */
  boundary = strstr (ct, "boundary");
  if (!boundary || !(boundary = strchr (boundary, '='))) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
		     "Missing boundary in multipart/form-data POST data");
      return;
    }

  boundary++;
  boundary_len = strlen (boundary);

  if (boundary[0] == '"') 
    {
      boundary++;
      boundary_end = strchr (boundary, '"');
      if (!boundary_end) 
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
			 "Invalid boundary in multipart/form-data POST data");
	  return;
	}
    }
  else
    {
      /* search for the end of the boundary */
      boundary_end = strchr (boundary, ',');
    }
  if (boundary_end) 
    {
      boundary_end[0] = '\0';
      boundary_len = boundary_end-boundary;
    }

  /* Initialize the buffer */
  if (!(mbuff = multipart_buffer_new (boundary, boundary_len))) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
		     "Unable to initialize the input buffer");
      return;
    }

  uploaded_files = 
    vm_htab_create (RFC1867_UPLOADED_FILES_HASH_INITIAL_SIZE, vm_htab_hash_string,
		    (vm_htab_eq) multipart_hash_eq, 
		    (vm_htab_del) multipart_hash_del);

  RCONTENT_RFC1867_UPLOADED_FILES (i) = uploaded_files;

  zend_llist_init (&header, sizeof (mime_header_entry), (llist_dtor_func_t) env_free_hdr_entry);

  while (!multipart_buffer_eof (er, mbuff))
    {
      char buff[FILLUNIT];
      char *cd = NULL, *param = NULL, *filename = NULL, *tmp = NULL;
      int blen = 0, wlen = 0;
      
      zend_llist_clean (&header);

      if (!multipart_buffer_headers (er, mbuff, &header)) 
	{
	  SAFE_RETURN;
	}
      
      if ((cd = multipart_mime_get_hdr_value (header, (char *) "Content-Disposition"))) 
	{
	  char *pair=NULL;
	  
	  while (ISSPACE (*cd)) 
	    {
	      ++cd;
	    }
	  
	  while (*cd && (pair = multipart_ap_getword (&cd, ';')))
	    {
	      char *key=NULL, *word = pair;
	      
	      while (ISSPACE (*cd)) 
		{
		  ++cd;
		}
	      
	      if (strchr (pair, '=')) 
		{
		  key = multipart_ap_getword (&pair, '=');
	      
		  if (!strcasecmp (key, "name")) 
		    {
		      if (param) 
			env_free (param);
		      param = multipart_ap_getword_conf (&pair);
		    } 
		  else if (!strcasecmp (key, "filename")) 
		    {
		      if (filename) 
			env_free (filename);
		      filename = multipart_ap_getword_conf (&pair);
		    }
		}
	      if (key) 
		env_free (key);
	      env_free (word);
	    }

	  /* Normal form variable, safe to read all data into memory */
	  if (!filename && param) 
	    {
	      char *value = multipart_buffer_read_body (er, mbuff);
	      
	      if (!value)
		value = env_strdup ("");

	      new_erequest_args (er, param, ETYPE_STRING, value, strlen (value));

	      if (!strcasecmp (param, "MAX_FILE_SIZE")) 
		max_file_size = atol (value);
	      
	      env_free (param);
	      env_free (value);
	      continue;
	    }

	  /* If file_uploads=off, skip the file part */
	  if (!OPTION_FILE_UPLOADS (ESERVER_OPTION (eserver)))
	    {
	      if (filename)
		env_free (filename);
	      if (param) 
		env_free (param);

	      continue;
	    }
	  
	  /* Return with an error if the posted data is garbled */
	  if (!param) 
	    {
	      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, "File Upload Mime headers garbled");
	      if (filename) 
		env_free (filename);
	      SAFE_RETURN;
	    }

	  /* New Rule: never repair potential malicious user input */
	  if (!skip_upload) 
	    {
	      char *tmp = param;
	      long c = 0;
	      
	      while (*tmp) 
		{
		  if (*tmp == '[') 
		    {
		      c++;
		    }
		  else if (*tmp == ']') 
		    {
		      c--;
		      if (tmp[1] && tmp[1] != '[') 
			{
			  skip_upload = 1;
			  break;
			}
		    }
		  if (c < 0) 
		    {
		      skip_upload = 1;
		      break;
		    }
		  tmp++;				
		}
	    }
	  
	  total_bytes = cancel_upload = 0;

	  if (!skip_upload) 
	    {
	      /* Handle file */
	      fp = multipart_open_tmp_file (OPTION_UPLOAD_TMP_DIR (ESERVER_OPTION (eserver)), "env", &temp_filename);
	      if (!fp) 
		{
		  env_log_error (ENVLOG_MARK, ENVLOG_WARNING, "File upload error - unable to create a temporary file");
		  cancel_upload = UPLOAD_ERROR_E;
		}
	    }
	  if (skip_upload) 
	    {
	      env_free (param);
	      env_free (filename);
	      continue;
	    }

	  if (strlen (filename) == 0) 
	    {
#ifdef DEBUG_FILE_UPLOAD
	      env_log_error (ENVLOG_MARK, ENVLOG_NOTICE, "No file uploaded");
#endif
	      cancel_upload = UPLOAD_ERROR_D;
	    }

	  while (!cancel_upload && (blen = multipart_buffer_read (er, mbuff, buff, sizeof (buff))))
	    {
	      if (OPTION_UPLOAD_MAX_FILESIZE (ESERVER_OPTION (eserver)) > 0 
		  && total_bytes > OPTION_UPLOAD_MAX_FILESIZE (ESERVER_OPTION (eserver))) 
		{
		  env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
				 "upload_max_filesize of %ld bytes exceeded - file [%s=%s] not saved", 
				 OPTION_UPLOAD_MAX_FILESIZE (ESERVER_OPTION (eserver)), param, filename);
		  cancel_upload = UPLOAD_ERROR_A;
		} 
	      else if (max_file_size && (total_bytes > max_file_size)) 
		{
		  env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
				 "MAX_FILE_SIZE of %ld bytes exceeded - file [%s=%s] not saved", 
				 max_file_size, param, filename);
		  cancel_upload = UPLOAD_ERROR_B;
		} 
	      else if (blen > 0) 
		{
		  wlen = fwrite (buff, 1, blen, fp);
		  
		  if (wlen < blen) 
		    {
		      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
				     "Only %d bytes were written, expected to write %d", wlen, blen);
		      cancel_upload = UPLOAD_ERROR_C;
		    } 
		  else 
		    total_bytes += wlen;
		} 
	    } 

	  if (fp)
	    fclose (fp);

#ifdef DEBUG_FILE_UPLOAD
	  if (strlen (filename) > 0 && total_bytes == 0 && !cancel_upload) 
	    {
	      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
			     "Uploaded file size 0 - file [%s=%s] not saved", param, filename);
	      cancel_upload = 5;
	    }
#endif

	  if (cancel_upload) 
	    {
	      if (temp_filename) 
		{
		  if (cancel_upload != UPLOAD_ERROR_E) 
		    { /* file creation failed */
		      unlink (temp_filename);
		    }
		  env_free (temp_filename);
		}
	      temp_filename= (char *) "";
	    }
	  else 
	    multipart_new_uploaded_files (uploaded_files, temp_filename);

	  /* Add $foo_name */
	  if (lbuf) 
	    env_free (lbuf);

	  lbuf = (char *) env_malloc (strlen (param) + MAX_SIZE_OF_INDEX + 1);
			
	  sprintf (lbuf, "%s_name", param);

	  s = strrchr (filename, '\\');
	  if ((tmp = strrchr (filename, '/')) > s)
	    s = tmp;
	
	  if (s && s > filename)
	    new_erequest_args (er, lbuf, ETYPE_STRING, s+1, strlen (s+1));
	  else
	    new_erequest_args (er, lbuf, ETYPE_STRING, filename, strlen (filename));

	  env_free (filename);
	  s = NULL;

	  /* Possible Content-Type: */
	  if (cancel_upload || !(cd = multipart_mime_get_hdr_value (header, (char *) "Content-Type"))) 
	    cd = (char *) "";
	  else 
	    {
	      /* fix for Opera 6.01 */
	      s = strchr (cd, ';');
	      if (s != NULL)
		*s = '\0';
	    }

	  /* Add $foo_type */
	  sprintf (lbuf, "%s_type", param);
	  new_erequest_args (er, lbuf, ETYPE_STRING, cd, strlen (cd));

	  /* Restore Content-Type Header */
	  if (s != NULL)
	    *s = ';';
	  s = (char *) "";

	  sprintf (lbuf, "%s_tmp_name", param);
	  new_erequest_args (er, lbuf, ETYPE_STRING, temp_filename, strlen (temp_filename));

	  sprintf (lbuf, "%s_size", param);
	  new_erequest_args (er, lbuf, ETYPE_LONG, total_bytes);
	  env_free (param);
	}
    }	

  SAFE_RETURN;
}

void
handle_post (er, arg)
     erequest_t *er;
     void *arg;
{
  erequest_info_t *i = REQUEST_INFO (er);
  env_post_entry_t *p = RCONTENT_POST_ENTRY (i);

  if (p && RCONTENT_TYPE (i))
    {
      p->post_handler (er, RCONTENT_TYPE (i), arg);
      if (RCONTENT_POST_DATA (i))
	{
	  env_free (RCONTENT_POST_DATA (i));
	  RCONTENT_POST_DATA (i) = NULL;
	}
    }
}

void
handle_get (er, ht)
     erequest_t *er;
     void *ht ATTRIBUTE_UNUSED;
{
  char *strtok_buf = NULL, *res = NULL, *var, *val;
  erequest_info_t *info = REQUEST_INFO (er);
  
  if (!RCONTENT_QUERY_STRING (info))
    return;
  
  res = apr_pstrdup (REQUEST_POOL (er), RCONTENT_QUERY_STRING (info));
  var = strtok_r (res, "&", &strtok_buf);
  while (var)
    {
      val = strchr (var, '=');
      if (val)
	{
	  int val_len;

	  *val++ = '\0';
	  url_decode (var, strlen (var));
	  val_len = url_decode (val, strlen (val));
	  /* variable  value  Hash table  Ѵ.  */
	  new_erequest_args (er, var, ETYPE_STRING, val, val_len);
	}
      else
	{
	  url_decode (var, strlen (var));
	  new_erequest_args (er, var, ETYPE_STRING, NULL, 0);
	}
      var = strtok_r (NULL, "&", &strtok_buf);
    }
}

static void
env_read_unknown_form_data (er)
     erequest_t *er;
{
  erequest_info_t *i = REQUEST_INFO (er);

  if (!RCONTENT_POST_DATA (i))
    env_read_standard_form_data (er);
}

/* POST data ó  handler  reader ü .  */
enum env_post_entry_enum
  {
    ENUM_DEFAULT_POST_CONTENT_TYPE = 0,
    ENUM_MULTIPART_CONTENT_TYPE,
    ENUM_UNKNOWN_POST_CONTENT_TYPE,
    ENUM_MAX_ENV_POST_ENTRY
  };

static struct env_post_entry env_post_entries[ENUM_MAX_ENV_POST_ENTRY + 1] =
  {
    { DEFAULT_POST_CONTENT_TYPE, sizeof (DEFAULT_POST_CONTENT_TYPE) - 1, env_read_standard_form_data,	env_standard_post_handler },
    /*  ׸  reader κ NULL   rfc1867_post_handler  reader/handler  θ ϱ ̴.  */
    { MULTIPART_CONTENT_TYPE,    sizeof (MULTIPART_CONTENT_TYPE) - 1,    NULL,				rfc1867_post_handler },
    { NULL, 0,								 env_read_unknown_form_data,	NULL},
    { NULL, 0, NULL, NULL }
  };

static void
run_default_post_reader (er)
     erequest_t *er;
{
  env_read_unknown_form_data (er);
}

static void
run_post_reader (er)
     erequest_t *er;
{
  struct env_post_entry *e;
  erequest_info_t *i = REQUEST_INFO (er);

  /* Default POST Content-Type  .  */
  if (!strncasecmp (DEFAULT_POST_CONTENT_TYPE, RCONTENT_TYPE (i), sizeof (DEFAULT_POST_CONTENT_TYPE) - 1))
    e = &env_post_entries[ENUM_DEFAULT_POST_CONTENT_TYPE];
  else if (!strncasecmp (MULTIPART_CONTENT_TYPE, RCONTENT_TYPE (i), sizeof (MULTIPART_CONTENT_TYPE) - 1))
    e = &env_post_entries[ENUM_MULTIPART_CONTENT_TYPE];
  else
    e = &env_post_entries[ENUM_UNKNOWN_POST_CONTENT_TYPE];

  /* POST data  ó , ϱ  env_post_entry ͸ Ѵ.  */
  RCONTENT_POST_ENTRY (i) = e;

  if (e && POST_ENTRY_READER (e))
    e->post_reader (er);
}

void
default_treat_data (er, type)
     erequest_t *er;
     enum treat_data_type type;
{
  erequest_info_t *i = REQUEST_INFO (er);

  /* ߸ type  , ó ʴ´.  */
  if (type != PARSE_GET && type != PARSE_POST && type != PARSE_COOKIE)
    return;

  /* TODO:  GET  COOKIE data  ٷ  ʴ.  ̸ óϱ 
     ƾ ϼؾ  ̴.  */
  switch (type)
    {
    case PARSE_GET:
      {
	handle_get (er, (void *) RCONTENT_ARGS (i));
	break;
      }
    case PARSE_POST:
      {
	handle_post (er, (void *) RCONTENT_ARGS (i));
	return;
      }
    case PARSE_COOKIE:
      break;
    }
}

/* POST data  óѴ.  */
void
process_post_data (er)
     erequest_t *er;
{
  erequest_info_t *i = REQUEST_INFO (er);

  /* GET Ȥ POST data  ޵ ڵ  ARGS Hash table  ȴ.  */
  RCONTENT_ARGS (i) = 
    vm_htab_create (RCONTENT_ARGS_HASH_INITIAL_SIZE, vm_htab_hash_string,
		    (vm_htab_eq) eargs_hash_eq, 
		    (vm_htab_del) eargs_hash_del);

  default_treat_data (er, PARSE_GET);
  default_treat_data (er, PARSE_POST);
}

void
process_interpretref (er)
     erequest_t *er;
{
  int i, j;
#define MAX_ARG_BITMAP	256
  unsigned char argbitmap[MAX_ARG_BITMAP];
  struct interpretref *node = RCONTENT_INTERPRETREF_LIST (REQUEST_INFO (er));

  if (!node)
    return;

  /* Argument Index  Ѵ.  */
  memset (argbitmap, '\0', MAX_ARG_BITMAP);
  while (node)
    {
      argbitmap[INTERPRETREF_ARG_INDEX (node)] = 1;
      node = INTERPRETREF_NEXT (node);
    }

  for (i=0; argbitmap[i] && i<MAX_ARG_BITMAP; i++)
    {
      int endidx = -1, maxidx = -1;
      earrayref_t *aref;
      eobjectref_t **oref;
      struct interpretref *bitmap[MAX_ARG_BITMAP];

      memset (bitmap, '\0', MAX_ARG_BITMAP * sizeof (struct interpretref *));

      node = RCONTENT_INTERPRETREF_LIST (REQUEST_INFO (er));
      while (node)
	{
	  if (INTERPRETREF_ARG_INDEX (node) == i)
	    {
	      if (INTERPRETREF_INDEX (node) > maxidx)
		maxidx = INTERPRETREF_INDEX (node);
	      bitmap[INTERPRETREF_INDEX (node)] = node;
	    }
	  node = INTERPRETREF_NEXT (node);
	}

      /* Ἲ üũ */
      for (j=MAX_ARG_BITMAP-1; j>=0; j--)
	if (bitmap[j])
	  {
	    endidx = j;
	    break;
	  }
      for (j=0; j<endidx; j++)
	if (!bitmap[j])
	  return;

      /* TODO: bitmap     Ϸ ǾǷ ̸ ؾ Ѵ.  
	 Ʒ óؾ  ܰ Ʒ .

	 1. String  ObjectRef  ȯѴ.
	 2. ObjectRef  ArrayRef  Ѵ.
	 3. RCONTENT_CALL_ARGV_IDX () ش ArrayRef  Ѵ.  */
      aref = new_arrayref (lookup_etype ((char *) ENV_STRING_TYPE), maxidx);
      if (!aref)
	return;

      oref = ((eobjectref_t **) ((unsigned char *) aref + EARRAYREF_DATA_OFFSET));
      for (j=0; j<maxidx; j++)
	{
	  evalue_t *ev = INTERPRETREF_VALUE (bitmap[j]);
	  oref[j] = char_to_eobject (EVALUE_STRING (ev), strlen (EVALUE_STRING (ev)));
	}

      RCONTENT_CALL_ARGV_IDX (REQUEST_INFO (er), i) = new_evalue (REQUEST_POOL (er), ETYPE_OBJECTREF, aref);
    }
}

/*  Լ mod_e.c  e_request_begin () Լ ȣȴ.  */
int
e_request_startup (er)
     erequest_t *er;
{
  erequest_info_t *i = REQUEST_INFO (er);

  /* Ʒ request method  ٷ.  */

  if (!REQUEST_SERVER_CONTEXT (er))
    return FAILURE;

  /* Header  Content-Type  ǵǾ ִ Ŀ  POST data ó
     ƾ ޸Ѵ.  */
  if (RCONTENT_TYPE (i))
    /* HTTP POST -> may contain form data to be read into variables
       depending on content type given.  */
    run_post_reader (er);
  else
    /* any other method with content payload will fill 
       $HTTP_RAW_POST_DATA if enabled by always_populate_raw_post_data 
       it is up to the webserver to decide whether to allow a method or not.  */
    run_default_post_reader (er);

  /*  Client   POST data   оǷ,  ̰ VM    
     ֵ data  ó  Ѵ.  */
  process_post_data (er);

  /* process_post_data () Լ ó interpretref  óѴ.  */
  process_interpretref (er);

  return SUCCESS;
}
