/* tsrm.c - Thread Safe Resource Management   .
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

/*
  Thread Safe Resource Manager.

  !!! Ʒ   ̴. Ʈ ʿ !!!

   ƾ erequest_t ü 𿡼 tsrm_fetch  ؼ  
  ֵ ϱ  Ѵ.

  eserver  ޸  ũ   .

    1. erequest_t ޸ Ʈ.
        ޸ Ʈ e_handler () Լ ȣǾ , Ҵǰ   
       ޸𸮰  Ǿ .  ׷ ϳ û  Ǵ ޸ 
       ΰ̴.  ׷ eserver ƾ  ȸ ޸   erequest_t
        ִ memory pool  ؾ Ѵ.

    2. VM ޸ Ʈ.
        ޸ Ʈ ׻ ޸𸮰 ҴǾ ִ κ̴.   ȸ ޸𸮰
       ** ƴϴ.   garbage collection ̳ finalizer thread κ
       ޸𸮰   ִ. 

  tsrm ƾ ϴ ޸ Ʈ (1)  ׸ شϴ ȸ ޸ Ʈ
   зȴٰ   ְڴ.  */

/** Read a resource from a thread's resource storage */
static int tsrm_error_level;
static FILE *tsrm_error_file;

/** Thread local storage */
static pthread_key_t tsrm_key;

static tsrm_entry_t * get_current_tsr (apr_pool_t *);
static void tsrm_error_set (int, char *);

static int 
tsrm_error (int level, const char *format, ...)
{
  if (level <= tsrm_error_level) 
    {
      va_list args;
      int size;
      
      fprintf (tsrm_error_file, "TSRM:  ");
      va_start (args, format);
      size = vfprintf (tsrm_error_file, format, args);
      va_end (args);
      fprintf (tsrm_error_file, "\n");
      fflush (tsrm_error_file);

      return size;
    } 
  else 
    return 0;
}

static void
tsrm_error_set (level, debug_filename)
     int level;
     char *debug_filename ATTRIBUTE_UNUSED;
{
  tsrm_error_level = level;
  
#if TSRM_DEBUG
  if (tsrm_error_file != stderr) 
    /* close files opened earlier */
    fclose(tsrm_error_file);
  
  if (debug_filename) 
    {
      tsrm_error_file = fopen (debug_filename, "w");
      if (!tsrm_error_file) 
	tsrm_error_file = stderr;
    }
  else 
    tsrm_error_file = stderr;
#endif
}

/* Current Thread Resource Main  ȯѴ.  */
static tsrm_entry_t *
get_current_tsr (p)
     apr_pool_t *p;
{
  tsrm_entry_t *r;

  r = pthread_getspecific (tsrm_key);
  if (r)
    return r;

  /* ⿡  ,  thread   resource main   ȵǾ ִٴ
     ̴.  */
  r = (tsrm_entry_t *) apr_pcalloc (p, sizeof (struct tsrm_entry));
  TSRM_ENTRY_COUNT (r) = 0;
  TSRM_ENTRY_SLOTS (r) = apr_pcalloc (p, sizeof (struct tsrm_resource *) * 36);

  /* Set thread local storage to this new thread resources structure */
  pthread_setspecific (tsrm_key, (void *) r);

  return r;
}

void *
tsrm_fetch (index)
     int index;
{
  tsrm_entry_t *r;
  tsrm_resource_t *m;

  r = pthread_getspecific (tsrm_key);
  if (!r)
    return NULL;

  if (!TSRM_ENTRY_SLOTS (r))
    return NULL;
  if (index >= TSRM_ENTRY_COUNT (r))
    return NULL;

  m = TSRM_ENTRY_ENTRY (r, index);
  if (!m)
    return NULL;

  return TSRM_RESOURCE_BUF (m);
}

/* Thread Safe Resource Manager ޸ Ҵ Լ.   tsrm_malloc 
   erequest_t ü Ҵ  ȴ.  */
void *
tsrm_malloc (p, size)
     apr_pool_t *p;
     size_t size;
{
  tsrm_entry_t *r;
  tsrm_resource_t *m;

  /* get_current_tsr () κ ȯ  ׻ 0  ƴ ̴.  */
  r = get_current_tsr (p);

  m = (tsrm_resource_t *) apr_pcalloc (p, sizeof (struct tsrm_resource));
  TSRM_RESOURCE_SIZE (m) = size;
  TSRM_RESOURCE_BUF (m) = (void *) apr_pcalloc (p, size);

  TSRM_ENTRY_COUNT (r)++;
  TSRM_ENTRY_ENTRY (r, TSRM_ENTRY_COUNT (r) - 1) = m;

  return TSRM_RESOURCE_BUF (m);
}

void
tsrm_destory (void)
{
  tsrm_entry_t *r;

  r = pthread_getspecific (tsrm_key);

  pthread_setspecific (tsrm_key, 0);
}

/* Startup TSRM (call once for the entire process) */
int
init_tsrm (debug_level, debug_filename)
     int debug_level;
     char *debug_filename;
{
  pthread_key_create (&tsrm_key, 0);

  tsrm_error_file = stderr;
  tsrm_error_set (debug_level, debug_filename);

  return 1;
}

/* Shutdown TSRM (call once for the entire process) */
void 
finish_tsrm (void)
{
  pthread_setspecific (tsrm_key, 0);
  pthread_key_delete (tsrm_key);
}
