/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include <math.h>

static int  env_htoi (char *);

static int 
env_htoi (s)
     char *s;
{
  int value;
  int c;
  
  c = ((unsigned char *) s)[0];
  if (ISUPPER (c))
    c = TOLOWER (c);
  value = (c >= '0' && c <= '9' ? c - '0' : c - 'a' + 10) * 16;
  
  c = ((unsigned char *) s)[1];
  if (ISUPPER (c))
    c = TOLOWER (c);
  value += c >= '0' && c <= '9' ? c - '0' : c - 'a' + 10;
  
  return (value);
}

/* rfc1738:
   
   ...The characters ";",
   "/", "?", ":", "@", "=" and "&" are the characters which may be
   reserved for special meaning within a scheme...

   ...Thus, only alphanumerics, the special characters "$-_.+!*'(),", and
   reserved characters used for their reserved purposes may be used
   unencoded within a URL...

   For added safety, we only leave -_. unencoded.  */

static unsigned char hexchars[] = "0123456789ABCDEF";

char *
url_encode (char *s, int len, int *new_length)
{
  register int x, y;
  unsigned char *str;
  
  str = (unsigned char *) env_malloc (3 * len + 1);
  for (x = 0, y = 0; len--; x++, y++) 
    {
      str[y] = (unsigned char) s[x];
      if (str[y] == ' ') 
	{
	  str[y] = '+';
	} 
      else if ((str[y] < '0' && str[y] != '-' && str[y] != '.') ||
	       (str[y] < 'A' && str[y] > '9') ||
	       (str[y] > 'Z' && str[y] < 'a' && str[y] != '_') ||
	       (str[y] > 'z')) 
	{
	  str[y++] = '%';
	  str[y++] = hexchars[(unsigned char) s[x] >> 4];
	  str[y] = hexchars[(unsigned char) s[x] & 15];
	}
    }

  str[y] = '\0';
  if (new_length) 
    {
      *new_length = y;
    }

  return ((char *) str);
}

int 
url_decode (str, len)
     char *str;
     int len;
{
  char *dest = str;
  char *data = str;
  
  while (len--) 
    {
      if (*data == '+')
	*dest = ' ';
      else if (*data == '%' && len >= 2 && ISXDIGIT ((int) *(data + 1)) && ISXDIGIT ((int) *(data + 2))) 
	{
	  *dest = (char) env_htoi (data + 1);
	  data += 2;
	  len -= 2;
	}
      else
	*dest = *data;
      data++;
      dest++;
    }
  *dest = '\0';

  return dest - str;
}

/* Print a warning message - output produced, but there may be problems.  */

void
warning (const char *format, ...)
{
  VA_OPEN (ap, format);
  VA_FIXEDARG (ap, const char *, format);

  fprintf (stderr, "warning: ");
  vfprintf (stderr, format, ap);
  VA_CLOSE (ap);
  fputc('\n', stderr);
}

/* Print an error message - we keep going but the output is unusable.  */

void
error (const char *format, ...)
{
  VA_OPEN (ap, format);
  VA_FIXEDARG (ap, const char *, format);

  vfprintf (stderr, format, ap);
  VA_CLOSE (ap);
  fputc('\n', stderr);
}


/**********************************************************************
  Memory Allocation  Լ.
 **********************************************************************/

void *
debug_env_malloc (size)
     size_t size;
{
  return malloc (size);
}

void *
debug_env_calloc (nmemb, size)
     size_t nmemb, size;
{
  return calloc (nmemb, size);
}

void *
debug_env_realloc (ptr, size)
     void *ptr;
     size_t size;
{
  return realloc (ptr, size);
}

void
debug_env_free (ptr)
     void *ptr;
{
  free (ptr);
}

/**********************************************************************
  String  Լ.
 **********************************************************************/
char *
env_strdup (s)
     const char *s;
{
  int length;
  char *p;
  
  length = strlen (s) + 1;

  p = (char *) env_malloc (length);
  if (!p) 
    return (char *)NULL;

  memcpy (p, s, length);

  return p;
}

char *
env_strndup (const char *s, size_t length)
{
  char *p;
  
  p = (char *) env_malloc (length+1);
  if (!p) 
    return (char *)NULL;
  
  memcpy (p, s, length);
  p[length] = 0;

  return p;
}

/**
   Return an string the same as (OLD_NAME, OLD_LENGTH).
   except that characters matching OLD_CHAR are substituted by NEW_CHAR.
   Also, PREFIX is prepended, and SUFFIX is appended.
*/
char *
env_subst (const char* old_name,
	   int old_length,
	   const char *prefix,
	   int old_char,
	   int new_char,
	   const char *suffix)
{
  int prefix_len = strlen (prefix);
  int suffix_len = strlen (suffix);
  int i = prefix_len + old_length + suffix_len + 1;
  char *buffer = env_malloc (i);

  strcpy (buffer, prefix);
  for (i = 0; i < old_length; i++)
    {
      char ch = old_name[i];
      if (ch == old_char)
	ch = new_char;
      buffer[prefix_len + i] = ch;
    }
  strcpy (buffer + prefix_len + old_length, suffix);

  return buffer;
}

/**********************************************************************
  Floating Pointer .
 **********************************************************************/

/*
 * Bit patterns according to IEEE 754.
 */

#define DEXPMASK	(((unsigned long long)0x7ff00000)<<32)
#define DMANMASK	((((unsigned long long)0x000fffff)<<32)|((unsigned long long)0xffffffff))
#define DSIGNBIT	(((unsigned long long)0x80000000)<<32)
#define DINFBITS	(((unsigned long long)0x7ff00000)<<32)
#define DNANBITS	(((unsigned long long)0x7ff80000)<<32)

#define FEXPMASK	0x7f800000
#define FMANMASK	0x007fffff
#define FSIGNBIT	0x80000000
#define FINFBITS	0x7f800000
#define FNANBITS	0x7fc00000

#define DISNAN(b)	(((b) & DEXPMASK) == DEXPMASK && ((b) & DMANMASK) != 0)
#define FISNAN(b)	(((b) & FEXPMASK) == FEXPMASK && ((b) & FMANMASK) != 0)

int
env_cvtdi (double v)
{
  long long vbits;
  
  vbits = (long long) (v);
  if (DISNAN (vbits)) 
    return (0);
  
  if (v < 0.0) 
    v = ceil (v);
  else 
    v = floor (v);

  /* If too small return smallest int */
  if (v <= -2147483648.0) 
    return (-2147483647-1);
  /* If too big return biggest int */
  else if (v >= 2147483647.0) 
    return (2147483647);
  else 
    return ((int) v);
}

long long
env_cvtdl (double v)
{
  long long vbits;
  
  vbits = (long long) v;
  if (DISNAN (vbits)) 
    return ((long long)0);
  
  if (v < 0.0) 
    v = ceil (v);
  else 
    v = floor (v);

  /* If too small return smallest long */
  if (v <= -9223372036854775808.0) 
    return ((long long)1) << 63;
  /* If too big return biggest long */
  else if (v >= 9223372036854775807.0) 
    return ~(((long long)1) << 63);
  else 
    return ((long long)v);
}

/*
 * The following functions round the float/double to an int/long.
 * They round the value toward zero.
 */

long long
env_cvtfl (float v)
{
  int vbits;
  
  vbits = (int) v;
  if (FISNAN (vbits)) 
    return (long long) 0;
  
  if (v < 0.0) 
    v = ceil (v);
  else 
    v = floor (v);

  /* If too small return smallest long */
  if (v <= -9223372036854775808.0) 
    return ((long long) 1) << 63;
  /* If too big return biggest long */
  else if (v >= 9223372036854775807.0) 
    return ~(((long long) 1) << 63);
  else 
    return ((long long) v);
}

int
env_cvtfi (float v)
{
  int vbits;
  
  vbits = (int) v;
  if (FISNAN (vbits)) 
    return 0;

  if (v < 0.0) 
    v = ceil (v);
  else 
    v = floor (v);

  /* If too small return smallest int */
  if (v <= -2147483648.0) 
    return (-2147483647-1);
  /* If too big return biggest int */
  else if (v >= 2147483647.0) 
    return 2147483647;
  else 
    return (int) v;
}

/*
 * env_fcmpg
 */
int
env_fcmpg (float v1, float v2)
{
  int ret;
  int v1bits;
  int v2bits;
  
  v1bits = (int) v1;
  v2bits = (int) v2;
  
  if (FISNAN (v1bits) || FISNAN (v2bits)) 
    ret = 1;
  else if (v1 > v2) 
    ret = 1;
  else if (v1 == v2) 
    ret = 0;
  else 
    ret = -1;
  
  return ret;
}

/*
 * env_fcmpg
 */
int
env_fcmpl (float v1, float v2)
{
  int ret;
  int v1bits;
  int v2bits;
  
  v1bits = (int) (v1);
  v2bits = (int) (v2);

  if (FISNAN (v1bits) || FISNAN (v2bits)) 
    ret = -1;
  else if (v1 > v2) 
    ret = 1;
  else if (v1 == v2) 
    ret = 0;
  else 
    ret = -1;

  return ret;
}

/*
 * env_dcmpg
 */
int
env_dcmpg (double v1, double v2)
{
  int ret;

  if ((!isinf (v1) && isnan (v1)) || (!isinf (v2) && isnan (v2))) 
    ret = 1;
  else if (v1 > v2) 
    ret = 1;
  else if (v1 == v2) 
    ret = 0;
  else 
    ret = -1;
  
  return ret;
}

/*
 * env_dcmpl
 */
int
env_dcmpl (double v1, double v2)
{
  int ret;
  if ((!isinf (v1) && isnan (v1)) || (!isinf (v2) && isnan (v2))) 
    ret = -1;
  else if (v1 > v2) 
    ret = 1;
  else if (v1 == v2) 
    ret = 0;
  else 
    ret = -1;

  return ret;
}

