/* vm_common.c -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"

#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include "etldefaults.h"
#include "econnect.h"

const char *etype_name[MAX_ETYPE] = {
#define DEF_ETYPES(ENUM, NAME, SIZE, SLOT)	NAME,
#include "etype.def"
#undef DEF_ETYPES
};

const int etype_size[MAX_ETYPE] = {
#define DEF_ETYPES(ENUM, NAME, SIZE, SLOT)	SIZE,
#include "etype.def"
#undef DEF_ETYPES
};

const int etype_slot[MAX_ETYPE] = {
#define DEF_ETYPES(ENUM, NAME, SIZE, SLOT)	SLOT,
#include "etype.def"
#undef DEF_ETYPES
};

static apr_pool_t *vmconf;		/* Pool for config stuff */
/** VM  Ǵ  ޸   pool  Ѵ.  */
static apr_pool_t *vmglobal;

static pid_t vm_my_pid;	/* it seems silly to call getpid all the time */
static pid_t vm_parent_pid;

static struct enet_stream *enet_send_ELOAD_initialize (struct enet_stream *, char *, int, char *, envkey_t *);
static env_t enet_send_ELOAD_process (struct enet_stream *, char *, int, char *, envkey_t *);
static void enet_send_ELOAD_finalize (struct enet_stream *);
static int etype_hash_eq (const void *, const void *);
static void etype_hash_del (void *);
static etype_t * new_etype (enum evalue_type, int, const char *, int, void *);
static int is_localhost (char *);

int vm_e_tree_compare (vm_splay_tree_key, vm_splay_tree_key);
int vm_eref_tree_compare (vm_splay_tree_key, vm_splay_tree_key);
UNIQUE_ID_TYPE get_unique_id (void);
env_t make_envstruct (envkey_t *);
int release_envstruct (env_t);
env_t lookup_envstruct (envkey_t *);
env_t make_envstruct_with_name (const char *);
env_t load_env (char *);
env_t load_env_types (char *);
etype_t * lookup_etype_array (etype_t *);
void vm_init_etypes (vm_t *);
int vm_init_global_struct (eserver_t *, apr_pool_t *);
int vm_finish_global_struct (void);
env_t alloc_env (enum env_struct_type);
env_t enet_send_ELOAD (char *, int, char *, envkey_t *);
env_t make_reo_envstruct (envkey_t *);
env_t get_reo_envstruct (envkey_t *);
env_t make_reo_envstruct_with_name (const char *);
struct enet_stream * alloc_enet_stream (void);
void free_enet_stream (struct enet_stream *);
inline int instanceof_array (env_t, env_t);
inline int instanceof_interface (env_t, env_t);
inline int instanceof_class (env_t, env_t);
int instanceof (env_t, env_t);
void init_native (eserver_t *, apr_pool_t *);
void dump_memory (void *, int);
void dump_arrayref (char *, earrayref_t *);
void dump_objectref (eobjectref_t *);
void dump_evalue_core (char *, evalue_t *, int);
void throw_NullPointerException (void);
void print_indent (FILE *, int);
void dump_objectref_core (eobjectref_t *, int);

static void 
vm_global_mutex_on (void)
{
  apr_status_t rv = apr_thread_mutex_lock (VM_GLOBAL_MUTEX (ESERVER_VM (eserver)));

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't grab the vm global mutex";

      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

static void 
vm_global_mutex_off (void)
{
  apr_status_t rv = apr_thread_mutex_unlock (VM_GLOBAL_MUTEX (ESERVER_VM (eserver)));

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't release the vm global mutex";
      
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

/** 
   Splay Tree  ƾ.  K1  K2   , 0  ȯѴ.  ׷  ,
   -1 Ȥ 1  ȯȴ.

   @param k1 Key 1 
   @param k2 Key 2 
   @return  K1  K2    0  ȯѴ.
*/
int
vm_e_tree_compare (k1, k2)
     vm_splay_tree_key k1;
     vm_splay_tree_key k2;
{
  envkey_t *s1, *s2;

  s1 = (envkey_t *) k1;
  s2 = (envkey_t *) k2;

  return compare_envkey (s1, s2);
}

/** 
   Objecref  Arrayref Splay Tree  ƾ.  K1  K2   , 0  ȯѴ.
   ׷  , -1 Ȥ 1  ȯȴ.

   @param k1 Key 1 
   @param k2 Key 2 
   @return  K1  K2    0  ȯѴ.
*/
int
vm_eref_tree_compare (k1, k2)
     vm_splay_tree_key k1;
     vm_splay_tree_key k2;
{
  int s1, s2;

  s1 = (int) k1;
  s2 = (int) k2;

  if (s1 == s2)
    return 0;
  else if (s1 > s2)
    return 1;
  else
    return -1;
}

/** Hash Table  Ǵ ϼ ׽Ʈ Լν, ENTRY  
   Hash Table  ִ SLOT  ̰, X_ARG  vm_htab_find_slot_with_hash 
   ڷ  ̴.  */
int
strconst_hash_eq (entry, x_arg)
     const void *entry, *x_arg;
{
  strconst_t *sa;
  char *sb;
  
  sa = (strconst_t *) entry;
  sb = (char *) x_arg;

  if (STRCONST_PTR (sa) && sb && !strcmp (STRCONST_PTR (sa), sb))
    return 1;

  return 0;
}

/** Hash Table   Slot  ϵǾ ִ entry  delete   Ǵ Լ̴.
   vm_htab_create () Լ ȣ , Լ Ͱ  ϵȴ.  */
void
strconst_hash_del (entry)
     void *entry;
{
  strconst_t *s = (strconst_t *) entry;

  env_free (STRCONST_PTR (s));
  env_free (s);
}

/** SA  SB   ,  , 1  ȯϰ, ׷  ,
   0  ȯѴ.  */
int
strconstcmp (sa, sb)
     strconst_t *sa, *sb;
{
  if (!strncmp (STRCONST_PTR (sa), STRCONST_PTR (sb), STRCONST_LENGTH (sa)))
    return 1;

  return 0;
}

/**
   ENV_CONSTANT_HASH_TABLE (e) hash table  ̰ LEN  ڿ S  ߰Ѵ.
     node  Ѵٸ  node  ȯѴ.
   LEN  0  , S  ̸ Ͽ Ѵ.  

   20050730 bugfix: CONSTANT_Utf8  ̰  0  ,  Ȯϰ
   ϵ . - weongyo
*/
strconst_t *
new_strconst (e, s, len)
     env_t e;
     const char *s;
     int len;
{
  char *buf;
  vm_hashval_t hash;
  strconst_t **slot, *entry;

  if (len < 0)
    return NULL;

  buf = (char *) apr_pcalloc (ENV_MPOOL (e), sizeof (char) * len + 1);
  memcpy (buf, s, len);
  buf[len] = '\0';

  hash = vm_htab_hash_string (buf);

  slot = (strconst_t **) vm_htab_find_slot_with_hash (ENV_CONSTANT_HASH_TABLE (e), buf, hash, INSERT);
  entry = *slot;

  /*  Ѵٸ, ش node  ȯѴ.  */
  if (entry)
    return entry;

  /*   ʴ´ٸ, slot  ο node  ߰Ѵ.  */
  entry = (strconst_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct string_constant));
  STRCONST_HASH (entry) = hash;
  STRCONST_PTR (entry) = buf;
  STRCONST_LENGTH (entry) = len;

  *slot = entry;
  
  return entry;
}


/** THIS environment , ̸ NAME , type  TYPE  method 
   pointer  ȯѴ.  */
emethod_t *
find_method (e, name, type)
     env_t e;
     strconst_t *name, *type;
{
  int n;
  emethod_t *mptr;

  if (!e || !name || !type)
    return NULL;

  n = ENV_METHODS_COUNT (e);

  for (mptr = ENV_METHODS (e); --n >= 0; ++mptr) 
    {
      if (strconstcmp (name, METHOD_NAME (mptr)) 
	  && strconstcmp (type, MTYPE_NAME (METHOD_TYPE (mptr)))) 
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
			 "find_method (%s,%s,%s) -> 0x%x",
			 STRCONST_PTR (ENV_NAME (e)), STRCONST_PTR (name), STRCONST_PTR (type), mptr);
	  return mptr;
	}
    }
  
  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "find_method_local(%s,%s,%s) -> NOT FOUND",
		 STRCONST_PTR (ENV_NAME (e)), STRCONST_PTR (name), STRCONST_PTR (type));

  return NULL;
}

/**
 * Lookup a named field.  Do not search super classes. Do not resolve the field.
 */
efield_t *
lookup_field (clp, name, is_static)
     env_t clp;
     strconst_t *name;
     int is_static;
{
  efield_t *fptr;
  int n;
  
  /* Search down class for field name */
  if (is_static) 
    {
      fptr = ENV_STATIC_FIELDS (clp);
      n = ENV_STATIC_FIELDS_COUNT (clp);
    }
  else 
    {
      fptr = &ENV_INSTANCE_FIELDS (clp);
      n = ENV_INSTANCE_FIELDS_COUNT (clp);
    }

  while (--n >= 0) 
    {
      if (strconstcmp (name, FIELD_NAME (fptr))) 
	return fptr;
      fptr++;
    }

  return NULL;
}

extern env_t get_envstruct_with_idx (int, env_t);

/** IDX  CONSTANT_Fieldref ν, struct field_info  RET  ش 
    ä ȯѴ.  machine.c Ͽ ȴ.  */
int
get_field (idx, e, is_static, info)
     int idx, is_static;
     env_t e;
     struct efield_info *info;
{
  int ci, ni;
  econst_pool_t *pool;
  efield_t *field;
  env_t e2;
  
  pool = ENV_CONSTANT_POOL (e);
  
  if (CONSTPOOL_TAGS_IDX (pool, idx) != CONSTANT_Fieldref) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "No Fieldref found");
      return -1;
    }
  
  ci = FIELDREF_ENVIRONMENT (pool, idx);

  /* ⼭ get_envstruct_with_idx ()  ȣϴ   óϰ ִ ENVIRONMENT  
     load Ǿ ִ Ȯϰ, ׷  , LOAD ϰ ּ ESTATE_LINKED  
     ó ־ ϱ ̴.  */
  e2 = get_envstruct_with_idx (ci, e);
  if (!e2)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "ENV  NULL");
      return -1;
    }

  ni = FIELDREF_NAMEANDTYPE (pool, idx);
  
  EFIELD_INFO_NAME (info) = (strconst_t *) CONSTPOOL_DATA_IDX (pool, NAMEANDTYPE_NAME (pool, ni));
  EFIELD_INFO_TYPE (info) = (strconst_t *) CONSTPOOL_DATA_IDX (pool, NAMEANDTYPE_TYPE (pool, ni));
  
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "*** get_field (%s,%s,%s)",
		 STRCONST_PTR (ENV_NAME (e2)), 
		 STRCONST_PTR (EFIELD_INFO_NAME (info)), STRCONST_PTR (EFIELD_INFO_TYPE (info)));
  
  field = lookup_field (e2, EFIELD_INFO_NAME (info), is_static);
  if (!field)
    return -1;
  
  EFIELD_INFO_FIELD (info) = field;
  EFIELD_INFO_ENV (info) = e2;

  return 0;
}

/** Lookup a method reference and get the various interesting bits.
 
   Return false if unsuccessful because of a malformed class file
   or if the class couldn't be found or processed.
   Returns true otherwise. 
 
   Note that even if it returns true, the method may not be found.
   call->method is set to NULL in this case.  */
int
get_method_type_environment (idx, e, load_env, is_special, call)
     int idx, load_env, is_special;
     env_t e;
     struct emethod_info *call;
{
  int ni, ci;
  env_t env = NULL;
  econst_pool_t *pool;
  emethod_t *mptr = NULL;
  
  EMETHOD_INFO_ENV (call) = 0;
  EMETHOD_INFO_METHOD (call) = 0;
  EMETHOD_INFO_NAME (call) = 0;
  EMETHOD_INFO_TYPE (call) = 0;

  pool = ENV_CONSTANT_POOL (e);
  if (CONSTPOOL_TAGS_IDX (pool, idx) != CONSTANT_Methodref &&
      CONSTPOOL_TAGS_IDX (pool, idx) != CONSTANT_InterfaceMethodref) 
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "No Methodref found for idx = %d", idx);
      return -1;
    }
  
  ni = METHODREF_NAMEANDTYPE (pool, idx);
  EMETHOD_INFO_NAME (call) = (strconst_t *) CONSTPOOL_DATA_IDX (pool, NAMEANDTYPE_NAME (pool, ni));
  EMETHOD_INFO_TYPE (call) = (strconst_t *) CONSTPOOL_DATA_IDX (pool, NAMEANDTYPE_TYPE (pool, ni));

  if (load_env)
    {
      ci = METHODREF_ENVIRONMENT (pool, idx);
      env = get_envstruct_with_idx (ci, e);
      if (!env)
	return -1;

      /* is_special  1 ̶  INVOKESPECIAL  ȣǾٴ ̴.  */
      if (is_special == 1
	  && (!strconstcmp (EMETHOD_INFO_NAME (call), new_strconst (e, "<init>", 6)))
	  && e != env
	  && instanceof (env, e))
	env = ENV_SUPER (e);

      EMETHOD_INFO_ENV (call) = env;

      /* is_special  1 ̶  INVOKEINTERFACE  ȣǾٴ ̴.  */
      if (is_special == 2)
	{
	  mptr = find_method (env, EMETHOD_INFO_NAME (call), EMETHOD_INFO_TYPE (call));
	  if (!mptr) 
	    {
	      int i;

	      for (i = ENV_TOTAL_INTERFACE_NUM (env) - 1; i >= 0; i--) 
		{
		  mptr = find_method (ENV_INTERFACE_IDX (env, i), EMETHOD_INFO_NAME (call), EMETHOD_INFO_TYPE (call));
		  if (mptr)
		    break;
		}
	      
	      if (!mptr)
		mptr = find_method (ENV_SUPER (EMETHOD_INFO_ENV (call)), EMETHOD_INFO_NAME (call), EMETHOD_INFO_TYPE (call));
	    }
	  EMETHOD_INFO_METHOD (call) = mptr;
	}
      else
	{
	  /* INVOKEVIRTUAL Ȥ INVOKESPECIAL, INVOKESTATIC   ´.  */
	  while (env)
	    {
	      mptr = find_method (env, EMETHOD_INFO_NAME (call), EMETHOD_INFO_TYPE (call));
	      if (mptr)
		{
		  EMETHOD_INFO_METHOD (call) = mptr;
		  break;
		}
	      env = ENV_SUPER (env);
	    }
	}
    }

  if (load_env)
    env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		   "get_method_type_environment (%s,%s,%s) -> succsess",
		   STRCONST_PTR (ENV_NAME (EMETHOD_INFO_ENV (call))), 
		   STRCONST_PTR (EMETHOD_INFO_NAME (call)), STRCONST_PTR (EMETHOD_INFO_TYPE (call)));

  return 0;
}


/** Environment ü ݵ ڱ⸸  unique id  .  ̰  Process 
   ID  ϴ.  ̰ ϴ VM_CURRENT_UNIQUE_ID   thread    ޴
   ¶ thread-safe ϰ Ǿ Ѵ.  */
UNIQUE_ID_TYPE
get_unique_id (void)
{
  int i;

  vm_global_mutex_on ();

  i = VM_CURRENT_UNIQUE_ID (ESERVER_VM (eserver))++;

  vm_global_mutex_off ();

  return i;
}

/**
    T  ´ environment ü ҴϿ ȯѴ.
*/
env_t
alloc_env (t)
     enum env_struct_type t;
{
  env_t e = NULL;

  switch (t)
    {
    case ENV_LEO:
      e = (env_t) egc_malloc (sizeof (struct env_leo), EGC_ALLOC_CLASS);
      break;
    case ENV_REO:
      e = (env_t) egc_malloc (sizeof (struct env_reo), EGC_ALLOC_CLASS);
      break;
    default:
      break;
    }

  ENV_TYPE (e) = t;

  return e;
}

/** 
   KEY   ּ  env_t ü Ҵϰ ȯѴ.   ü
    environment ̳ʸ ؼ ȵ ̴.  
*/
env_t
make_envstruct (key)
     envkey_t *key;
{
  env_t e;
  apr_pool_t *mp;
  apr_status_t rv;

  apr_pool_create (&mp, VM_GLOBAL_MEMORY_POOL (ESERVER_VM (eserver)));
  apr_pool_tag (mp, "local env mpool");

  e = (env_t) alloc_env (ENV_LEO);
  ENV_UNIQUE_ID (e) = get_unique_id ();  
  ENV_MPOOL (e) = mp;

  rv = apr_thread_mutex_create (&ENV_MUTEX (e), APR_THREAD_MUTEX_DEFAULT, ENV_MPOOL (e));
  if (rv != APR_SUCCESS) 
    {
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, ap_server_conf,
		    "Couldn't create env global lock");
      return NULL;
    }

  ENV_KEY (e) = key;

  ENV_SET_STATE (e, ENV_STATE_0_STRUCT_INITIALIZED);

  ENV_CONSTANT_HASH_TABLE (e) = 
    vm_htab_create (CONSTANT_HASH_INITIAL_SIZE, vm_htab_hash_string,
		    (vm_htab_eq) strconst_hash_eq, 
		    (vm_htab_del) strconst_hash_del);

  ENV_CONSTANT_POOL (e) = 
    (econst_pool_t *) apr_pcalloc (ENV_MPOOL (e), sizeof (struct econst_pool));

  return e;
}

/**
   Key   Ͽ Remote Environment ü Ѵ.
*/
env_t
make_reo_envstruct (key)
     envkey_t *key;
{
  env_t e;
  apr_pool_t *mp;
  apr_status_t rv;

  apr_pool_create (&mp, VM_GLOBAL_MEMORY_POOL (ESERVER_VM (eserver)));
  apr_pool_tag (mp, "remote env mpool");

  e = (env_t) alloc_env (ENV_REO);
  ENV_UNIQUE_ID (e) = get_unique_id ();
  ENV_MPOOL (e) = mp;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "make_envstruct: memory pool 0x%x", ENV_MPOOL (e));

  rv = apr_thread_mutex_create (&ENV_MUTEX (e), APR_THREAD_MUTEX_DEFAULT, ENV_MPOOL (e));
  if (rv != APR_SUCCESS) 
    {
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, ap_server_conf,
		    "Couldn't create env global lock");
      return NULL;
    }

  ENV_KEY (e) = key;
  ENV_SET_STATE (e, ENV_STATE_COMPLETE);

  return e;
}

/** env_t ü E   Ҵ ҽ Ѵ.  */
int
release_envstruct (e)
     env_t e;
{
  if (ENV_MPOOL (e))
    apr_pool_destroy (ENV_MPOOL (e));

  FREE_ENVKEY (ENV_KEY (e));

  vm_htab_delete (ENV_CONSTANT_HASH_TABLE (e));

  env_free (e);

  return 0;
}

/**
   KEY   environment  splay tree  ϵǾ ִ Ȯ ,
   node   , ش env_t  ȯϰ, ׷ ʴٸ NULL  ȯѴ.
*/
env_t
lookup_envstruct (key)
     envkey_t *key;
{
  vm_splay_tree_node_t n;

  if (!key)
    return NULL;

  n = vm_splay_tree_lookup (VM_E_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) key);
  if (n)
    return (env_t) n->value;

  return NULL;
}

env_t load_env_core (char *, int, envkey_t *, char *);

/** E  constant pool  ִ IDX  ENVIRONMENT  LOAD Ѵ.  */
env_t
get_envstruct_with_idx (idx, e)
     int idx;
     env_t e;
{
  int tag;
  int host, hostname_idx, hostport_idx, hostport;
  int sha1nname, sha1_idx, name_idx;
  envkey_t key;
  strconst_t *name, *hostname;
  ENVKEY_CHAR *sha1;
  econst_pool_t *pool = ENV_CONSTANT_POOL (e);

  tag = CONSTPOOL_TAGS_IDX (pool, idx);
  switch (tag)
    {
    case CONSTANT_Environment:
      /* CONSTANT_Environment  ΰ field   ִ.
	 High κ HOST   ִ.
	 Low  κ SHA1   ̸   ִ.  */
      host = ENVIRONMENT_HOST (pool, idx);
      hostname_idx = (u2) HOST_NAME (pool, host);
      hostport_idx = (u2) HOST_PORT (pool, host);
      sha1nname = ENVIRONMENT_SHA1ANDNAME (pool, idx);
      sha1_idx = (u2) SHA1ANDNAME_SHA1 (pool, sha1nname);
      name_idx = (u2) SHA1ANDNAME_NAME (pool, sha1nname);

      hostname = (strconst_t *) CONSTPOOL_DATA_IDX (pool, hostname_idx);
      hostport = (int) CONSTPOOL_DATA_IDX (pool, hostport_idx);
      sha1 = (ENVKEY_CHAR *) CONSTPOOL_DATA_IDX (pool, sha1_idx);
      name = (strconst_t *) CONSTPOOL_DATA_IDX (pool, name_idx);
      PUT_ENVKEY (sha1, &key);

      return load_env_core (STRCONST_PTR (hostname), hostport, &key, STRCONST_PTR (name));

    default:
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		     "get_envstruct_with_idx: tag error = 0x%08x. idx = 0x%08x", tag, idx);
      break;
    }

  return NULL;
}

/** KEY   environment  splay tree  ϵǾ ִ Ȯ ,
   node   , ش env_t  ȯϰ, ׷ ʴٸ ο env_t ü
   Ҵϰ, splay tree  ߰Ѵ.  */
env_t
get_envstruct (key)
    envkey_t *key;
{
  env_t env;
  envkey_t *splaytree_key;

  env = lookup_envstruct (key);
  if (env)
    return env;

  /* DUPLICATE_ENVKEY   ޸𸮴 release_envstruct () Լ  ̴.  */
  DUPLICATE_ENVKEY (key, splaytree_key);
  
  /* Splay Tree  key   node   ʴ´ٸ, ش node  ϰ
     Ѵ.  */
  env = make_envstruct (splaytree_key);
  if (!env)
    return NULL;

  vm_splay_tree_insert (VM_E_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) splaytree_key,
			(vm_splay_tree_value) env);

  return env;
}

/**
   get_envstruct ()  remote ̴.
*/
env_t
get_reo_envstruct (key)
    envkey_t *key;
{
  env_t env;
  envkey_t *splaytree_key;

  env = lookup_envstruct (key);
  if (env)
    return env;

  /* DUPLICATE_ENVKEY   ޸𸮴 release_envstruct () Լ  ̴.  */
  DUPLICATE_ENVKEY (key, splaytree_key);
  
  /* Splay Tree  key   node   ʴ´ٸ, ش node  ϰ
     Ѵ.  */
  env = make_reo_envstruct (splaytree_key);
  if (!env)
    return NULL;

  vm_splay_tree_insert (VM_E_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) splaytree_key,
			(vm_splay_tree_value) env);

  return env;
}

/**
   get_envstruct ()  , KEY  ̸ ƴ, ڿ STR ̴.

   ̰  ʿ  primitive environment   ؼ̴.  
   Primitive Environment  environment ̳ʸ ó ߿ Ǿ 
   ν STR  ̿Ͽ KEY   , splay tree  ϴ 
   ϰ ִ.  
*/
env_t
make_envstruct_with_name (str)
     const char *str;
{
  int keysize;
  /*  迭 static  Ǿ ִµ,  Լ  primitive type  
     ؼ Ǿ Ѵ.  */
  char tmp[256];
  ENVKEY_CHAR s[ENVKEY_SIZE];
  envkey_t key;

  /* STR  , SHA1  .  */
  apr_sha1_base64 (str, strlen (str), tmp);
  keysize = apr_base64_decode_binary (s, tmp);
  if (keysize != ENVKEY_SIZE)
    return NULL;

  PUT_ENVKEY (s, &key);

  return get_envstruct (&key); 
}

/**
   make_envstruct_with_name ()  , Remote ü  Ҵ ̷
   .
*/
env_t
make_reo_envstruct_with_name (str)
     const char *str;
{
  int keysize;
  /*  迭 static  Ǿ ִµ,  Լ  primitive type  
     ؼ Ǿ Ѵ.  */
  char tmp[256];
  ENVKEY_CHAR s[ENVKEY_SIZE];
  env_t e;
  envkey_t key;
  apr_status_t rv;

  /* STR  , SHA1  .  */
  apr_sha1_base64 (str, strlen (str), tmp);
  keysize = apr_base64_decode_binary (s, tmp);
  if (keysize != ENVKEY_SIZE)
    return NULL;

  PUT_ENVKEY (s, &key);

  e = get_reo_envstruct (&key);

  /*   ٷ   E ü memory pool  .  */
  rv = euri_parse (ENV_MPOOL (e), str, &RENV_PARSED_EURI (e));
  if (rv != APR_SUCCESS)
    {
      /* TODO: release_reo_envstruct () Լ ؾ Ѵ.  */
      return NULL;
    }
  RENV_UNPARSED_EURI (e) = apr_pstrdup (ENV_MPOOL (e), str);
  RENV_HOSTNAME (e) = EURI_HOSTNAME (&RENV_PARSED_EURI (e));
  RENV_PORT (e) = EURI_PORT (&RENV_PARSED_EURI (e));
  RENV_NAME (e) = EURI_PATH (&RENV_PARSED_EURI (e));
  RENV_NAMELEN (e) = EURI_PATH (&RENV_PARSED_EURI (e)) ? strlen (EURI_PATH (&RENV_PARSED_EURI (e))) : 0;

  return e;
}

/**
   Environment Struct   ϵ lock Ѵ.  
*/
void 
lock_envstruct (e)
     env_t e;
{
  apr_status_t rv;

  if (!e)
    return;

  if (ENV_NAME (e))
    env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		   "Locking %s", STRCONST_PTR (ENV_NAME (e)));

  rv = apr_thread_mutex_lock (ENV_MUTEX (e));

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't grab the environment struct mutex";

      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

/** 
    Environment Struct   Ѵ.  
*/
void 
unlock_envstruct (e)
     env_t e;
{
  apr_status_t rv;

  if (!e)
    return;

  if (ENV_NAME (e))
    env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		   "Unlocking %s", STRCONST_PTR (ENV_NAME (e)));

  rv = apr_thread_mutex_unlock (ENV_MUTEX (e));

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't release the environment struct mutex";
      
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

/**
    ObjectRef  locking Ų.  mutex  ObjectRef     
   ȴ.
 */
void
lock_eobjectref (objref)
     eobjectref_t *objref;
{
  apr_status_t rv;

  if (!objref)
    return;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "Locking ObjectRef(0x%x) %s",
		 objref,
		 STRCONST_PTR (ENV_NAME (EOBJECTREF_ENV (objref))));

  rv = apr_thread_mutex_lock (EOBJECTREF_MUTEX (objref));

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't grab the environment struct mutex";

      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

/**
    ObjectRef  unlocking Ų.
 */
void
unlock_eobjectref (objref)
     eobjectref_t *objref;
{
  apr_status_t rv;

  if (!objref)
    return;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "Unlocking ObjectRef(0x%x) %s",
		 objref,
		 STRCONST_PTR (ENV_NAME (EOBJECTREF_ENV (objref))));

  rv = apr_thread_mutex_unlock (EOBJECTREF_MUTEX (objref));

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't release the environment struct mutex";
      
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

/**
    method  synchronizing  ʿ , object  Ѵ.  
     static    ƴ , local  ù° ڸ locking 
   Ѵ. 
   
   @param af	method  access flag ̴.
*/
void
lock_synchronized (m, local)
     emethod_t *m;
     evalue_t *local;
{
  if (METHOD_ACCESS_FLAGS (m) & AF_SYNCHRONIZED)
    {
      if (METHOD_ACCESS_FLAGS (m) & AF_STATIC)
	lock_envstruct (METHOD_ENV (m));
      else if (METHOD_ACCESS_FLAGS (m) & AF_ENI)
	lock_eobjectref ((eobjectref_t *) EVALUE_ADDRESS (local + 1));
      else
	lock_eobjectref ((eobjectref_t *) EVALUE_ADDRESS (local));
    }
}

void
unlock_synchronized (m, local)
     emethod_t *m;
     evalue_t *local;
{
  if (METHOD_ACCESS_FLAGS (m) & AF_SYNCHRONIZED)
    {
      if (METHOD_ACCESS_FLAGS (m) & AF_STATIC)
	unlock_envstruct (METHOD_ENV (m));
      else if (METHOD_ACCESS_FLAGS (m) & AF_ENI)
	unlock_eobjectref ((eobjectref_t *) EVALUE_ADDRESS (local + 1));
      else
	unlock_eobjectref ((eobjectref_t *) EVALUE_ADDRESS (local));
    }
}

/** HOSTNAME  localhost ȣƮ ˻Ѵ.   ׷ٸ 1 . ׷ 
   ʴٸ 0  ȯѴ.   HOSTNAME   ִٸ -1  ȯѴ.  */
static int
is_localhost (hostname)
     char *hostname;
{
  eoption_t *op = ESERVER_OPTION (eserver);

  if (!hostname)
    return -1;

  /* TODO:  localhost  ˻ϴ ƾ   ʴ.   ʿϴ.  */
  if (!strncmp (hostname, "127.0.0.1", 9))
    return 1;
  else if (!strncmp (hostname, "localhost", 9))
    return 1;
  else if (OPTION_HOSTNAME (op) && !strcmp (hostname, OPTION_HOSTNAME (op)))
    return 1;

  return 0;
}

struct enet_stream
{
  int status;

  unsigned char *buffer;
  unsigned char *buffer_end;
  unsigned int len;
};

/** ENET Stream  ҴѴ.  */
struct enet_stream *
alloc_enet_stream (void)
{
  struct enet_stream *es = env_malloc (sizeof (struct enet_stream));

  /* TODO: ʱȭ ƹ͵ ϴ .  */

  return es;
}

/** ENET Stream  Ѵ.  */
void 
free_enet_stream (es)
     struct enet_stream *es;
{
  env_free (es->buffer);
  env_free (es);
}

/** 
   Remote Class  û  ް Ǵ response   callback ԼԴϴ.  
    arg  buffer  ϴ   Լ ̴.  
*/
static apr_status_t
enet_send_ELOAD_stream_handler (ec_response_t *res, void *arg)
{
  struct enet_stream *es = (struct enet_stream *) arg;

  es->status = res->status;
  es->len = res->bodylen;
  es->buffer = env_malloc (res->bodylen);
  es->buffer_end = es->buffer + res->bodylen;
  memcpy (es->buffer, res->body, res->bodylen);
  
  return 0;
}

/**
   enet_send_ELOAD ʱȭѴ.  
*/
static struct enet_stream *
enet_send_ELOAD_initialize (es, hostname, port, name, key)
     struct enet_stream *es;
     char *hostname, *name;
     int port;
     envkey_t *key;
{
  ec_client_t *conn;
  ec_request_t *req;
  ec_response_t *res;
  char keystr[ENVKEY_STRSIZE];

  conn = ec_client_new (hostname, port);
  if (!conn)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "Failed to process ec_client_new () function.");
      return NULL;
    }
  if (ec_client_connect (conn))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "Can't connect to e://%s:%d/%s", hostname, port, name);
      return NULL;
    }

  req = ec_request_new (conn, "ELOAD", name);
  if (key)
    {
      GET_ENVKEY_STRING (key, keystr);
      ec_request_add_arg (req, "__KEY", keystr);
    }
  ec_request_set_handler (req, enet_send_ELOAD_stream_handler, es);
  res = ec_request_send (req);
  if (!res)
    {
      ec_client_close (conn);
      free_enet_stream (es);
      return NULL;
    }
  ec_client_close (conn);

  return es;
}

/**
   enet_send_ELOAD  óѴ.  
*/
static env_t
enet_send_ELOAD_process (es, hostname, port, name, key)
     struct enet_stream *es;
     char *hostname, *name;
     int port;
     envkey_t *key;
{
  char fullname[MAX_STRING_SIZE];
  char keystr[ENVKEY_STRSIZE];

  /*     HTTP_OK  Ա , ڷ Ѿ  
     fake env struct  Ѵ.  */
  if (es->status != HTTP_OK)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		     "enet_send_ELOAD: !!޼ Ȯϰ  ּ!!");
      return NULL;
    }

  GET_ENVKEY_STRING (key, keystr);

  /* make_reo_envstruct_with_name () Լ ڰ Ѿ ڿ
     e://<hostname>:<port>/<name>('<key>')    Ѿ KEY  
     ߺ  , ׻ unique  ¿ Ѵ.   object   KEY 
     ޶󼭴  ȵȴ.  */
  snprintf (fullname, MAX_STRING_SIZE,
	    "e://%s:%d%s('%s')", hostname, port, name, keystr);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "enet_send_ELOAD_process: %s", fullname);

  return make_reo_envstruct_with_name (fullname);
}

/**
   enet_send_ELOAD ó Ѵ.  
*/
static void
enet_send_ELOAD_finalize (es)
     struct enet_stream *es;
{
  free_enet_stream (es);  
}

/**
   ǻ ּҰ HOSTNAME ̰ Ʈ ȣ PORT  ǻͿ NAME  KEY 
    environment  ´. 
   
   TODO:  Լ ٸ Ǵ Լ   ٸ Ϸ и ʿ䰡 
   ̴.
*/
env_t
enet_send_ELOAD (hostname, port, name, key)
     char *hostname, *name;
     int port;
     envkey_t *key;
{
  env_t e;
  struct enet_stream *es = alloc_enet_stream ();

  /*  */
  enet_send_ELOAD_initialize (es, hostname, port, name, key);
  /* ó */
  e = enet_send_ELOAD_process (es, hostname, port, name, key);
  /* Ҹ */
  enet_send_ELOAD_finalize (es);

  return NULL;
}

env_t
load_env_core (hostname, port, key, name)
     char *hostname, *name;
     int port;
     envkey_t *key;
{
  env_t r = NULL;
  eoption_t *op = ESERVER_OPTION (eserver);

  if (is_localhost (hostname))
    {
      /* HostName  local  , Ʒ óѴ.  */
      int fpathlen;
      char *fpath;

      fpathlen = 
	strlen (OPTION_ERE_ROOT (op))
	+ 2 /* DIR_SEPARATOR .  */
	+ strlen (name)
	+ strlen (ENV_FILE_EXTENSION) /* Ȯ   */
	+ 1; /* NULL  */
      fpath = (char *) env_malloc (fpathlen);

      if (name[0] == DIR_SEPARATOR)
	snprintf (fpath, fpathlen - 1, "%s%s%s", OPTION_ERE_ROOT (op), name, ENV_FILE_EXTENSION);
      else
	snprintf (fpath, fpathlen - 1, "%s%c%s%s", OPTION_ERE_ROOT (op), DIR_SEPARATOR, name, ENV_FILE_EXTENSION);

      r = efile_readnload (fpath);

      env_free (fpath);
    }
  else
    r = enet_send_ELOAD (hostname, port, name, key);

  return r;
}

/** STR  ݴ e:// scheme  ؾ Ѵ.  */
env_t
load_env (euri)
     char *euri;
{
  int rv;
  vm_t *vm = ESERVER_VM (eserver);
  euri_t *eu = (euri_t *) apr_pcalloc (VM_GLOBAL_MEMORY_POOL (vm), sizeof (struct euri));

  rv = euri_parse (VM_GLOBAL_MEMORY_POOL (vm), euri, eu);
  if (rv != APR_SUCCESS)
    return NULL;

  return load_env_core (EURI_HOSTNAME (eu), 
			EURI_PORT_STRING (eu) ? atoi (EURI_PORT_STRING (eu)) : APR_URI_ENV_DEFAULT_PORT,
			EURI_KEY (eu), EURI_PATH (eu));
}

/** ETYPE ڿ ݵ Le://e.mybox.com/env/core/Object('SHA1KEY');   ؾ Ѵ.  */
env_t
load_env_types (etype)
     char *etype;
{
  env_t e;
  char *tmpstr;
  char *ts, *te;

  if (!etype
      || (etype && !strlen (etype))
      || (etype && etype[0] != 'L'))
    return NULL;

  ts = etype + 1;
  te = etype + strlen (etype) - 1;

  if (te[0] != ';')
    return NULL;

  tmpstr = env_strndup (ts, te - ts);
  e = load_env (tmpstr);
  env_free (tmpstr);

  return e;
}

/**
   ̸ eclass  ã´.

   @param name	ε eclass ̸.  ´ "/env/core/string"   ̴.
*/
env_t
lookup_env (name)
     const char *name;
{
  env_t e;

  e = load_env ((char *) name);
  if (!e)
    return NULL;

  if (process_env (e, 1))
    return NULL;

  return e;
}


/** Hash Table  Ǵ ϼ ׽Ʈ Լν, ENTRY  
   Hash Table  ִ SLOT  ̰, X_ARG  vm_htab_find_slot_with_hash 
   ڷ  ̴.  */
static int
etype_hash_eq (entry, x_arg)
     const void *entry, *x_arg;
{
  etype_t *sa;
  char *sb;
  
  sa = (etype_t *) entry;
  sb = (char *) x_arg;

  if (ETYPE_SIGNITURE (sa) && sb
      && !strcmp (ETYPE_SIGNITURE (sa), sb))
    return 1;

  return 0;
}

/** Hash Table   Slot  ϵǾ ִ entry  delete   Ǵ Լ̴.
   vm_htab_create () Լ ȣ , Լ Ͱ  ϵȴ.  */
static void
etype_hash_del (entry)
     void *entry;
{
  etype_t *s = (etype_t *) entry;

  env_free (ETYPE_SIGNITURE (s));
  env_free (s);
}

/** ο ETYPE  ϴ.  */
static etype_t *
new_etype (kind, is_resolved, tp, size, arg)
     enum evalue_type kind;
     int is_resolved, size;
     const char *tp;
     void *arg;
{
  vm_t *vm = ESERVER_VM (eserver);
  etype_t **slot, *entry;
  vm_hashval_t hash;

  hash = vm_htab_hash_string (tp);
  
  slot = (etype_t **) vm_htab_find_slot_with_hash (VM_ETYPES_HASH (vm), tp, hash, INSERT);
  entry = *slot;
  
  if (entry)
    return entry;	/* ̹  */

  /* TODO: ETYPE  ޸  κ   ؾ Ѵ.  Malloc  ̿ 
      ̰, Memory Pool  ̿   ̴.  ݱ  memory pool 
     ̿ϸ ȵȴٴ ̴.  */
  entry = (etype_t *) env_malloc (sizeof (struct etype));
  ETYPE_KIND (entry) = kind;
  /* Ӱ  ŭ  resolve  ʾҴ.  */
  ETYPE_IS_RESOLVED (entry) = is_resolved;
  ETYPE_SIGNITURE (entry) = env_strdup (tp);
  ETYPE_SIZE (entry) = size;

  switch (kind)
    {
    case ETYPE_BYTE:
    case ETYPE_SHORT:
    case ETYPE_INT:
    case ETYPE_LONG:
    case ETYPE_FLOAT:
    case ETYPE_DOUBLE:
    case ETYPE_CHAR:
    case ETYPE_BOOLEAN:
    case ETYPE_VOID:
      /* ETYPE  primitive type   , ̰ Ѵ.  ̰  type 
	 primitive θ Ȯϱ ؼ Ѵ.  */
      ETYPE_IS_PRIMITIVE (entry) = 1;
      break;
    case ETYPE_OBJECTREF:
      ETYPE_U_ENV (entry) = (env_t) arg;
      break;
    case ETYPE_ARRAYREF:
      ETYPE_U_PTYPE (entry) = (etype_t *) arg;
      break;
    default:
      break;
    }

  *slot = entry;

  return entry;	/*  */
}

/** ETYPE  ã⸸ Ѵ.  */
etype_t *
lookup_etype (tp)
     char *tp;
{
  vm_t *vm = ESERVER_VM (eserver);
  etype_t **slot;
  vm_hashval_t hash;

  hash = vm_htab_hash_string (tp);

  slot = (etype_t **) vm_htab_find_slot_with_hash (VM_ETYPES_HASH (vm), tp, hash, NO_INSERT);

  if (slot)
    return *slot;

  return NULL;
}

/** ARRAY  ETYPE  ã ȯѴ.  */
etype_t *
lookup_etype_array (type)
     etype_t *type;
{
  char *at;
  etype_t *t;

  if (!type)
    return NULL;

  at = env_malloc (strlen (ETYPE_SIGNITURE (type)) + 4);

  switch (ETYPE_KIND (type))
    {
    case ETYPE_OBJECTREF:
      sprintf (at, "[L%s;", ETYPE_SIGNITURE (type));
      break;
    default:
      sprintf (at, "[%s", ETYPE_SIGNITURE (type));
      break;
    }

  t = new_etype (ETYPE_ARRAYREF, RESOLVED_ETYPE, at, 4, type);
  env_free (at);

  return t;
}

/**
   ڿ TP  type  ã ȯѴ.   L Ȥ [  , ã ȯϵ  
   ϰ ȴ.
*/
etype_t *
get_etype (tp)
     char *tp;
{
  etype_t *type;

  /* 迭  L type   ⺻ type  ̹ VM  εǸ鼭   Ǿ
      ̴.  */
  type = lookup_etype (tp);
  if (type)
    return type;

  /* .  */
  switch (tp[0])
    {
    case 'L':
      {
	char **strp = &tp;
	char *start, *end;
	char *L;
	
	start = *strp;
	for (end = start; *end != 0 && *end != ';'; end++)
	  ;
	*strp = end;
	if (*end == ';')
	  end++;
	else
	  return NULL;

	L = env_strndup (start, end - start);
	type = new_etype (ETYPE_OBJECTREF, RESOLVED_ETYPE, L, 4, load_env_types (L));
	env_free (L);

	return type;
      }
    case '[':
      return lookup_etype_array (get_etype (tp + 1));
    default:
      /* ⿡    .  */
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "get_etype: . tp = %s", tp);
      break;
    }

  return NULL;
}

etype_t *
get_etype_with_kind (kind)
     enum evalue_type kind;
{
  const char *tp;

  switch (kind)
    {
    case ETYPE_BYTE:	tp = "B"; break;
    case ETYPE_SHORT:   tp = "S"; break;
    case ETYPE_INT:     tp = "I"; break;
    case ETYPE_LONG:    tp = "J"; break;
    case ETYPE_FLOAT:   tp = "F"; break;
    case ETYPE_DOUBLE:  tp = "D"; break;
    case ETYPE_CHAR:    tp = "C"; break;
    case ETYPE_BOOLEAN: tp = "Z"; break;
      //case ETYPE_OBJECT:  tp = "L"; break;
      //case ETYPE_ARRAY:   tp = "["; break;
    case ETYPE_STRING:  tp = "T"; break;
    case ETYPE_KEY:	tp = "K"; break;
    case ETYPE_VOID:    tp = "V"; break;
    default:
      tp = NULL;
      break;
    }

  return get_etype ((char *) tp);
}

/**
   Environment Type  ʱȭѴ.  
*/
void
vm_init_etypes (vm)
     vm_t *vm;
{
  etype_t *rv;

  /* Ʒ ʱȭ ϴ hash table  new_etype () Լ Ǳ , ݵ
     ù° new_etype () Լ ȣǱ  ʱȭǾ Ѵ.  */
  VM_ETYPES_HASH (vm) = 
    vm_htab_create (ETYPES_HASH_INITIAL_SIZE, vm_htab_hash_string,
		    (vm_htab_eq) etype_hash_eq,
		    (vm_htab_del) etype_hash_del);

  /* Environment  󿡼  Ǵ type 鿡  .  */
  new_etype (ETYPE_BYTE,	RESOLVED_ETYPE, "B", etype_size[ETYPE_BYTE], NULL);
  new_etype (ETYPE_SHORT,	RESOLVED_ETYPE, "S", etype_size[ETYPE_SHORT], NULL);
  new_etype (ETYPE_INT,		RESOLVED_ETYPE, "I", etype_size[ETYPE_INT], NULL);
  new_etype (ETYPE_LONG,	RESOLVED_ETYPE, "J", etype_size[ETYPE_LONG], NULL);
  new_etype (ETYPE_FLOAT,	RESOLVED_ETYPE, "F", etype_size[ETYPE_FLOAT], NULL);
  new_etype (ETYPE_DOUBLE,	RESOLVED_ETYPE, "D", etype_size[ETYPE_DOUBLE], NULL);
  new_etype (ETYPE_CHAR,	RESOLVED_ETYPE, "C", etype_size[ETYPE_CHAR], NULL);
  new_etype (ETYPE_BOOLEAN,	RESOLVED_ETYPE, "Z", etype_size[ETYPE_BOOLEAN], NULL);
  /* L  [   type    , ش ̸  
        ⿡   ʴ´.  ̰ environment 
     ؼϸ鼭  Ҵȴ.  */
  new_etype (ETYPE_VOID,	RESOLVED_ETYPE, "V", etype_size[ETYPE_VOID], NULL);

  /* Eserver   ȿ ó   ϴ type .  */
  new_etype (ETYPE_STRING,	RESOLVED_ETYPE, "T", etype_size[ETYPE_STRING], NULL);
  new_etype (ETYPE_KEY,		RESOLVED_ETYPE, "K", etype_size[ETYPE_KEY], NULL);

  /* ⺻ Type  ʱȭ.
     ) Le://localhost/lang/env/String('n7J2znmJv5pCocpie/1eATzhXBY='); */
  rv = get_etype ((char *) ENV_OBJECT_TYPE);
  process_env (ETYPE_U_ENV (rv), 1);
  rv = get_etype ((char *) ENV_STRING_TYPE);
  process_env (ETYPE_U_ENV (rv), 1);
}

/* TODO: ϼؾ Ѵ.  */
void *
lookup_interfacemethod (obj, ifclass, idx)
     eobjectref_t *obj ATTRIBUTE_UNUSED;
     env_t ifclass ATTRIBUTE_UNUSED;
     int idx ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "lookup_interfacemethod: TODO");

  return NULL;
}

/* GDB  ̴.  Memory  dump Ѵ.  */
void
dump_memory (ptr, size)
     void *ptr;
     int size;
{
  int i;

  if (!ptr)
    {
      fprintf (stdout, "(null)");
      fflush (stdout);
      return;
    }

  for (i=0; i<size; i++)
    {
      char *p = ((char *) ptr) + i;

      fprintf (stdout, "0x%02x ", *p & 0xff);
    }
  fflush (stdout);
}

/**
   FILE stream FP  INDENT ũ⸸ŭ  Ѵ.
 */
void
print_indent (fp, indent)
     FILE *fp;
     int indent;
{
  int i;

  for (i=0; i<indent; i++)
    fprintf (fp, " ");
}

/**
   ArrayRef  ϴµ, ش array   type  TYPE  ; ϸ, ش
   ͷ AREF  Էµȴ.
 */
void
dump_arrayref (type, aref)
     char *type;
     earrayref_t *aref;
{
  printf ("(array ");

  if (!type || !aref)
    {
      printf ("null)");
      fflush (stdout);
      return;
    }

  if (EARRAYREF_MAGIC (aref) != 0xbdbdbdbd)
    {
      printf ("isn't valid)");
      fflush (stdout);
      return;
    }

  switch (type[0])
    {
    case 'C':
      {
	char *str = EARRAYREF_DATA (aref);
	int len = EARRAYREF_SIZE (aref);
	
	printf ("(char \"%.*s\") ", len, str);
	printf ("(length %d) ", len);
	break;
      }
    default:
      printf ("(unknown)");
      break;
    }

  printf (")");
  fflush (stdout);
}

/**
   GDB  ̴.  Type   evalue  dump Ѵ.  
*/
void
dump_evalue_core (type, v, indent)
     char *type;
     evalue_t *v;
     int indent;
{
  printf ("(evalue ");

  if (!v)
    {
      printf ("null)");
      fflush (stdout);
      return;
    }

  switch (type[0])
    {
    case '[':
      printf ("(array 0x%x ", (unsigned int) EVALUE_ADDRESS (v));
      dump_arrayref (type + 1, (earrayref_t *) EVALUE_ADDRESS (v));
      printf (")");
      break;
    case 'L':
      printf ("\n");
      dump_objectref_core ((eobjectref_t *) EVALUE_ADDRESS (v), indent + 8);
      break;
    case 'I':
      printf ("(int %d)", EVALUE_INT (v));
      break;
    case 'Z':
      printf ("(boolean %d)", EVALUE_BOOLEAN (v));
      break;
    default:
      printf ("(unknown)");
      break;
    }

  printf (")");
  fflush (stdout);
}

/* GDB  ̴.  ObjectRef   dump Ѵ.  */
void
dump_objectref_core (obj, indent)
     eobjectref_t *obj;
     int indent;
{
  env_t e;
  efield_t *ifs, *fld;
  int nifs, sfc, index = 0;

  if (!obj)
    {
      fprintf (stdout, "(obj null)");
      fflush (stdout);
      return;
    }

  if (EOBJECTREF_MAGIC (obj) != 0xacacacac)
    {
      fprintf (stdout, "(obj isn't valid)");
      fflush (stdout);
      return;
    }    
  
  e = EOBJECTREF_ENV (obj);
  if (!e)
    {
      fprintf (stdout, "(env null)");
      fflush (stdout);
      return;
    }

  ifs = &ENV_INSTANCE_FIELDS (e);
  nifs = ENV_INSTANCE_FIELDS_COUNT (e);

  if (!nifs)
    return;

  print_indent (stdout, indent);
  fprintf (stdout, "(obj (env_name \"%s\")\n", STRCONST_PTR (ENV_NAME (e)));

  fld = ENV_STATIC_FIELDS (e);
  sfc = ENV_STATIC_FIELDS_COUNT (e);

  if (ENV_STATIC_FIELDS_COUNT (e) > 0)
    {
      print_indent (stdout, indent + 4);
      fprintf (stdout, "(static_fields [\n");
    }
  for (; --sfc >= 0; fld++) 
    {
      if (!(FIELD_ACCESS_FLAGS (fld) & AF_STATIC))
	{
	  fprintf (stdout, "(non-static error)");
	  return;
	}

      print_indent (stdout, indent + 8);
      fprintf (stdout, "(field \"%s\" \"%s\" ", 
	       STRCONST_PTR (FIELD_NAME (fld)),
	       ETYPE_SIGNITURE (FIELD_TYPE (fld)));

      dump_evalue_core (ETYPE_SIGNITURE (FIELD_TYPE (fld)), 
			(ENV_STATIC_FIELDS_DATA (e) + index),
			indent + 4);
      fprintf (stdout, ")\n");
    }
  if (ENV_STATIC_FIELDS_COUNT (e) > 0)
    {
      print_indent (stdout, indent + 4);
      fprintf (stdout, "])");
    }

  if (ENV_INSTANCE_FIELDS_COUNT (e) > 0
      && ENV_STATIC_FIELDS_COUNT (e) > 0)
    fprintf (stdout, "\n");

  if (ENV_INSTANCE_FIELDS_COUNT (e) > 0)
    {
      print_indent (stdout, indent + 4);
      fprintf (stdout, "(non_static_fields [\n");
    }
  for (; --nifs >= 0; ifs++) 
    {
      void *ret;

      print_indent (stdout, indent + 8);
      fprintf (stdout, "(field \"%s\" \"%s\" ",
	       STRCONST_PTR (FIELD_NAME (ifs)),
	       ETYPE_SIGNITURE (FIELD_TYPE (ifs)));

      ret = (void *) (((unsigned char *) (obj)) + (off_t) FIELD_OBJECT_DATA_OFFSET (ifs));

      dump_evalue_core (ETYPE_SIGNITURE (FIELD_TYPE (ifs)), (evalue_t *) ret, indent + 4);
      fprintf (stdout, ")\n");
    }
  if (ENV_INSTANCE_FIELDS_COUNT (e) > 0)
    {
      print_indent (stdout, indent + 4);
      fprintf (stdout, "])");
    }
  fprintf (stdout, ")");
  fflush (stdout);
}

/** 
    GDB  Լ.  ObjectRef  OBJ   Ѵ.
*/
void
dump_objectref (obj)
     eobjectref_t *obj;
{
  dump_objectref_core (obj, 0);
  fprintf (stdout, "\n");
  fflush (stdout);
}


/* ============================================================================

   VM  software  Լ 

   ============================================================================ */


/**
   soft_athrow.

   TODO:  ϼ ʾҴ!!!!  ڵ Ȯ 𸣰ڱ.  
*/
void
soft_athrow (o)
     eobjectref_t *o;
{
  if (!o)
    throw_NullPointerException ();
  else
    throw_exception (o);
}

/**
   soft_checkcast.

   TODO:  ϼ ʾҴ!!!!  ڵ Ȯ 𸣰ڱ.  
*/
void *
soft_checkcast (e, o)
     env_t e ATTRIBUTE_UNUSED;
     eobjectref_t *o ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "soft_checkcast: TODO");

  return NULL;
}

int
soft_instanceof (e, o)
     env_t e ATTRIBUTE_UNUSED;
     eobjectref_t *o ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "soft_instanceof: TODO");

  return -1;
}

void *
soft_multianewarray (class, dims, args)
     env_t class ATTRIBUTE_UNUSED;
     int dims ATTRIBUTE_UNUSED;
     evalue_t *args ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "soft_multianewarray: TODO");

  return NULL;
}


struct stack_trace_info * build_stack_trace (void);
int find_method_exception_block (uintp, env_t, emethod_t *, uintp *);
void print_stack_trace (eobjectref_t *);
void unhandled_exception (eobjectref_t *);
void dispatch_exception (eobjectref_t *, struct stack_trace_info *);

/*  request  ׿ִ exception   struct stack_trace_info ü 
 */
struct stack_trace_info *
build_stack_trace (void)
{
  int count = 0;
  erequest_t *er;
  struct stack_trace_info *info;
  struct VMexception *v;

  er = FETCH_EREQUEST;
  if (!er)
    return NULL;

  v = (struct VMexception *) REQ_DATA_EXCEPTION (REQUEST_DATA (er));

  while (v)
    {
      if (VMEXCEPTION_METHOD (v) != (emethod_t  *) 1)
	count++;

      v = VMEXCEPTION_PREV (v);
    }

  /* Build an array of stack_trace_info ü. 
     TODO: Ʒ ޸ Ҵ garbage collection    ֵ ؾ Ѵ.  */
  info = env_malloc (sizeof (struct stack_trace_info) * (count + 1));
  if (!info)
    return NULL;

  count = 0;
  v = (struct VMexception *) REQ_DATA_EXCEPTION (REQUEST_DATA (er));
  for (; v; v = VMEXCEPTION_PREV (v))
    {
      if (VMEXCEPTION_METHOD (v) == (emethod_t  *) 1)
	continue;

      info[count].pc = VMEXCEPTION_PC (v);
      info[count].m = VMEXCEPTION_METHOD (v);
      count++;
    }
  info[count].pc = 0;
  info[count].m = (emethod_t *) -1;

  return info;
}

/*
 * Look for exception block in method.
 * Returns true if there is an exception handler, false otherwise.
 */
int
find_method_exception_block (pc, e, m, handler)
     uintp pc;
     env_t e;
     emethod_t *m;
     uintp *handler;
{
  uint32 i;
  env_t eptr;

  if (!e || !m)
    return -1;

  /* Right method - look for exception */
  if (!METHOD_EXCEPTION_TABLE (m))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR,
		     "%s %s %s has no handlers.",
		     STRCONST_PTR (ENV_NAME (METHOD_ENV (m))),
		     STRCONST_PTR (METHOD_NAME (m)),
		     STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (m))));
      return -1;
    }

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "%s.%s has %d handlers (throw was pc=%#x):",
		 STRCONST_PTR (ENV_NAME (METHOD_ENV (m))),
		 STRCONST_PTR (METHOD_NAME (m)),
		 MEXCEPTION_LENGTH (METHOD_EXCEPTION_TABLE (m)),
		 pc);

  for (i=0; i<MEXCEPTION_LENGTH (METHOD_EXCEPTION_TABLE (m)); i++)
    {
      uintp start_pc = MEXCEPTION_START_PC (METHOD_EXCEPTION_TABLE (m), i);
      uintp end_pc = MEXCEPTION_END_PC (METHOD_EXCEPTION_TABLE (m), i);
      uintp handler_pc = MEXCEPTION_HANDLER_PC (METHOD_EXCEPTION_TABLE (m), i);

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "  Handler %d covers %#lx-%#lx", (long) start_pc, (long) end_pc);

      if (pc < start_pc || pc >= end_pc)
	continue;

      /* Found exception - is it right type */
      if (!MEXCEPTION_CATCH_IDX (METHOD_EXCEPTION_TABLE (m), i))
	{
	  *handler = handler_pc;

	  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
			 "  Found handler @ %#lx: catches all exceptions.",
			 (long) handler_pc);
	  return 0;
	}

      /* Did I try to resolve that catch type before */
      if (MEXCEPTION_CATCH_TYPE (METHOD_EXCEPTION_TABLE (m), i) == (env_t) -1)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR,
			 "  Found handler @ %#lx: Unresolvable catch type.", 
			 (long) handler_pc);
	  return -1;
	}

      /* Resolve catch class if necessary */
      if (!MEXCEPTION_CATCH_TYPE (METHOD_EXCEPTION_TABLE (m), i))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR,
			 "find_method_exception_block: TODO");
	  return -1;
	}
      for (eptr = e; eptr != 0; eptr = ENV_SUPER (eptr)) 
	{
	  if (eptr == MEXCEPTION_CATCH_TYPE (METHOD_EXCEPTION_TABLE (m), i)) 
	    {
	      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
			     "  Found matching handler at %#lx: Handles %s.",
			     (long) handler_pc,
			     STRCONST_PTR (ENV_NAME (MEXCEPTION_CATCH_TYPE (METHOD_EXCEPTION_TABLE (m), i))));
	      *handler = handler_pc;
	      return 0;
	    }
	}

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "  Handler at %#lx (handles %s), does not match.",
		     (long) handler_pc, 
		     STRCONST_PTR (ENV_NAME (MEXCEPTION_CATCH_TYPE (METHOD_EXCEPTION_TABLE (m), i))));
    }

  return -1;
}

void
print_stack_trace (obj)
     eobjectref_t *obj ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "print_stack_trace: TODO");
}

/**
  ObjectRef  Ű exception  ܰ迡 ó    ,  Լ
  ȣȴ.
*/
void
unhandled_exception (obj)
     eobjectref_t *obj;
{
  erequest_t *er;

  er = FETCH_EREQUEST;

  REQ_DATA_EXCEPTION (REQUEST_DATA (er)) = NULL;

  /* We don't know what to do here. */
  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "Internal error: caught an unexpected exception. Please check your CLASSPATH and your installation.");

  /* Display the exception and stack trace to help in debugging.  */
  print_stack_trace (obj);

  ENV_ABORT ();
}

void
dispatch_exception (obj, info)
     eobjectref_t *obj;
     struct stack_trace_info *info ATTRIBUTE_UNUSED;
{
  int rv;
  erequest_t *er;
  struct VMexception *v;
  uintp handler;

  er = FETCH_EREQUEST;
  
  /* Search down exception stack for a match */
  env_log_error (ENVLOG_MARK, ENVLOG_INFO, 
		 "dispatch_exception: %s",
		 STRCONST_PTR (ENV_NAME (EOBJECTREF_ENV (obj))));

  v = (struct VMexception *) REQ_DATA_EXCEPTION (REQUEST_DATA (er));
  for (; v; v = VMEXCEPTION_PREV (v))
    {
      eobjectref_t *o;

      if (VMEXCEPTION_METHOD (v) == (emethod_t  *) 1)
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "dispatch_exception: TODO 1");
	}

      rv = find_method_exception_block (VMEXCEPTION_PC (v), 
					EOBJECTREF_ENV (obj),
					VMEXCEPTION_METHOD (v),
					&handler);
      if (!VMEXCEPTION_METHOD (v) 
	  || (METHOD_ACCESS_FLAGS (VMEXCEPTION_METHOD (v)) & AF_SYNCHRONIZED) == 0)
	o = NULL;
      else if (METHOD_ACCESS_FLAGS (VMEXCEPTION_METHOD (v)) & AF_STATIC)
	env_log_error (ENVLOG_MARK, ENVLOG_ERR, "dispatch_exception: TODO 2");
      else
	o = VMEXCEPTION_OBJECTREF (v);

      /* If handler found, call it */
      if (!rv)
	{
	  VMEXCEPTION_PC (v) = handler;
	  longjmp (VMEXCEPTION_JMP_BUF (v), 1);
	  /* don't return.  */
	}

      /* If not here, exit monitor if synchronised. */
      if (o && METHOD_ACCESS_FLAGS (VMEXCEPTION_METHOD (v)) & AF_SYNCHRONIZED)
	env_log_error (ENVLOG_MARK, ENVLOG_ERR, "dispatch_exception: TODO 3");
    }

  unhandled_exception (obj);
  /* Should not return */  
}

/**
   Create and throw an exception resulting from an error during VM processing.
   
   longjmp  ϱ , jump   return    ̴.
*/
void
throw_error (kind)
     enum exception_kind kind ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: throw_error");

  return;
}

/**
   Throw an exception with backtrace recomputed.
 
   Semantic: take stacktrace right now (overwrite whatever stacktrace
   is in the exception object) and dispatch.

   @param obj	env.core.Throwable ObjectRef Ÿ
*/
void
throw_exception (obj)
     eobjectref_t *obj;
{
  if (!obj)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Exception thrown on null object ... aborting");
      ENV_ABORT ();
      /* don't return  */
    }

  dispatch_exception (obj, build_stack_trace ());
}

/**
   env.core.NullPointerException  ~ ! ޾ƶ!
*/
void
throw_NullPointerException (void)
{
  throw_exception (call_constructor ("env.core.NullPointerException", NULL, "()V"));
}

inline int
instanceof_array (c, oc)
     env_t c ATTRIBUTE_UNUSED, oc ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: instanceof_array");

  return 0;
}

inline int
instanceof_interface (c, oc)
     env_t c ATTRIBUTE_UNUSED, oc ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: instanceof_interface");
  
  return 0;
}

inline int
instanceof_class (c, oc)
     env_t c, oc;
{
  env_t tc;

  /* Check for superclass matches */
  for (tc = ENV_SUPER (oc); tc != 0; tc = ENV_SUPER (tc)) 
    if (c == tc) 
      return 1;

  return 0;
}

int
instanceof (c, oc)
     env_t c, oc;
{
  /* Handle the simplest case first - they are the same */
  if (c == oc)
    return 1;
  /* Else if an array check that */
  else if (ENV_IS_ARRAY (c))
    return instanceof_array (c, oc);
  /* Else if an interface check that */
  else if (ENV_IS_INTERFACE (c))
    return instanceof_interface (c, oc);

  /* Else is must to a class */
  return instanceof_class (c, oc);
}

/* eserver  daemon ̱    ߻Ͽ process  ߸ ȵȴ.  
   ׷ abort  env_abort ()   üѴ.  */
void
env_abort (void)
{
  env_log_error (ENVLOG_MARK, ENVLOG_CRIT, "TODO: env_abort");
}


/*===========================================================================
  Native Call  ƾ.
  ===========================================================================*/

static apr_thread_mutex_t *native_lock;

int open_nativelib (apr_pool_t *, struct _lib_handler *, char *);
int load_nativelib (apr_pool_t *, char *);
void *load_nativelib_symbol (char *);

int
open_nativelib (p, lib, path)
     apr_pool_t *p;
     struct _lib_handler *lib;
     char *path;
{
  /*  Ʒ ׽Ʈ  ϴٸ, libltdl  system-dependent 
     library ̸ ã   ̴.  */
  lib->desc = nativelib_load (path);
  if (!lib->desc)
    return -1;

  lib->name = apr_pstrdup (p, path);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, 
		 "Opending native lib %s\n\tLOAD desc=0x%x", lib->name, lib->desc);

  return 0;
}

/* PATH  Ǿ ִ library  load Ѵ.  */
int
load_nativelib (p, path)
     apr_pool_t *p;
     char *path;
{
  apr_status_t rv;
  struct _lib_handler *node, *root;

  rv = apr_thread_mutex_lock (native_lock);

  root = VM_NATIVE_LIBS (ESERVER_VM (eserver));
  node = apr_pcalloc (p, sizeof (struct _lib_handler));

  while (root)
    {
      if (!strcmp (root->name, path) && root->desc)
	return 0;

      root = root->next;
    }

  if (open_nativelib (p, node, path))
    {
      rv = apr_thread_mutex_unlock (native_lock);
      return -1;
    }

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, 
		 "Native lib %s LOAD desc=0x%x", node->name, node->desc);

  node->next = VM_NATIVE_LIBS (ESERVER_VM (eserver));
  VM_NATIVE_LIBS (ESERVER_VM (eserver)) = node;

  rv = apr_thread_mutex_unlock (native_lock);
  return 0;
}

void *
load_nativelib_symbol (funcname)
     char *funcname;
{
  apr_status_t rv;
  void *func = NULL;
  struct _lib_handler *node = VM_NATIVE_LIBS (ESERVER_VM (eserver));

  rv = apr_thread_mutex_lock (native_lock);

  while (!func && node) 
    {
      func = nativelib_getsymbol (node->desc, funcname);

      if (!func) 
	env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
		       "Couldn't find %s in library handle %s.\nError message is %s.",
		       funcname, node->name, nativelib_geterror ());
      else
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_INFO, 
			 "Found \"%s\" in library handle %s.", funcname, node->name);
	  break;
	}

      node = node->next;
    }

  rv = apr_thread_mutex_unlock (native_lock);

  return func;
}

/**
   Լ ̸ ִ ̸ Ѵ.   C 󿡼 Լ ̸ 256  ѵǴ 
   ˰ ,   Ȯ д.  
*/
#define	MAXSTUBLEN	1024

void *
call_native (m)
     emethod_t *m;
{
  char *fullname, stub[MAXSTUBLEN];
  void *func;

  /* ǥ ̸  Ʒ  ̴.

       ȯ)
         Environment Name + '/' + Method Name

       ȯ)
          '/' ڴ '_' ڷ ġȯȴ.

       Environment Name  env/core/Runtime ̰ Method Name  _load  ,
     ȯ fullname  env_core_Runtime__load  ̴.  */
  snprintf (stub, MAXSTUBLEN, "%s/%s", 
	    STRCONST_PTR (ENV_NAME (METHOD_ENV (m))), STRCONST_PTR (METHOD_NAME (m)));

  fullname = env_subst (stub, strlen (stub), "", '/', '_', "");

  func = load_nativelib_symbol (fullname);
  if (!func)
    env_log_error (ENVLOG_MARK, ENVLOG_ERR, "Failed to locate native function: %s", fullname);

  env_free (fullname);
  
  return func;
}

/* ڰ DLL  .so  load ϴ ڵ带 ϱ , ڵ library  ã
   ʿ .  */
void
init_native (s, pool)
     eserver_t *s;
     apr_pool_t *pool;
{
  char *p, *np;
  char lib[MAXLIBPATH];
  apr_status_t rv;
  const char *libpath;
  unsigned int len;
  eoption_t *op = ESERVER_OPTION (s);

  nativelib_init ();

  rv = apr_thread_mutex_create (&native_lock, APR_THREAD_MUTEX_DEFAULT, pool);
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_EMERG, "Couldn't create gcmon lock");
      return;
    }
  
  /* ȯ  E_LIBPATH  д´.  */
  libpath = getenv ("E_LIBPATH");
  if (!libpath && !OPTION_LIBPATH (op))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_WARNING, 
		     "init_native: ȯ溯 E_LIBPATH  ǵǾ  ʽϴ.");
      return;
    }

  OPTION_LIBPATH (op) = apr_pstrcat (pool, OPTION_LIBPATH (op), PATH_SEPARATOR, libpath, NULL);
  len = strlen (OPTION_LIBPATH (op));

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "init_native: ȯ溯 E_LIBPATH  = \"%s\"", OPTION_LIBPATH (op));
  
  for (p = (char *) OPTION_LIBPATH (op); p; p = np)
    {
      np = strstr (p, PATH_SEPARATOR);
      if (!np)
	strcpy (lib, p);
      else if (p == np)
	{
	  np += strlen (PATH_SEPARATOR);
	  continue;
	}
      else
	{
	  strncpy (lib, p, (size_t) (np - p));
	  lib[np - p] = '\0';
	  np += strlen (PATH_SEPARATOR);
	}
      /* should be file_separator, libltdl does not handle backslashes yet */
      strcat (lib, "/");
      strcat (lib, NATIVELIBRARY);
      strcat (lib, LIBRARYSUFFIX);

      env_log_error (ENVLOG_MARK, ENVLOG_INFO, "trying to load %s", lib);

      if (!load_nativelib (pool, lib))
	{
	  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "%s ε ", lib);
	  return;
	}
    }

  env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
		 "Failed to locate native library \"%s\" in path: %s", NATIVELIBRARY, OPTION_LIBPATH (op));
}


/**
   ü vm_t  ʱȭѴ.  
*/
int
vm_init_global_struct (s, p)
     eserver_t *s;
     apr_pool_t *p;
{
  apr_status_t rv;
  vm_t *vm;

  srand (time (NULL));

  /* apr_pcalloc   ҴǴ ޸𸮸 \0  Ѵ.  */
  vm = ESERVER_VM (s) = (vm_t *) apr_pcalloc (p, sizeof (struct vm));

  VM_CURRENT_UNIQUE_ID (vm) = 0;
  VM_GLOBAL_MEMORY_POOL (vm) = p;

  /* VM   garbage collector  ʱȭѴ.   boehm gc  ϰ
     ִ.  */
  VM_GC (vm) = initialize_gc (p);
  EGC_init (VM_GC (vm));
  /* finalize   process (finalize_daemon)  garbage collection (gc_daemon)
       process  Ѵ.  */
  EGC_enable (VM_GC (vm), p);

  VM_E_TREE (vm) = 
    vm_splay_tree_new ((vm_splay_tree_compare_fn) vm_e_tree_compare,
		       (vm_splay_tree_delete_key_fn) NULL,
		       (vm_splay_tree_delete_value_fn) NULL);

  VM_EOBJECTREF_TREE (vm) = 
    vm_splay_tree_new ((vm_splay_tree_compare_fn) vm_eref_tree_compare,
		       (vm_splay_tree_delete_key_fn) NULL,
		       (vm_splay_tree_delete_value_fn) NULL);

  VM_EARRAYREF_TREE (vm) = 
    vm_splay_tree_new ((vm_splay_tree_compare_fn) vm_eref_tree_compare,
		       (vm_splay_tree_delete_key_fn) NULL,
		       (vm_splay_tree_delete_value_fn) NULL);
    

  rv = apr_thread_mutex_create (&VM_GLOBAL_MUTEX (vm), APR_THREAD_MUTEX_DEFAULT, VM_GLOBAL_MEMORY_POOL (vm));
  if (rv != APR_SUCCESS) 
    {
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, ap_server_conf,
		    "Couldn't create vm global lock");
      return -1;
    }

  /* Native Interface   ʱ  Ѵ.  */
  init_native (s, vmglobal);

  /* Primitve Type  ʱȭѴ.  */
  vm_init_etypes (vm);

  return 0;
}

/**
   VM MPM (Multi-Process Model)  Signal  ޾ , ȣ Ǿ Ѵ.  
   
   TODO: shutdown ô  Լ ȣǾ ϰ, restart   Լ ȣǾ
   ϴ°?
*/
int
vm_finish_global_struct (void)
{
  apr_status_t rv;

#if VM_THREAD_SAFE_SPLAY_TREE
  vm_finish_splaytree ();
#endif

  vm_splay_tree_delete (VM_E_TREE (ESERVER_VM (eserver)));

  rv = apr_thread_mutex_destroy (VM_GLOBAL_MUTEX (ESERVER_VM (eserver)));

  if (rv != APR_SUCCESS) 
    {
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, ap_server_conf,
		    "Couldn't destroy vm global lock");
      return -1;
    }

  return 0;
}

void
vm_pre_config (void)
{
  static int vm_restart_num = 0;

  /* sigh, want this only the second time around */
  if (vm_restart_num++ == 1) 
    {
      /* TODO:  ⼭    ϰ  ʴ.  */
      vm_parent_pid = vm_my_pid = getpid ();
    }
}

/**
   VM ʱȭ.  
*/
void
vm_init_common (s, p)
     eserver_t *s;
     apr_pool_t *p;
{
  vmconf = p;

  /* VM  Ǵ  ޸  Ҵϰ  ޸ pool  ʱȭѴ.  */
  apr_pool_create (&vmglobal, vmconf);
  apr_pool_tag (vmglobal, "vmglobal");

#if VM_THREAD_SAFE_SPLAY_TREE
  /* Environment ü ϰ Ǵ splay tree  thread-safe ϰ ϱ  
     ʱȭ ǽѴ.  */
  vm_init_splaytree (vmglobal);
#endif

  vm_init_global_struct (s, vmglobal);
}
