/* A splay-tree datatype.  
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

/* Splay-tree 鿡  б  å  մϴ:

     Lewis, Harry R. and Denenberg, Larry.  Data Structures and Their
     Algorithms.  Harper-Collins, Inc.  1991.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"

#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor() */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

#if VM_THREAD_SAFE_SPLAY_TREE
/*
  TODO:  Splay TREE  thread-safe ϰ Ǿ ִ.   mutex  on/off
  ġ ȭǾ  ʴ.
 */

/* Thread safe   mutex.  */
static apr_thread_mutex_t *splaytree_mutex;

static void splaytree_mutex_on		PARAMS((void));
static void splaytree_mutex_off		PARAMS((void));
#endif

static void * splay_tree_xmalloc_allocate PARAMS ((int, void *));
static void splay_tree_xmalloc_deallocate PARAMS ((void *, void *));

static void splay_tree_delete_helper    PARAMS((vm_splay_tree_t, 
						vm_splay_tree_node_t));
static void splay_tree_splay            PARAMS((vm_splay_tree_t,
						vm_splay_tree_key));
static vm_splay_tree_node_t splay_tree_splay_helper     
                                        PARAMS((vm_splay_tree_t,
						vm_splay_tree_key,
						vm_splay_tree_node_t*,
						vm_splay_tree_node_t*,
						vm_splay_tree_node_t*));
static int splay_tree_foreach_helper    PARAMS((vm_splay_tree_t,
					        vm_splay_tree_node_t,
						vm_splay_tree_foreach_fn,
						void*));

/* Deallocate NODE (a member of SP), and all its sub-trees.  */

static void 
splay_tree_delete_helper (sp, node)
     vm_splay_tree_t sp;
     vm_splay_tree_node_t node;
{
  if (!node)
    return;

  splay_tree_delete_helper (sp, node->left);
  splay_tree_delete_helper (sp, node->right);

  if (sp->delete_key)
    (*sp->delete_key)(node->key);
  if (sp->delete_value)
    (*sp->delete_value)(node->value);

  (*sp->deallocate) ((char*) node, sp->allocate_data);
}

/* Help splay SP around KEY.  PARENT and GRANDPARENT are the parent
   and grandparent, respectively, of NODE.  */

static vm_splay_tree_node_t
splay_tree_splay_helper (sp, key, node, parent, grandparent)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
     vm_splay_tree_node_t *node;
     vm_splay_tree_node_t *parent;
     vm_splay_tree_node_t *grandparent;
{
  vm_splay_tree_node_t *next;
  vm_splay_tree_node_t n;
  int comparison;
  
  n = *node;

  if (!n)
    return *parent;

  comparison = (*sp->comp) (key, n->key);

  if (comparison == 0)
    /* We've found the target.  */
    next = 0;
  else if (comparison < 0)
    /* The target is to the left.  */
    next = &n->left;
  else 
    /* The target is to the right.  */
    next = &n->right;

  if (next)
    {
      /* Continue down the tree.  */
      n = splay_tree_splay_helper (sp, key, next, node, parent);

      /* The recursive call will change the place to which NODE
	 points.  */
      if (*node != n)
	return n;
    }

  if (!parent)
    /* NODE is the root.  We are done.  */
    return n;

  /* First, handle the case where there is no grandparent (i.e.,
     *PARENT is the root of the tree.)  */
  if (!grandparent) 
    {
      if (n == (*parent)->left)
	{
	  *node = n->right;
	  n->right = *parent;
	}
      else
	{
	  *node = n->left;
	  n->left = *parent;
	}
      *parent = n;
      return n;
    }

  /* Next handle the cases where both N and *PARENT are left children,
     or where both are right children.  */
  if (n == (*parent)->left && *parent == (*grandparent)->left)
    {
      vm_splay_tree_node_t p = *parent;

      (*grandparent)->left = p->right;
      p->right = *grandparent;
      p->left = n->right;
      n->right = p;
      *grandparent = n;
      return n; 
    }
  else if  (n == (*parent)->right && *parent == (*grandparent)->right)
    {
      vm_splay_tree_node_t p = *parent;

      (*grandparent)->right = p->left;
      p->left = *grandparent;
      p->right = n->left;
      n->left = p;
      *grandparent = n;
      return n;
    }

  /* Finally, deal with the case where N is a left child, but *PARENT
     is a right child, or vice versa.  */
  if (n == (*parent)->left) 
    {
      (*parent)->left = n->right;
      n->right = *parent;
      (*grandparent)->right = n->left;
      n->left = *grandparent;
      *grandparent = n;
      return n;
    } 
  else
    {
      (*parent)->right = n->left;
      n->left = *parent;
      (*grandparent)->left = n->right;
      n->right = *grandparent;
      *grandparent = n;
      return n;
    }
}

/* Splay SP around KEY.  */

static void
splay_tree_splay (sp, key)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
{
  if (sp->root == 0)
    return;

  splay_tree_splay_helper (sp, key, &sp->root, 
			   /*grandparent=*/0, /*parent=*/0); 
}

/* Call FN, passing it the DATA, for every node below NODE, all of
   which are from SP, following an in-order traversal.  If FN every
   returns a non-zero value, the iteration ceases immediately, and the
   value is returned.  Otherwise, this function returns 0.  */

static int
splay_tree_foreach_helper (sp, node, fn, data)
     vm_splay_tree_t sp;
     vm_splay_tree_node_t node;
     vm_splay_tree_foreach_fn fn;
     void* data;
{
  int val;

  if (!node)
    return 0;

  val = splay_tree_foreach_helper (sp, node->left, fn, data);
  if (val)
    return val;

  val = (*fn)(node, data);
  if (val)
    return val;

  return splay_tree_foreach_helper (sp, node->right, fn, data);
}

/* An allocator and deallocator based on xmalloc.  */
static void *
splay_tree_xmalloc_allocate (size, data)
     int size;
     void *data ATTRIBUTE_UNUSED;
{
  return (void *) env_malloc (size);
}

static void
splay_tree_xmalloc_deallocate (object, data)
     void *object;
     void *data ATTRIBUTE_UNUSED;
{
  env_free (object);
}

/* Allocate a new splay tree, using COMPARE_FN to compare nodes,
   DELETE_KEY_FN to deallocate keys, and DELETE_VALUE_FN to deallocate
   values.  Use xmalloc to allocate the splay tree structure, and any
   nodes added.  */

vm_splay_tree_t
vm_splay_tree_new (compare_fn, delete_key_fn, delete_value_fn)
     vm_splay_tree_compare_fn compare_fn;
     vm_splay_tree_delete_key_fn delete_key_fn;
     vm_splay_tree_delete_value_fn delete_value_fn;
{
  return (vm_splay_tree_new_with_allocator
          (compare_fn, delete_key_fn, delete_value_fn,
           splay_tree_xmalloc_allocate, splay_tree_xmalloc_deallocate, 0));
}

/* Allocate a new splay tree, using COMPARE_FN to compare nodes,
   DELETE_KEY_FN to deallocate keys, and DELETE_VALUE_FN to deallocate
   values.  */

vm_splay_tree_t
vm_splay_tree_new_with_allocator (compare_fn, delete_key_fn, delete_value_fn,
				  allocate_fn, deallocate_fn, allocate_data)
     vm_splay_tree_compare_fn compare_fn;
     vm_splay_tree_delete_key_fn delete_key_fn;
     vm_splay_tree_delete_value_fn delete_value_fn;
     vm_splay_tree_allocate_fn allocate_fn;
     vm_splay_tree_deallocate_fn deallocate_fn;
     void *allocate_data;
{
  vm_splay_tree_t sp = (vm_splay_tree_t) (*allocate_fn) (sizeof (struct vm_splay_tree_s),
							 allocate_data);
  sp->root = 0;
  sp->comp = compare_fn;
  sp->delete_key = delete_key_fn;
  sp->delete_value = delete_value_fn;
  sp->allocate = allocate_fn;
  sp->deallocate = deallocate_fn;
  sp->allocate_data = allocate_data;

  return sp;
}

/* Deallocate SP.  */

void 
vm_splay_tree_delete (sp)
     vm_splay_tree_t sp;
{
  splay_tree_delete_helper (sp, sp->root);
  (*sp->deallocate) ((char*) sp, sp->allocate_data);
}

/* Insert a new node (associating KEY with DATA) into SP.  If a
   previous node with the indicated KEY exists, its data is replaced
   with the new value.  Returns the new node.  */

vm_splay_tree_node_t
vm_splay_tree_insert (sp, key, value)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
     vm_splay_tree_value value;
{
  int comparison = 0;

#if VM_THREAD_SAFE_SPLAY_TREE
  splaytree_mutex_on ();
#endif

  splay_tree_splay (sp, key);

  if (sp->root)
    comparison = (*sp->comp)(sp->root->key, key);

  if (sp->root && comparison == 0)
    {
      /* If the root of the tree already has the indicated KEY, just
	 replace the value with VALUE.  */
      if (sp->delete_value)
	(*sp->delete_value)(sp->root->value);
      sp->root->value = value;
    } 
  else 
    {
      /* Create a new node, and insert it at the root.  */
      vm_splay_tree_node_t node;
      
      node = ((vm_splay_tree_node_t)
              (*sp->allocate) (sizeof (struct vm_splay_tree_node_s),
                               sp->allocate_data));
      node->key = key;
      node->value = value;
      
      if (!sp->root)
	node->left = node->right = 0;
      else if (comparison < 0)
	{
	  node->left = sp->root;
	  node->right = node->left->right;
	  node->left->right = 0;
	}
      else
	{
	  node->right = sp->root;
	  node->left = node->right->left;
	  node->right->left = 0;
	}

      sp->root = node;
    }

#if VM_THREAD_SAFE_SPLAY_TREE
  splaytree_mutex_off ();
#endif

  return sp->root;
}

/* Remove KEY from SP.  It is not an error if it did not exist.  */

void
vm_splay_tree_remove (sp, key)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
{
  splay_tree_splay (sp, key);

  if (sp->root && (*sp->comp) (sp->root->key, key) == 0)
    {
      vm_splay_tree_node_t left, right;

      left = sp->root->left;
      right = sp->root->right;

      /* Delete the root node itself.  */
      if (sp->delete_value)
	(*sp->delete_value) (sp->root->value);
      (*sp->deallocate) (sp->root, sp->allocate_data);

      /* One of the children is now the root.  Doesn't matter much
	 which, so long as we preserve the properties of the tree.  */
      if (left)
	{
	  sp->root = left;

	  /* If there was a right child as well, hang it off the 
	     right-most leaf of the left child.  */
	  if (right)
	    {
	      while (left->right)
		left = left->right;
	      left->right = right;
	    }
	}
      else
	sp->root = right;
    }
}

/* Lookup KEY in SP, returning VALUE if present, and NULL 
   otherwise.  */

vm_splay_tree_node_t
vm_splay_tree_lookup (sp, key)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
{
#if VM_THREAD_SAFE_SPLAY_TREE
  splaytree_mutex_on ();
#endif

  splay_tree_splay (sp, key);

  if (sp->root && (*sp->comp)(sp->root->key, key) == 0)
    {
#if VM_THREAD_SAFE_SPLAY_TREE
      splaytree_mutex_off ();
#endif
      return sp->root;
    }
  else
    {
#if VM_THREAD_SAFE_SPLAY_TREE
  splaytree_mutex_off ();
#endif
      return 0;
    }
}

/* Return the node in SP with the greatest key.  */

vm_splay_tree_node_t
vm_splay_tree_max (sp)
     vm_splay_tree_t sp;
{
  vm_splay_tree_node_t n = sp->root;

  if (!n)
    return NULL;

  while (n->right)
    n = n->right;

  return n;
}

/* Return the node in SP with the smallest key.  */

vm_splay_tree_node_t
vm_splay_tree_min (sp)
     vm_splay_tree_t sp;
{
  vm_splay_tree_node_t n = sp->root;

  if (!n)
    return NULL;

  while (n->left)
    n = n->left;

  return n;
}

/* Return the immediate predecessor KEY, or NULL if there is no
   predecessor.  KEY need not be present in the tree.  */

vm_splay_tree_node_t
vm_splay_tree_predecessor (sp, key)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
{
  int comparison;
  vm_splay_tree_node_t node;

  /* If the tree is empty, there is certainly no predecessor.  */
  if (!sp->root)
    return NULL;

  /* Splay the tree around KEY.  That will leave either the KEY
     itself, its predecessor, or its successor at the root.  */
  splay_tree_splay (sp, key);
  comparison = (*sp->comp)(sp->root->key, key);

  /* If the predecessor is at the root, just return it.  */
  if (comparison < 0)
    return sp->root;

  /* Otherwise, find the leftmost element of the right subtree.  */
  node = sp->root->left;
  if (node)
    while (node->right)
      node = node->right;

  return node;
}

/* Return the immediate successor KEY, or NULL if there is no
   predecessor.  KEY need not be present in the tree.  */

vm_splay_tree_node_t
vm_splay_tree_successor (sp, key)
     vm_splay_tree_t sp;
     vm_splay_tree_key key;
{
  int comparison;
  vm_splay_tree_node_t node;

  /* If the tree is empty, there is certainly no predecessor.  */
  if (!sp->root)
    return NULL;

  /* Splay the tree around KEY.  That will leave either the KEY
     itself, its predecessor, or its successor at the root.  */
  splay_tree_splay (sp, key);
  comparison = (*sp->comp)(sp->root->key, key);

  /* If the successor is at the root, just return it.  */
  if (comparison > 0)
    return sp->root;

  /* Otherwise, find the rightmost element of the left subtree.  */
  node = sp->root->right;
  if (node)
    while (node->left)
      node = node->left;

  return node;
}

/* Call FN, passing it the DATA, for every node in SP, following an
   in-order traversal.  If FN every returns a non-zero value, the
   iteration ceases immediately, and the value is returned.
   Otherwise, this function returns 0.  */

int
vm_splay_tree_foreach (sp, fn, data)
     vm_splay_tree_t sp;
     vm_splay_tree_foreach_fn fn;
     void *data;
{
  return splay_tree_foreach_helper (sp, sp->root, fn, data);
}

/* Splay-tree  Լ, key  int  մϴ.  */

int
vm_splay_tree_compare_ints (k1, k2)
     vm_splay_tree_key k1;
     vm_splay_tree_key k2;
{
  if ((int) k1 < (int) k2)
    return -1;
  else if ((int) k1 > (int) k2)
    return 1;
  else 
    return 0;
}

/* Splay-tree comparison function, treating the keys as pointers.  */

int
vm_splay_tree_compare_pointers (k1, k2)
     vm_splay_tree_key k1;
     vm_splay_tree_key k2;
{
  if ((char*) k1 < (char*) k2)
    return -1;
  else if ((char*) k1 > (char*) k2)
    return 1;
  else 
    return 0;
}

#if VM_THREAD_SAFE_SPLAY_TREE
static void 
splaytree_mutex_on (void)
{
  apr_status_t rv = apr_thread_mutex_lock (splaytree_mutex);

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't grab the splay tree mutex";

      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

static void 
splaytree_mutex_off (void)
{
  apr_status_t rv = apr_thread_mutex_unlock (splaytree_mutex);

  if (rv != APR_SUCCESS) 
    {
      const char *msg = "couldn't release the splay tree mutex";
      
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, NULL, msg);
      exit (APEXIT_CHILDFATAL);
    }
}

/* Thread-safe  Splay Tree    ʱȭ.  */
int
vm_init_splaytree (p)
     apr_pool_t *p;
{
  apr_status_t rv;

  rv = apr_thread_mutex_create (&splaytree_mutex, APR_THREAD_MUTEX_DEFAULT, p);
  if (rv != APR_SUCCESS) 
    {
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, ap_server_conf,
		    "Couldn't create splay tree lock");
      return -1;
    }

  return 0;
}

/* Splay Tree  .  */
int
vm_finish_splaytree (void)
{
  apr_status_t rv;

  rv = apr_thread_mutex_destroy (splaytree_mutex);

  if (rv != APR_SUCCESS) 
    {
      ap_log_error (APLOG_MARK, APLOG_EMERG, rv, ap_server_conf,
		    "Couldn't destroy splay tree lock");
      return -1;
    }

  return 0;
}
#endif
