/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#define HTTP_NOT_CONNECTED -1
#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "errors.h"
#include "hashtab.h"
#include "econnect.h"
#include "gensupport.h"

#define HUGE_STRING_LEN		8192
#define OK 0

/** @} */
/* Just in case your linefeed isn't the one the other end is expecting. */
#if !APR_CHARSET_EBCDIC
/** linefeed */
#define LF 10
/** carrige return */
#define CR 13
/** carrige return /Line Feed Combo */
#define CRLF "\015\012"
#else /* APR_CHARSET_EBCDIC */
/* For platforms using the EBCDIC charset, the transition ASCII->EBCDIC is done
 * in the buff package (bread/bputs/bwrite).  Everywhere else, we use
 * "native EBCDIC" CR and NL characters. These are therefore
 * defined as
 * '\r' and '\n'.
 */
#define CR '\r'
#define LF '\n'
#define CRLF "\r\n"
#endif /* APR_CHARSET_EBCDIC */

#define HTTP_REQUEST_URI_TOO_LARGE         414
#define HTTP_VERSION(major,minor) (1000*(major)+(minor))
#define HTTP_BAD_REQUEST                   400

/**
 * Iterate through a ring
 * @param ep The current element
 * @param hp The ring to iterate over
 * @param elem The name of the element struct
 * @param link The name of the APR_RING_ENTRY in the element struct
 * @remark This is the same as either:
 * <pre>
 *      ep = APR_RING_FIRST(hp);
 *      while (ep != APR_RING_SENTINEL(hp, elem, link)) {
 *          ...
 *          ep = APR_RING_NEXT(ep, link);
 *      }
 *   OR
 *      for (ep = APR_RING_FIRST(hp);
 *           ep != APR_RING_SENTINEL(hp, elem, link);
 *           ep = APR_RING_NEXT(ep, link)) {
 *          ...
 *      }
 * </pre>
 * @warning Be aware that you cannot change the value of ep within
 * the foreach loop, nor can you destroy the ring element it points to.
 * Modifying the prev and next pointers of the element is dangerous
 * but can be done if you're careful.  If you change ep's value or
 * destroy the element it points to, then APR_RING_FOREACH
 * will have no way to find out what element to use for its next
 * iteration.  The reason for this can be seen by looking closely
 * at the equivalent loops given in the tip above.  So, for example,
 * if you are writing a loop that empties out a ring one element
 * at a time, APR_RING_FOREACH just won't work for you.  Do it
 * by hand, like so:
 * <pre>
 *      while (!APR_RING_EMPTY(hp, elem, link)) {
 *          ep = APR_RING_FIRST(hp);
 *          ...
 *          APR_RING_REMOVE(ep, link);
 *      }
 * </pre>
 * @deprecated This macro causes more headaches than it's worth.  Use
 * one of the alternatives documented here instead; the clarity gained
 * in what's really going on is well worth the extra line or two of code.
 * This macro will be removed at some point in the future.
 */
#define APR_RING_FOREACH(ep, hp, elem, link)                            \
    for ((ep)  = APR_RING_FIRST((hp));                                  \
         (ep) != APR_RING_SENTINEL((hp), elem, link);                   \
         (ep)  = APR_RING_NEXT((ep), link))

/**
 * Iterate through a bucket brigade
 * @param e The current bucket
 * @param b The brigade to iterate over
 * @remark This is the same as either:
 * <pre>
 *      e = APR_BRIGADE_FIRST(b);
 *      while (e != APR_BRIGADE_SENTINEL(b)) {
 *          ...
 *          e = APR_BUCKET_NEXT(e);
 *      }
 *  OR
 *      for (e = APR_BRIGADE_FIRST(b);
 *           e != APR_BRIGADE_SENTINEL(b);
 *           e = APR_BUCKET_NEXT(e)) {
 *          ...
 *      }
 * </pre>
 * @warning Be aware that you cannot change the value of e within
 * the foreach loop, nor can you destroy the bucket it points to.
 * Modifying the prev and next pointers of the bucket is dangerous
 * but can be done if you're careful.  If you change e's value or
 * destroy the bucket it points to, then APR_BRIGADE_FOREACH
 * will have no way to find out what bucket to use for its next
 * iteration.  The reason for this can be seen by looking closely
 * at the equivalent loops given in the tip above.  So, for example,
 * if you are writing a loop that empties out a brigade one bucket
 * at a time, APR_BRIGADE_FOREACH just won't work for you.  Do it
 * by hand, like so:
 * <pre>
 *      while (!APR_BRIGADE_EMPTY(b)) {
 *          e = APR_BRIGADE_FIRST(b);
 *          ...
 *          apr_bucket_delete(e);
 *      }
 * </pre>
 * @deprecated This macro causes more headaches than it's worth.  Use
 * one of the alternatives documented here instead; the clarity gained
 * in what's really going on is well worth the extra line or two of code.
 * This macro will be removed at some point in the future.
 */
#define APR_BRIGADE_FOREACH(e, b)                                       \
        APR_RING_FOREACH((e), &(b)->list, apr_bucket, link)

#define APR_STATUS_IS_SUCCESS(s)           ((s) == APR_SUCCESS)

static char * ec_getword_white (apr_pool_t *, const char **);
static char * ec_escape_html (apr_pool_t *, const char *);
static void ec_get_mime_headers_core (ec_response_t *, apr_bucket_brigade *);
static int ec_get_brigade (ec_response_t *, apr_bucket_brigade *, enum ec_input_mode_t, apr_read_type_e, apr_off_t);
static int ec_request_write (ec_request_t *, const char *, size_t);
static int ec_get_msgblock (ec_response_t *, char **, apr_size_t);
static int ec_read_response_line (ec_response_t *, apr_bucket_brigade *);
static apr_status_t ec_rgetline (ec_response_t *, char **, apr_size_t, apr_size_t *, apr_pool_t *, int, apr_bucket_brigade *);
static apr_status_t ec_send_all_header_fields (ec_request_t *);

void ec_debug (const char *, ...);

void
ec_debug (const char *fmt, ...)
{
#ifdef EC_DEBUG
  va_list args;

  va_start (args, fmt);
  vfprintf (stderr, fmt, args);
  fprintf (stderr, "\n");
  va_end (args);
#endif
}

void
ec_file_init (ec_file_t *ef, unsigned char *buf, int len)
{
  ef->base = ef->cur = buf;
  ef->size = len;
}

/**
 * Remove all zero length buckets from the brigade.
 */
#define BRIGADE_NORMALIZE(b) \
do { \
    apr_bucket *e = APR_BRIGADE_FIRST(b); \
    do {  \
        if (e->length == 0 && !APR_BUCKET_IS_METADATA(e)) { \
            apr_bucket *d; \
            d = APR_BUCKET_NEXT(e); \
            apr_bucket_delete(e); \
            e = d; \
        } \
        e = APR_BUCKET_NEXT(e); \
    } while (!APR_BRIGADE_EMPTY(b) && (e != APR_BRIGADE_SENTINEL(b))); \
} while (0)

static int
ec_get_brigade (ec_response_t *res, apr_bucket_brigade *b,
		enum ec_input_mode_t mode, apr_read_type_e block,
		apr_off_t readbytes)
{
  apr_status_t rv;
  const char *str;
  apr_size_t len;

  if (APR_BRIGADE_EMPTY(res->bb))
    return APR_EOF;

  /* ### This is bad. */
  BRIGADE_NORMALIZE(res->bb);

  /* check for empty brigade again *AFTER* BRIGADE_NORMALIZE()
   * If we have lost our socket bucket (see above), we are EOF.
   *
   * Ideally, this should be returning SUCCESS with EOS bucket, but
   * some higher-up APIs (spec. read_response_line via ec_rgetline)
   * want an error code. */
  if (APR_BRIGADE_EMPTY(res->bb))
    return APR_EOF;

  if (mode == EC_MODE_GETLINE) 
    {
      /* we are reading a single LF line, e.g. the HTTP headers */
      rv = apr_brigade_split_line(b, res->bb, block, HUGE_STRING_LEN);
      /* We should treat EAGAIN here the same as we do for EOF (brigade is
       * empty).  We do this by returning whatever we have read.  This may
       * or may not be bogus, but is consistent (for now) with EOF logic.
       */
      if (APR_STATUS_IS_EAGAIN(rv))
	rv = APR_SUCCESS;

      return rv;
    }  

  /* read up to the amount they specified. */
  if (mode == EC_MODE_READBYTES || mode == EC_MODE_SPECULATIVE) 
    {
      apr_bucket *e;
      apr_bucket_brigade *newbb;
      
      e = APR_BRIGADE_FIRST(res->bb);
      rv = apr_bucket_read (e, &str, &len, block);
      
      if (APR_STATUS_IS_EAGAIN(rv))
	return APR_SUCCESS;

      else if (rv != APR_SUCCESS)
	return rv;
      else if (block == APR_BLOCK_READ && len == 0) 
	{
	  /* We wanted to read some bytes in blocking mode.  We read
	   * 0 bytes.  Hence, we now assume we are EOS.
	   *
	   * When we are in normal mode, return an EOS bucket to the
	   * caller.
	   * When we are in speculative mode, leave ctx->b empty, so
	   * that the next call returns an EOS bucket.
	   */
	  apr_bucket_delete (e);
	  
	  if (mode == EC_MODE_READBYTES) 
	    {
	      e = apr_bucket_eos_create(res->bb_alloc);
	      APR_BRIGADE_INSERT_TAIL(b, e);
            }
	  return APR_SUCCESS;
        }
      
      /* We can only return at most what we read. */
      if (len < (apr_size_t) readbytes) 
	{
	  readbytes = len;
        }
      
      rv = apr_brigade_partition(res->bb, readbytes, &e);
      if (rv != APR_SUCCESS) 
	return rv;
      
      /* Must do split before CONCAT */
      newbb = apr_brigade_split(res->bb, e);
      
      if (mode == EC_MODE_READBYTES) 
	{
	  APR_BRIGADE_CONCAT(b, res->bb);
	}
      else if (mode == EC_MODE_SPECULATIVE) 
	{
	  apr_bucket *copy_bucket;
	  APR_BRIGADE_FOREACH(e, res->bb) 
	    {
	      rv = apr_bucket_copy(e, &copy_bucket);
	      if (rv != APR_SUCCESS) 
		{
		  return rv;
		}
	      APR_BRIGADE_INSERT_TAIL(b, copy_bucket);
	    }
	}
      
      /* Take what was originally there and place it back on ctx->b */
      APR_BRIGADE_CONCAT(res->bb, newbb);
    }

  return APR_SUCCESS;
}

static apr_status_t 
ec_rgetline (ec_response_t *res, char **s, apr_size_t n,
	     apr_size_t *read, apr_pool_t *pool,
	     int fold, apr_bucket_brigade *bb)
{
/* Min # of bytes to allocate when reading a respons line */
#define MIN_LINE_ALLOC 80
  apr_status_t rv;
  apr_bucket *e;
  apr_size_t bytes_handled = 0, current_alloc = 0;
  char *pos, *last_char = *s;
  int do_alloc = (*s == NULL), saw_eos = 0;

  for (;;) 
    {
      apr_brigade_cleanup (bb);
      rv = ec_get_brigade (res, bb, EC_MODE_GETLINE, APR_BLOCK_READ, 0);
      
      if (rv != APR_SUCCESS)
        return rv;

      /* Something horribly wrong happened.  Someone didn't block! */
      if (APR_BRIGADE_EMPTY(bb))
        return APR_EGENERAL;
      
      APR_BRIGADE_FOREACH(e, bb) 
	{
	  const char *str;
	  apr_size_t len;
	  
	  /* If we see an EOS, don't bother doing anything more. */
	  if (APR_BUCKET_IS_EOS(e)) 
	    {
	      saw_eos = 1;
	      break;
	    }
	  
	  rv = apr_bucket_read (e, &str, &len, APR_BLOCK_READ);
	  
	  if (rv != APR_SUCCESS)
            return rv;

	  if (len == 0) 
	    {
	      /* no use attempting a zero-byte alloc (hurts when
	       * using --with-efence --enable-pool-debug) or
	       * doing any of the other logic either
	       */
	      continue;
	    }
	  
	  /* Would this overrun our buffer?  If so, we'll die. */
	  if (n < bytes_handled + len) 
	    {
	      *read = bytes_handled;
	      if (*s) 
		{
		  /* ensure this string is NUL terminated */
		  if (bytes_handled > 0) 
		    {
		      (*s)[bytes_handled-1] = '\0';
		    }
		  else 
		    {
		      (*s)[0] = '\0';
		    }
		}
	      return APR_ENOSPC;
	    }
	  
	  /* Do we have to handle the allocation ourselves? */
	  if (do_alloc) 
	    {
	      /* We'll assume the common case where one bucket is enough. */
	      if (!*s) 
		{
		  current_alloc = len;
		  if (current_alloc < MIN_LINE_ALLOC) 
		    {
		      current_alloc = MIN_LINE_ALLOC;
		    }
		  *s = apr_palloc(pool, current_alloc);
		}
	      else if (bytes_handled + len > current_alloc) 
		{
		  /* Increase the buffer size */
		  apr_size_t new_size = current_alloc * 2;
		  char *new_buffer;
		  
		  if (bytes_handled + len > new_size) 
		    {
		      new_size = (bytes_handled + len) * 2;
		    }
		  
		  new_buffer = apr_palloc(pool, new_size);
		  
		  /* Copy what we already had. */
		  memcpy(new_buffer, *s, bytes_handled);
		  current_alloc = new_size;
		  *s = new_buffer;
		}
	    }
	  
	  /* Just copy the rest of the data to the end of the old buffer. */
	  pos = *s + bytes_handled;
	  memcpy(pos, str, len);
	  last_char = pos + len - 1;
	  
	  /* We've now processed that new data - update accordingly. */
	  bytes_handled += len;
	}
      
      /* If we got a full line of input, stop reading */
      if (last_char && (*last_char == APR_ASCII_LF)) 
	{
	  break;
        }
    }
  
  /* Now NUL-terminate the string at the end of the line; 
   * if the last-but-one character is a CR, terminate there */
  if (last_char > *s && last_char[-1] == APR_ASCII_CR) 
    {
      last_char--;
    }
  *last_char = '\0';
  bytes_handled = last_char - *s;
  
  /* If we're folding, we have more work to do.
   *
   * Note that if an EOS was seen, we know we can't have another line.
   */
  if (fold && bytes_handled && !saw_eos) 
    {
      for (;;) 
	{
	  const char *str;
	  apr_size_t len;
	  char c;
	  
	  apr_brigade_cleanup (bb);
	  
	  /* We only care about the first byte. */
	  rv = ec_get_brigade (res, bb, EC_MODE_SPECULATIVE, APR_BLOCK_READ, 1);
	  
	  if (rv != APR_SUCCESS)
	    return rv;
	  
	  if (APR_BRIGADE_EMPTY(bb)) {
	    break;
	  }
	
        e = APR_BRIGADE_FIRST(bb);
	
        /* If we see an EOS, don't bother doing anything more. */
        if (APR_BUCKET_IS_EOS(e)) {
	  break;
        }
	
        rv = apr_bucket_read(e, &str, &len, APR_BLOCK_READ);
	
        if (rv != APR_SUCCESS) {
	  apr_brigade_cleanup(bb);
	  return rv;
        }
	
        /* Found one, so call ourselves again to get the next line.
         *
         * FIXME: If the folding line is completely blank, should we
         * stop folding?  Does that require also looking at the next
         * char?
         */
	/* When we call destroy, the buckets are deleted, so save that
	 * one character we need.  This simplifies our execution paths
	 * at the cost of one character read.
	 */
	c = *str;
        if (c == APR_ASCII_BLANK || c == APR_ASCII_TAB) {
	  /* Do we have enough space? We may be full now. */
	  if (bytes_handled >= n) {
	    *read = n;
	    /* ensure this string is terminated */
	    (*s)[n-1] = '\0';
	    return APR_ENOSPC;
	  }
	  else {
	    apr_size_t next_size, next_len;
	    char *tmp;
	    
	    /* If we're doing the allocations for them, we have to
	     * give ourselves a NULL and copy it on return.
	     */
	    if (do_alloc) {
	      tmp = NULL;
	    } else {
	      /* We're null terminated. */
	      tmp = last_char;
	    }
	    
	    next_size = n - bytes_handled;
	    
	    rv = ec_rgetline (res, &tmp, next_size,
			      &next_len, pool, 0, bb);
	    
	    if (rv != APR_SUCCESS) {
	      return rv;
	    }
	    
	    if (do_alloc && next_len > 0) {
	      char *new_buffer;
	      apr_size_t new_size = bytes_handled + next_len + 1;
	      
	      /* we need to alloc an extra byte for a null */
	      new_buffer = apr_palloc(pool, new_size);
	      
	      /* Copy what we already had. */
	      memcpy(new_buffer, *s, bytes_handled);
	      
	      /* copy the new line, including the trailing null */
	      memcpy(new_buffer + bytes_handled, tmp, next_len + 1);
	      *s = new_buffer;
	    }
	    
	    bytes_handled += next_len;
	  }
	}
	else { /* next character is not tab or space */
	  break;
	}
      }
    }
  
  *read = bytes_handled;
  return APR_SUCCESS;
}

ec_client_t *
ec_client_new (hostname, port)
     const char *hostname;
     const int port;
{
  /*  proxy   .  */
  const char *proxy_host = NULL;
  int proxy_port = 0;
  const char *use_interface = NULL;
  apr_pool_t *pool;
  apr_status_t status;
  ec_client_t *conn;

  apr_initialize ();
  apr_pool_create (&pool, NULL);

  conn = (ec_client_t *) apr_pcalloc (pool, sizeof (struct ec_client));

  if (!conn)
    return NULL;

  EC_CLIENT_POOL (conn) = pool;

  EC_CLIENT_PROXY_HOST (conn) = proxy_host ? apr_pstrdup (EC_CLIENT_POOL (conn), proxy_host) : NULL;
  EC_CLIENT_PROXY_PORT (conn) = proxy_port;
  EC_CLIENT_HOST (conn) = apr_pstrdup (EC_CLIENT_POOL (conn), hostname);
  EC_CLIENT_PORT (conn) = (apr_port_t) port;
  EC_CLIENT_USE_INTERFACE (conn) = apr_pstrdup (EC_CLIENT_POOL (conn), use_interface);

  if (EC_CLIENT_PROXY_PORT (conn))
    status = apr_sockaddr_info_get (&EC_CLIENT_SOCKADDR (conn), EC_CLIENT_PROXY_HOST (conn), APR_INET, 
				    EC_CLIENT_PROXY_PORT (conn), 0, EC_CLIENT_POOL (conn));
  else
    status = apr_sockaddr_info_get (&EC_CLIENT_SOCKADDR (conn), EC_CLIENT_HOST (conn), APR_INET, 
				    EC_CLIENT_PORT (conn), 0, EC_CLIENT_POOL (conn));

  if (!APR_STATUS_IS_SUCCESS (status))
    {
      fprintf (stderr, "Error creating new client connection %d \n", status);
      return NULL;
    }

  status = apr_sockaddr_info_get (&EC_CLIENT_INTFADDR (conn), EC_CLIENT_USE_INTERFACE (conn), APR_INET, 0, 0, EC_CLIENT_POOL (conn));

  if (!APR_STATUS_IS_SUCCESS (status))
    {
      fprintf (stderr, "Error creating new client connection %d \n", status);
      return NULL;
    }

  return conn;
}

apr_status_t 
ec_client_connect (conn)
     ec_client_t *conn;
{
  apr_status_t status;

  if (!conn)
    return -1;

  if (EC_CLIENT_PROXY_HOST (conn)) 
    ec_debug ("connect to %s:%d via %s:%d on intf %s\n", 
	      conn->host, conn->port, conn->proxy_host, conn->proxy_port, conn->use_interface);
  else
    ec_debug ("connect to %s:%d on intf %s\n", 
	      conn->host, conn->port, conn->use_interface);
  
  status = apr_socket_create (&EC_CLIENT_SOCKET (conn), APR_INET, SOCK_STREAM, 0, EC_CLIENT_POOL (conn));

  if (!APR_STATUS_IS_SUCCESS(status)) 
    {
      fprintf (stderr, "socket create failed.\n");
      return -1;
    }

  status = apr_socket_bind (EC_CLIENT_SOCKET (conn), EC_CLIENT_INTFADDR (conn));

  if (!APR_STATUS_IS_SUCCESS(status)) 
    {
      fprintf (stderr, "socket bind failed.\n");
      return -1;
    }

  status = apr_socket_connect (EC_CLIENT_SOCKET (conn), EC_CLIENT_SOCKADDR (conn));

  if (!APR_STATUS_IS_SUCCESS(status))
    {
      fprintf (stderr, "socket connect failed.\n");
      return -1;
    }

  return OK;
}

static int 
ec_request_write (req, buf, size)
     ec_request_t *req;
     const char *buf;
     size_t size;
{
  apr_status_t status = APR_SUCCESS;
  apr_size_t written = size;
  
  while (size > 0) 
    {
      status = apr_socket_send (req->conn->socket, buf, &written);
      if (written > 0) 
	{
	  buf += written;
	  size -= written;
	  written = size;
        } 
      else 
	break;
    }
  
  return APR_STATUS_IS_SUCCESS(status) ? 0 : -1;
}

ec_request_t *
ec_request_new (conn, method, uri)
     ec_client_t *conn;
     const char *method;
     const char *uri;
{
  int res;
  ec_request_t *req = NULL;
  char s[1024];

  req = (ec_request_t *) apr_pcalloc (EC_CLIENT_POOL (conn), sizeof(struct ec_request));
  if (!req)
    return NULL;

  req->pool = EC_CLIENT_POOL (conn);
  req->conn = conn;
  req->bb_alloc = apr_bucket_alloc_create (req->pool);
  req->bb = apr_brigade_create (req->pool, req->bb_alloc);
  req->headers_out = apr_table_make (EC_CLIENT_POOL (conn), 12);
  req->args = apr_table_make (EC_CLIENT_POOL (conn), 12);
  req->method = method;
  req->uri = uri;
  conn->req = req;

  return req;
}

static apr_status_t
ec_send_all_header_fields (req)
     ec_request_t *req;
{
  const apr_array_header_t *elts;
  const apr_table_entry_t *t_elt;
  const apr_table_entry_t *t_end;
  struct iovec *vec;
  struct iovec *vec_next;
  
  elts = apr_table_elts (req->headers_out);
  if (elts->nelts == 0)
    return APR_SUCCESS;

  t_elt = (const apr_table_entry_t *) (elts->elts);
  t_end = t_elt + elts->nelts;
  vec = (struct iovec *) apr_palloc (req->pool, 4 * elts->nelts * sizeof (struct iovec));
  vec_next = vec;

  /* For each field, generate
   *    name ": " value CRLF
   */
  do {
    vec_next->iov_base = (void*)(t_elt->key);
    vec_next->iov_len = strlen(t_elt->key);
    vec_next++;
    vec_next->iov_base = (void *) ": ";
    vec_next->iov_len = sizeof(": ") - 1;
    vec_next++;
    vec_next->iov_base = (void*)(t_elt->val);
    vec_next->iov_len = strlen(t_elt->val);
    vec_next++;
    vec_next->iov_base = (void *) CRLF;
    vec_next->iov_len = sizeof(CRLF) - 1;
    vec_next++;
    t_elt++;
  } while (t_elt < t_end);
  
  return apr_brigade_writev (req->bb, NULL, NULL, vec, vec_next - vec);
}

#define apr_isspace(c) (ISSPACE(((unsigned char)(c))))
#define apr_isdigit(c) (ISDIGIT(((unsigned char)(c))))

static char * 
ec_getword_white (apr_pool_t *atrans, const char **line)
{
  const char *pos = *line;
  int len;
  char *res;
  
  while (!apr_isspace(*pos) && *pos) 
    {
      ++pos;
    }
  
  len = pos - *line;
  res = (char *) apr_palloc (atrans, len + 1);
  memcpy (res, *line, len);
  res[len] = 0;
  
  while (apr_isspace (*pos)) 
    {
      ++pos;
    }
  
  *line = pos;
  
  return res;
}

/* Limits on the size of various request items.  These limits primarily
 * exist to prevent simple denial-of-service attacks on a server based
 * on misuse of the protocol.  The recommended values will depend on the
 * nature of the server resources -- CGI scripts and database backends
 * might require large values, but most servers could get by with much
 * smaller limits than we use below.  The request message body size can
 * be limited by the per-dir config directive LimitRequestBody.
 *
 * Internal buffer sizes are two bytes more than the DEFAULT_LIMIT_REQUEST_LINE
 * and DEFAULT_LIMIT_REQUEST_FIELDSIZE below, which explains the 8190.
 * These two limits can be lowered (but not raised) by the server config
 * directives LimitRequestLine and LimitRequestFieldsize, respectively.
 *
 * DEFAULT_LIMIT_REQUEST_FIELDS can be modified or disabled (set = 0) by
 * the server config directive LimitRequestFields.
 */
#ifndef DEFAULT_LIMIT_REQUEST_LINE
#define DEFAULT_LIMIT_REQUEST_LINE 8190
#endif /* default limit on bytes in Request-Line (Method+URI+HTTP-version) */
#ifndef DEFAULT_LIMIT_REQUEST_FIELDSIZE
#define DEFAULT_LIMIT_REQUEST_FIELDSIZE 8190
#endif /* default limit on bytes in any one header field  */
#ifndef DEFAULT_LIMIT_REQUEST_FIELDS
#define DEFAULT_LIMIT_REQUEST_FIELDS 100
#endif /* default limit on number of request header fields */

static int 
ec_read_response_line (res, bb)
     ec_response_t *res;
     apr_bucket_brigade *bb;
{
  const char *ll, *pro;
  
  unsigned int major = 1, minor = 0;   /* Assume HTTP/1.0 if non-"HTTP" protocol */
  char http[5];
  apr_size_t len;
  int num_blank_lines = 0;
  int max_blank_lines = 0;

  if (max_blank_lines <= 0) 
    max_blank_lines = DEFAULT_LIMIT_REQUEST_FIELDS;
  
  /* Read past empty lines until we get a real request line,
   * a read error, the connection closes (EOF), or we timeout.
   *
   * We skip empty lines because browsers have to tack a CRLF on to the end
   * of POSTs to support old CERN webservers.  But note that we may not
   * have flushed any previous response completely to the client yet.
   * We delay the flush as long as possible so that we can improve
   * performance for clients that are pipelining requests.  If a request
   * is pipelined then we won't block during the (implicit) read() below.
   * If the requests aren't pipelined, then the client is still waiting
   * for the final buffer flush from us, and we will block in the implicit
   * read().  B_SAFEREAD ensures that the BUFF layer flushes if it will
   * have to block during a read.
   */
  
  do 
    {
      apr_status_t rv;
      
      /* insure ec_rgetline allocates memory each time thru the loop
       * if there are empty lines
       */
      res->the_request = NULL;
      rv = ec_rgetline (res, &(res->the_request), (apr_size_t)(DEFAULT_LIMIT_REQUEST_LINE + 2),
			&len, res->pool, 0, bb);
      
      if (rv != APR_SUCCESS) 
	{
	  res->response_time = apr_time_now();
	  
	  /* ec_rgetline returns APR_ENOSPC if it fills up the
	   * buffer before finding the end-of-line.  This is only going to
	   * happen if it exceeds the configured limit for a request-line.
	   */
	  if (rv == APR_ENOSPC) 
	    {
	      res->status    = HTTP_REQUEST_URI_TOO_LARGE;
	      res->proto_num = HTTP_VERSION(1,0);
	      res->protocol  = apr_pstrdup (res->pool, "HTTP/1.0");
            }
	  
	  return 0;
        }
    } while ((len <= 0) && (++num_blank_lines < max_blank_lines));

  /* we've probably got something to do, ignore graceful restart requests */
  
  res->response_time = apr_time_now ();
  ll = res->the_request;
  pro = res->protocol = ec_getword_white (res->pool, &ll);
  len = strlen (pro);

  /* XXX ap_update_connection_status(conn->id, "Protocol", res->protocol); */
  
  /* Avoid sscanf in the common case */
  if (len == 8
      && pro[0] == 'H' && pro[1] == 'T' && pro[2] == 'T' && pro[3] == 'P'
      && pro[4] == '/' && apr_isdigit(pro[5]) && pro[6] == '.'
      && apr_isdigit(pro[7])) {
    res->proto_num = HTTP_VERSION(pro[5] - '0', pro[7] - '0');
  }
  else if (3 == sscanf (res->protocol, "%4s/%u.%u", http, &major, &minor)
	   && (strcasecmp("http", http) == 0)
	   && (minor < HTTP_VERSION(1, 0)) ) /* don't allow HTTP/0.1000 */
    res->proto_num = HTTP_VERSION(major, minor);
  else
    res->proto_num = HTTP_VERSION(1, 0);

  res->status = atoi (ec_getword_white (res->pool, &ll));
  res->status_msg = apr_pstrmemdup (res->pool, ll, strlen (ll));

  return 1;
}

/* ap_escape_uri is now a macro for os_escape_path */

static char * 
ec_escape_html (apr_pool_t *p, const char *s)
{
    int i, j;
    char *x;

    /* first, count the number of extra characters */
    for (i = 0, j = 0; s[i] != '\0'; i++)
        if (s[i] == '<' || s[i] == '>')
            j += 3;
        else if (s[i] == '&')
            j += 4;

    if (j == 0)
        return apr_pstrmemdup(p, s, i);

    x = apr_palloc(p, i + j + 1);
    for (i = 0, j = 0; s[i] != '\0'; i++, j++)
        if (s[i] == '<') {
            memcpy(&x[j], "&lt;", 4);
            j += 3;
        }
        else if (s[i] == '>') {
            memcpy(&x[j], "&gt;", 4);
            j += 3;
        }
        else if (s[i] == '&') {
            memcpy(&x[j], "&amp;", 5);
            j += 4;
        }
        else
            x[j] = s[i];

    x[j] = '\0';
    return x;
}

static void 
ec_get_mime_headers_core (res, bb)
     ec_response_t *res;
     apr_bucket_brigade *bb;
{
  char *last_field = NULL;
  apr_size_t last_len = 0;
  apr_size_t alloc_len = 0;
  char *field;
  char *value;
  apr_size_t len;
  int fields_read = 0;
  char *tmp_field;
  
  /*
   * Read header lines until we get the empty separator line, a read error,
   * the connection closes (EOF), reach the server limit, or we timeout.
   */
  while(1) 
    {
      apr_status_t rv;
      int folded = 0;
      
      field = NULL;
      rv = ec_rgetline (res, &field, DEFAULT_LIMIT_REQUEST_LINE + 2,
			&len, res->pool, 0, bb);

      /* ap_rgetline returns APR_ENOSPC if it fills up the buffer before
       * finding the end-of-line.  This is only going to happen if it
       * exceeds the configured limit for a field size.
       */
      if (rv == APR_ENOSPC && field) 
	{
	  /* insure ec_escape_html will terminate correctly */
	  field[len - 1] = '\0';
	  apr_table_setn(res->notes, "error-notes",
			 apr_pstrcat(res->pool,
				     "Size of a request header field "
				     "exceeds server limit.<br />\n"
				     "<pre>\n",
				     ec_escape_html(res->pool, field),
				     "</pre>\n", NULL));
	  return;
        }

        if (rv != APR_SUCCESS) 
	  {
            res->status = HTTP_BAD_REQUEST;
            return;
	  }
	
        if (last_field != NULL) 
	  {
            if ((len > 0) && ((*field == '\t') || *field == ' ')) 
	      {
                /* This line is a continuation of the preceding line(s),
                 * so append it to the line that we've set aside.
                 * Note: this uses a power-of-two allocator to avoid
                 * doing O(n) allocs and using O(n^2) space for
                 * continuations that span many many lines.
                 */
                apr_size_t fold_len = last_len + len + 1; /* trailing null */
		
                if ((fold_len - 1) > DEFAULT_LIMIT_REQUEST_LINE) 
		  {
                    res->status = HTTP_BAD_REQUEST;
                    /* report what we have accumulated so far before the
                     * overflow (last_field) as the field with the problem
                     */
                    apr_table_setn(res->notes, "error-notes",
                                   apr_pstrcat(res->pool,
                                               "Size of a request header field " 
                                               "after folding "
                                               "exceeds server limit.<br />\n"
                                               "<pre>\n",
                                               ec_escape_html(res->pool, last_field),
                                               "</pre>\n", NULL));
                    return;
		  }
		
                if (fold_len > alloc_len) 
		  {
                    char *fold_buf;
                    alloc_len += alloc_len;
                    if (fold_len > alloc_len) 
		      {
                        alloc_len = fold_len;
		      }
                    fold_buf = (char *)apr_palloc(res->pool, alloc_len);
                    memcpy(fold_buf, last_field, last_len);
                    last_field = fold_buf;
		  }
                memcpy(last_field + last_len, field, len +1); /* +1 for nul */
                last_len += len;
                folded = 1;
	      }
            else /* not a continuation line */ 
	      {
		
                if (DEFAULT_LIMIT_REQUEST_FIELDS
                    && (++fields_read > DEFAULT_LIMIT_REQUEST_FIELDS)) 
		  {
		    res->status = HTTP_BAD_REQUEST;
		    apr_table_setn(res->notes, "error-notes",
				   "The number of request header fields "
                                   "exceeds this server's limit.");
                    return;
		  }
		
                if (!(value = strchr(last_field, ':'))) 
		  { /* Find ':' or    */
                    res->status = HTTP_BAD_REQUEST;      /* abort bad request */
                    apr_table_setn(res->notes, "error-notes",
                                   apr_pstrcat(res->pool,
                                               "Request header field is "
                                               "missing ':' separator.<br />\n"
                                               "<pre>\n",
                                               ec_escape_html(res->pool,
                                                              last_field),
                                               "</pre>\n", NULL));
                    return;
		  }
                
                tmp_field = value - 1; /* last character of field-name */
		
                *value++ = '\0'; /* NUL-terminate at colon */
		
                while (*value == ' ' || *value == '\t') 
		  {
                    ++value;            /* Skip to start of value   */
		  }
		
                /* Strip LWS after field-name: */
                while (tmp_field > last_field 
                       && (*tmp_field == ' ' || *tmp_field == '\t')) 
		  {
                    *tmp_field-- = '\0';
		  }
                
                /* Strip LWS after field-value: */
                tmp_field = last_field + last_len - 1;
                while (tmp_field > value
                       && (*tmp_field == ' ' || *tmp_field == '\t')) 
		  {
		    *tmp_field-- = '\0';
		  }
		
                apr_table_addn(res->headers_in, last_field, value);
                
                /* reset the alloc_len so that we'll allocate a new
                 * buffer if we have to do any more folding: we can't
                 * use the previous buffer because its contents are
                 * now part of res->headers_in
                 */
                alloc_len = 0;
		
	      } /* end if current line is not a continuation starting with tab */
	  }
	
        /* Found a blank line, stop. */
        if (len == 0) 
	  {
            break;
	  }
	
        /* Keep track of this line so that we can parse it on
         * the next loop iteration.  (In the folded case, last_field
         * has been updated already.)
         */
        if (!folded) 
	  {
            last_field = field;
            last_len = len;
	  }
    }
  
  apr_table_compress (res->headers_in, APR_OVERLAP_TABLES_MERGE);
}

static int
ec_get_msgblock (ec_response_t *res, char **buffer,
		 apr_size_t bufsiz)
{
  apr_status_t rv;
  apr_bucket_brigade *bb;
  
  if (res->remaining < 0 || (!res->read_chunked && res->remaining == 0))
    return 0;
  
  bb = apr_brigade_create(res->pool, res->bb_alloc);
  if (bb == NULL)
    return -1;
  
  rv = ec_get_brigade (res, bb, EC_MODE_READBYTES, APR_BLOCK_READ, bufsiz);
  
  /* We lose the failure code here.  This is why ap_get_client_block should
   * not be used.
   */
  if (rv != APR_SUCCESS) 
    {
      apr_brigade_destroy (bb);
      return -1;
    }
  
  /* Check to see if EOS in the brigade.
   *
   * If so, we have to leave a nugget for the *next* ap_get_client_block
   * call to return 0.
   */
  if (APR_BUCKET_IS_EOS(APR_BRIGADE_LAST (bb))) 
    {
      if (res->read_chunked) 
	{
	  res->remaining = -1;
        }
      else 
	{
	  res->remaining = 0;
	}
    }
  
  rv = apr_brigade_pflatten (bb, buffer, &bufsiz, res->pool);
  if (rv != APR_SUCCESS) 
    {
      apr_brigade_destroy(bb);
      return -1;
    }
  
  /* XXX yank me? */
  res->read_length += bufsiz;
  
  apr_brigade_destroy (bb);
  
  return bufsiz;
}

apr_status_t
ec_request_set_handler (req, handler, arg)
     ec_request_t *req;
     apr_status_t (*handler) (ec_response_t *, void *);
     void *arg;
{
  req->handler = handler;
  req->handler_arg = arg;

  return 0;
}

/**
   TODO: Էµ NAME  VALUE   ݿ ʾҴ.
*/
apr_status_t
ec_request_add_arg (req, name, value)
     ec_request_t *req;
     const char *name, *value;
{
  apr_table_addn (req->args, name, value);

  return 0;
}

/*
 * More machine-dependent networking gooo... on some systems,
 * you've got to be *really* sure that all the packets are acknowledged
 * before closing the connection, since the client will not be able
 * to see the last response if their TCP buffer is flushed by a RST
 * packet from us, which is what the server's TCP stack will send
 * if it receives any request data after closing the connection.
 *
 * In an ideal world, this function would be accomplished by simply
 * setting the socket option SO_LINGER and handling it within the
 * server's TCP stack while the process continues on to the next request.
 * Unfortunately, it seems that most (if not all) operating systems
 * block the server process on close() when SO_LINGER is used.
 * For those that don't, see USE_SO_LINGER below.  For the rest,
 * we have created a home-brew lingering_close.
 *
 * Many operating systems tend to block, puke, or otherwise mishandle
 * calls to shutdown only half of the connection.  You should define
 * NO_LINGCLOSE in ap_config.h if such is the case for your system.
 */
#ifndef MAX_SECS_TO_LINGER
#define MAX_SECS_TO_LINGER 30
#endif

ec_response_t *
ec_request_send (req)
     ec_request_t *req;
{
#define SECONDS_TO_LINGER  2
  char dummybuf[512];
  apr_size_t nbytes = sizeof (dummybuf);
  apr_status_t rc;
  apr_bucket *e;
  apr_int32_t timeout;
  apr_int32_t total_linger_time = 0;
  apr_socket_t *csd = EC_CLIENT_SOCKET (EC_REQUEST_CLIENT (req));
  ec_response_t *res;
  char *buf, *header_buf;
  apr_size_t header_length = 0;
  apr_bucket_brigade *tmp_bb;

  res = (ec_response_t *) apr_pcalloc (EC_REQUEST_POOL (req), sizeof (struct ec_response));
  res->pool = EC_REQUEST_POOL (req);
  res->bb_alloc = req->bb_alloc;
  res->remaining = 1;
  res->read_chunked = 0;
  res->read_length = 0;
  res->bb = apr_brigade_create (EC_REQUEST_POOL (req), res->bb_alloc);
  res->headers_in	= apr_table_make (EC_REQUEST_POOL (req), 25);
  res->notes		= apr_table_make (EC_REQUEST_POOL (req), 5);
  res->conn	 = req->conn;
  req->conn->res = res;

  /* Request header ϼ.  */
  ec_send_all_header_fields (req);
  rc = apr_brigade_pflatten (req->bb, &header_buf, &header_length, EC_REQUEST_POOL (req));
  ec_debug ("headerbuf = %s, length = %d\n", header_buf, header_length);
  buf = apr_pstrcat (EC_REQUEST_POOL (req),
		       apr_psprintf (EC_REQUEST_POOL (req),
				     "%s %s HTTP/1.0\r\n", 
				     EC_REQUEST_METHOD (req), EC_REQUEST_URI (req)), 
		       apr_pstrndup (EC_REQUEST_POOL (req), header_buf, header_length),
		       "\r\n", NULL);
  ec_debug ("lines:\n%s\n+  --\n", buf);
  ec_request_write (req, buf, strlen (buf));

  /* Response Header  Body ó.  */

  /* Read all data from the peer until we reach "end-of-file" (FIN
   * from peer) or we've exceeded our overall timeout. If the client does
   * not send us bytes within 2 seconds (a value pulled from Apache 1.3
   * which seems to work well), close the connection.
   */
  timeout = apr_time_from_sec (SECONDS_TO_LINGER);
  apr_socket_timeout_set (csd, timeout);
  apr_socket_opt_set (csd, APR_INCOMPLETE_READ, 1);

  /* seed the brigade with the client socket. */
  e = apr_bucket_socket_create(csd, res->bb_alloc);
  APR_BRIGADE_INSERT_TAIL(res->bb, e);

  tmp_bb = apr_brigade_create (res->pool, res->bb_alloc);

  if (!ec_read_response_line (res, tmp_bb))
    {
      /*  ó.  */
    }

  ec_get_mime_headers_core (res, tmp_bb);

  res->bodylen = ec_get_msgblock (res, &res->body, 1024);

  apr_brigade_destroy (tmp_bb);

  /* Res  ó handler  ϵǾ  , ش Լ ȣ ݴϴ.  */
  if (req->handler)
    {
      if (req->handler (res, req->handler_arg))
        return NULL;
    }

  return res;
}

void
ec_client_close (conn)
     ec_client_t *conn;
{
  ec_request_t *req = conn->req;
  ec_response_t *res = conn->res;

  if (conn->socket) 
    {
      apr_socket_shutdown (conn->socket, APR_SHUTDOWN_READWRITE);
      apr_socket_close (conn->socket);
      conn->socket = NULL;
    }

  apr_pool_destroy (conn->pool);

  apr_terminate ();
}
