/* ETL utility routines.
   Copyright (C) 1987, 1988, 1991, 1994, 1997, 1998, 1999, 2000, 2001, 2002,
   2003 Free Software Foundation, Inc.

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "errors.h"

/* Indexed by etx code, gives number of operands for an etx with that code.
   Does NOT include etx header data (code and links).  */

#define DEF_ETL_EXPR(ENUM, NAME, FORMAT, CLASS)   sizeof FORMAT - 1 ,

const unsigned char etx_length[NUM_ETX_CODE] = {
#include "etl.def"
};

#undef DEF_ETL_EXPR

/* Indexed by etx code, gives the name of that kind of etx, as a C string.  */

#define DEF_ETL_EXPR(ENUM, NAME, FORMAT, CLASS)   NAME ,

const char * const etx_name[NUM_ETX_CODE] = {
#include "etl.def"		/* etl expressions are documented here */
};

#undef DEF_ETL_EXPR

/* Indexed by etx code, gives a sequence of operand-types for
   etx's of that code.  The sequence is a C string in which
   each character describes one operand.  */

const char * const etx_format[NUM_ETX_CODE] = {
  /* "*" undefined.
         can cause a warning message
     "0" field is unused (or used in a phase-dependent manner)
         prints nothing
     "i" an integer
         prints the integer
     "n" like "i", but prints entries from `note_insn_name'
     "w" an integer of width HOST_BITS_PER_WIDE_INT
         prints the integer
     "s" a pointer to a string
         prints the string
     "S" like "s", but optional:
	 the containing etx may end before this operand
     "T" like "s", but treated specially by the ETL reader;
         only found in machine description patterns.
     "e" a pointer to an etl expression
         prints the expression
     "E" a pointer to a vector that points to a number of etl expressions
         prints a list of the etl expressions
     "V" like "E", but optional:
	 the containing etx may end before this operand
     "u" a pointer to another insn
         prints the uid of the insn.
     "b" is a pointer to a bitmap header.
     "B" is a basic block pointer.
     "t" is a etree pointer.  */

#define DEF_ETL_EXPR(ENUM, NAME, FORMAT, CLASS)   FORMAT ,
#include "etl.def"		/* etl expressions are defined here */
#undef DEF_ETL_EXPR
};

/* Indexed by etx code, gives a character representing the "class" of
   that etx code.  See etl.def for documentation on the defined classes.  */

const char etx_class[NUM_ETX_CODE] = {
#define DEF_ETL_EXPR(ENUM, NAME, FORMAT, CLASS)   CLASS,
#include "etl.def"		/* etl expressions are defined here */
#undef DEF_ETL_EXPR
};

/* Indexed by etx code, gives the size of the etx in bytes.  */

const unsigned char etx_size[NUM_ETX_CODE] = {
#define DEF_ETL_EXPR(ENUM, NAME, FORMAT, CLASS)				\
  ((ENUM) == ETX_CONST_INT 			\
   ? ETX_HDR_SIZE + (sizeof FORMAT - 1) * sizeof (HOST_WIDE_INT)	\
   : ETX_HDR_SIZE + (sizeof FORMAT - 1) * sizeof (etunion)),

#include "etl.def"
#undef DEF_ETL_EXPR
};

/* Names for kinds of NOTEs and REG_NOTEs.  */
#if 0
const char * const note_insn_name[NOTE_INSN_MAX - NOTE_INSN_BIAS] =
{
  "", "NOTE_INSN_DELETED",
  "NOTE_INSN_BLOCK_BEG", "NOTE_INSN_BLOCK_END",
  "NOTE_INSN_LOOP_BEG", "NOTE_INSN_LOOP_END",
  "NOTE_INSN_LOOP_CONT", "NOTE_INSN_LOOP_VTOP",
  "NOTE_INSN_LOOP_END_TOP_COND", "NOTE_INSN_FUNCTION_END",
  "NOTE_INSN_PROLOGUE_END", "NOTE_INSN_EPILOGUE_BEG",
  "NOTE_INSN_DELETED_LABEL", "NOTE_INSN_FUNCTION_BEG",
  "NOTE_INSN_EH_REGION_BEG", "NOTE_INSN_EH_REGION_END",
  "NOTE_INSN_REPEATED_LINE_NUMBER",
  "NOTE_INSN_BASIC_BLOCK", "NOTE_INSN_EXPECTED_VALUE",
  "NOTE_INSN_PREDICTION"
};

const char * const reg_note_name[] =
{
  "", "REG_DEAD", "REG_INC", "REG_EQUIV", "REG_EQUAL",
  "REG_RETVAL", "REG_LIBCALL", "REG_NONNEG",
  "REG_NO_CONFLICT", "REG_UNUSED", "REG_CC_SETTER", "REG_CC_USER",
  "REG_LABEL", "REG_DEP_ANTI", "REG_DEP_OUTPUT", "REG_BR_PROB",
  "REG_VALUE_PROFILE", "REG_NOALIAS", "REG_SAVE_AREA", "REG_BR_PRED",
  "REG_FRAME_RELATED_EXPR", "REG_EH_CONTEXT", "REG_EH_REGION",
  "REG_SAVE_NOTE", "REG_MAYBE_DEAD", "REG_NORETURN",
  "REG_NON_LOCAL_GOTO", "REG_SETJMP", "REG_ALWAYS_RETURN",
  "REG_VTABLE_REF"
};
#endif


#ifdef GATHER_STATISTICS
static int etx_alloc_counts[(int) LAST_AND_UNUSED_ETX_CODE];
static int etx_alloc_sizes[(int) LAST_AND_UNUSED_ETX_CODE];
static int etvec_alloc_counts;
static int etvec_alloc_sizes;
#endif

#define ggc_alloc_etx(POOL, CODE)                    \
  ((etx) apr_pcalloc (POOL, ETX_SIZE (CODE)))

#define ggc_alloc_etvec(POOL, NELT)						  \
  ((etvec) apr_pcalloc (POOL, sizeof (struct etvec_def) \
		      + ((NELT) - 1) * sizeof (etx)))


/* Allocate an etx vector of N elements.
   Store the length, and initialize all elements to zero.  */

etvec
etvec_alloc (apr_pool_t *pool, int n)
{
  etvec rt;

  rt = ggc_alloc_etvec (pool, n);
  /* clear out the vector */
  memset (&rt->elem[0], 0, n * sizeof (etx));

  ETX_PUT_NUM_ELEM (rt, n);

#ifdef GATHER_STATISTICS
  etvec_alloc_counts++;
  etvec_alloc_sizes += n * sizeof (etx);
#endif

  return rt;
}

/* Allocate an etx of code CODE.  The CODE is stored in the etx;
   all the rest is initialized to zero.  */

etx
etx_alloc (apr_pool_t *pool, ETX_CODE code)
{
  etx rt;

  rt = ggc_alloc_etx (pool, code);

  /* We want to clear everything up to the FLD array.  Normally, this
     is one int, but we don't want to assume that and it isn't very
     portable anyway; this is.  */

  memset (rt, 0, ETX_HDR_SIZE);
  ETX_PUT_CODE (rt, code);

#ifdef GATHER_STATISTICS
  etx_alloc_counts[code]++;
  etx_alloc_sizes[code] += ETX_SIZE (code);
#endif

  return rt;
}


/* Create a new copy of an etx.
   Recursively copies the operands of the etx,
   except for those few etx codes that are sharable.  */

etx
copy_etx (apr_pool_t *pool, etx orig)
{
  etx copy;
  int i, j;
  ETX_CODE code;
  const char *format_ptr;

  code = ETX_GET_CODE (orig);

  switch (code)
    {
    case ETX_CONST_INT:
      return orig;

      /* A MEM with a constant address is not sharable.  The problem is that
	 the constant address may need to be reloaded.  If the mem is shared,
	 then reloading one copy of this mem will cause all copies to appear
	 to have been reloaded.  */

    default:
      break;
    }

  copy = etx_alloc (pool, code);

  /* Copy the various flags, and other information.  We assume that
     all fields need copying, and then clear the fields that should
     not be copied.  That is the sensible default behavior, and forces
     us to explicitly document why we are *not* copying a flag.  */
  memcpy (copy, orig, ETX_HDR_SIZE);

  /* We do not copy the USED flag, which is used as a mark bit during
     walks over the ETL.  */
  ETX_FLAG (copy, used) = 0;

  /* We do not copy FRAME_RELATED for INSNs.  */
  if (GET_ETX_CLASS (code) == 'i')
    ETX_FLAG (copy, frame_related) = 0;
  ETX_FLAG (copy, jump) = ETX_FLAG (orig, jump);
  ETX_FLAG (copy, call) = ETX_FLAG (orig, call);

  format_ptr = GET_ETX_FORMAT (ETX_GET_CODE (copy));

  for (i = 0; i < GET_ETX_LENGTH (ETX_GET_CODE (copy)); i++)
    {
      copy->u.fld[i] = orig->u.fld[i];
      switch (*format_ptr++)
	{
	case 'e':
	  if (ETX_XEXP (orig, i) != NULL)
	    ETX_XEXP (copy, i) = copy_etx (pool, ETX_XEXP (orig, i));
	  break;

	case 'E':
	case 'V':
	  if (ETX_XVEC (orig, i) != NULL)
	    {
	      ETX_XVEC (copy, i) = etvec_alloc (pool, ETX_XVECLEN (orig, i));
	      for (j = 0; j < ETX_XVECLEN (copy, i); j++)
		ETX_XVECEXP (copy, i, j) = copy_etx (pool, ETX_XVECEXP (orig, i, j));
	    }
	  break;

	case 't':
	case 'w':
	case 'i':
	case 's':
	case 'S':
	case 'T':
	case 'u':
	case 'B':
	case '0':
	  /* These are left unchanged.  */
	  break;

	default:
	  abort ();
	}
    }
  return copy;
}

/* Create a new copy of an etx.  Only copy just one level.  */

etx
shallow_copy_etx (apr_pool_t *pool, etx orig)
{
  etx copy;

  copy = ggc_alloc_etx (pool, ETX_GET_CODE (orig));
  memcpy (copy, orig, ETX_SIZE (ETX_GET_CODE (orig)));
  return copy;
}

/* This is 1 until after the etl generation pass.  */
int etx_equal_function_value_matters;

/* Nonzero when we are generating CONCATs.  */
int generating_concat_p;

/* Return 1 if X and Y are identical-looking etx's.
   This is the Lisp function EQUAL for etx arguments.  */

int
etx_equal_p (etx x, etx y)
{
  int i;
  int j;
  enum etx_code code;
  const char *fmt;

  if (x == y)
    return 1;
  if (x == 0 || y == 0)
    return 0;

  code = ETX_GET_CODE (x);
  /* Rtx's of different codes cannot be equal.  */
  if (code != ETX_GET_CODE (y))
    return 0;

  /* (MULT:SI x y) and (MULT:HI x y) are NOT equivalent.
     (REG:SI x) and (REG:HI x) are NOT equivalent.  */

  /* Some ETL can be compared nonrecursively.  */
  switch (code)
    {
    case ETX_CONST_INT:
      return 0;

    default:
      break;
    }

  /* Compare the elements.  If any pair of corresponding elements
     fail to match, return 0 for the whole thing.  */

  fmt = GET_ETX_FORMAT (code);
  for (i = GET_ETX_LENGTH (code) - 1; i >= 0; i--)
    {
      switch (fmt[i])
	{
	case 'w':
	  if (ETX_XWINT (x, i) != ETX_XWINT (y, i))
	    return 0;
	  break;

	case 'n':
	case 'i':
	  if (ETX_XINT (x, i) != ETX_XINT (y, i))
	    return 0;
	  break;

	case 'V':
	case 'E':
	  /* Two vectors must have the same length.  */
	  if (ETX_XVECLEN (x, i) != ETX_XVECLEN (y, i))
	    return 0;

	  /* And the corresponding elements must match.  */
	  for (j = 0; j < ETX_XVECLEN (x, i); j++)
	    if (etx_equal_p (ETX_XVECEXP (x, i, j), ETX_XVECEXP (y, i, j)) == 0)
	      return 0;
	  break;

	case 'e':
	  if (etx_equal_p (ETX_XEXP (x, i), ETX_XEXP (y, i)) == 0)
	    return 0;
	  break;

	case 'S':
	case 's':
	  if ((ETX_XSTR (x, i) || ETX_XSTR (y, i))
	      && (! ETX_XSTR (x, i) || ! ETX_XSTR (y, i)
		  || strcmp (ETX_XSTR (x, i), ETX_XSTR (y, i))))
	    return 0;
	  break;

	case 'u':
	  /* These are just backpointers, so they don't matter.  */
	  break;

	case '0':
	case 't':
	  break;

	  /* It is believed that etx's at this level will never
	     contain anything but integers and other etx's,
	     except for within LABEL_REFs and SYMBOL_REFs.  */
	default:
	  abort ();
	}
    }
  return 1;
}

void dump_etx_statistics (void)
{
#ifdef GATHER_STATISTICS
  int i;
  int total_counts = 0;
  int total_sizes = 0;
  fprintf (stderr, "\nETX Kind               Count      Bytes\n");
  fprintf (stderr, "---------------------------------------\n");
  for (i = 0; i < LAST_AND_UNUSED_ETX_CODE; i++)
    if (etx_alloc_counts[i])
      {
        fprintf (stderr, "%-20s %7d %10d\n", GET_ETX_NAME (i),
                 etx_alloc_counts[i], etx_alloc_sizes[i]);
        total_counts += etx_alloc_counts[i];
        total_sizes += etx_alloc_sizes[i];
      }
  if (etvec_alloc_counts)
    {
      fprintf (stderr, "%-20s %7d %10d\n", "etvec",
               etvec_alloc_counts, etvec_alloc_sizes);
      total_counts += etvec_alloc_counts;
      total_sizes += etvec_alloc_sizes;
    }
  fprintf (stderr, "---------------------------------------\n");
  fprintf (stderr, "%-20s %7d %10d\n",
           "Total", total_counts, total_sizes);
  fprintf (stderr, "---------------------------------------\n");
#endif  
}

#if defined ENABLE_ETL_CHECKING && (GCC_VERSION >= 2007)
void
etl_check_failed_bounds (etx r, int n, const char *file, int line,
			 const char *func)
{
  internal_error
    ("ETL check: access of elt %d of `%s' with last elt %d in %s, at %s:%d",
     n, GET_ETX_NAME (ETX_GET_CODE (r)), GET_ETX_LENGTH (ETX_GET_CODE (r)) - 1,
     func, trim_filename (file), line);
}

void
etl_check_failed_type1 (etx r, int n, int c1, const char *file, int line,
			const char *func)
{
  internal_error
    ("ETL check: expected elt %d type '%c', have '%c' (etx %s) in %s, at %s:%d",
     n, c1, GET_ETX_FORMAT (ETX_GET_CODE (r))[n], GET_ETX_NAME (ETX_GET_CODE (r)),
     func, trim_filename (file), line);
}

void
etl_check_failed_type2 (etx r, int n, int c1, int c2, const char *file,
			int line, const char *func)
{
  internal_error
    ("ETL check: expected elt %d type '%c' or '%c', have '%c' (etx %s) in %s, at %s:%d",
     n, c1, c2, GET_ETX_FORMAT (ETX_GET_CODE (r))[n], GET_ETX_NAME (ETX_GET_CODE (r)),
     func, trim_filename (file), line);
}

void
etl_check_failed_code1 (etx r, enum etx_code code, const char *file,
			int line, const char *func)
{
  internal_error ("ETL check: expected code `%s', have `%s' in %s, at %s:%d",
		  GET_ETX_NAME (code), GET_ETX_NAME (ETX_GET_CODE (r)), func,
		  trim_filename (file), line);
}

void
etl_check_failed_code2 (etx r, enum etx_code code1, enum etx_code code2,
			const char *file, int line, const char *func)
{
  internal_error
    ("ETL check: expected code `%s' or `%s', have `%s' in %s, at %s:%d",
     GET_ETX_NAME (code1), GET_ETX_NAME (code2), GET_ETX_NAME (ETX_GET_CODE (r)),
     func, trim_filename (file), line);
}

/* XXX Maybe print the vector?  */
void
etvec_check_failed_bounds (etvec r, int n, const char *file, int line,
			   const char *func)
{
  internal_error
    ("ETL check: access of elt %d of vector with last elt %d in %s, at %s:%d",
     n, GET_NUM_ELEM (r) - 1, func, trim_filename (file), line);
}
#endif /* ENABLE_ETL_CHECKING */

#if defined ENABLE_ETL_FLAG_CHECKING
void
etl_check_failed_flag (const char *name, etx r, const char *file,
		       int line, const char *func)
{
  internal_error
    ("ETL flag check: %s used with unexpected etx code `%s' in %s, at %s:%d",
     name, GET_ETX_NAME (ETX_GET_CODE (r)), func, trim_filename (file), line);
}
#endif /* ENABLE_ETL_FLAG_CHECKING */
