/* Support routines for the various generation passes.
   Copyright (C) 2000, 2001, 2002, 2003, 2004
   Free Software Foundation, Inc.

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "errors.h"
#include "hashtab.h"
#include "econnect.h"
#include "gensupport.h"


/* We initially queue all patterns, process the define_insn and
   define_cond_exec patterns, then return them one at a time.  */

struct queue_elem
{
  etx data;
  const char *filename;
  int lineno;
  struct queue_elem *next;
};

extern int ec_ungetc (int c, ec_file_t *ef);

static void queue_pattern (apr_pool_t *, etx, struct queue_elem ***,
			   const char *, int);
static void process_etx (struct etx_classobj *, etx, int);
static char *save_string (apr_pool_t *, const char *, int);

void
message_with_line (struct etx_classobj *rc, int lineno, const char *msg, ...)
{
  va_list ap;

  va_start (ap, msg);

  fprintf (stderr, "%s:%d: ", rc->read_etx_filename, lineno);
  vfprintf (stderr, msg, ap);
  fputc ('\n', stderr);

  va_end (ap);
}


/* Queue PATTERN on LIST_TAIL.  */

static void
queue_pattern (apr_pool_t *pool, etx pattern, struct queue_elem ***list_tail,
	       const char *filename, int lineno)
{
  struct queue_elem *e = apr_pcalloc (pool, sizeof (*e));
  e->data = pattern;
  e->filename = filename;
  e->lineno = lineno;
  e->next = NULL;
  **list_tail = e;
  *list_tail = &e->next;
}

/* Process a top level etx in some way, queuing as appropriate.  */

static void
process_etx (struct etx_classobj *rc, etx desc, int lineno)
{
  switch (ETX_GET_CODE (desc))
    {
    default:
      queue_pattern (rc->pool, desc, &rc->other_tail, rc->read_etx_filename, lineno);
      break;
    }
}

static char *
save_string (apr_pool_t *pool, const char *s, int len)
{
  char *result = apr_pcalloc (pool, len + 1);

  memcpy (result, s, len);
  result[len] = 0;
  return result;
}


/* The entry point for initializing the reader.  */

int
ecd_init_reader_args (struct etx_classobj **rc, ec_file_t *ef)
{
  apr_pool_t *pool;
  struct etx_classobj *etxclass;

  apr_initialize ();
  apr_pool_create (&pool, NULL);  

  etxclass = (struct etx_classobj *) apr_pcalloc (pool, sizeof (struct etx_classobj));

  /* -- ʱȭ  ---- */
  etxclass->pool = pool;
  etxclass->read_etx_filename = "<unknown>";
  etxclass->read_etx_lineno = 1;
  etxclass->etl_obstack = &etxclass->obstack;
  etxclass->define_attr_tail = &etxclass->define_attr_queue;
  etxclass->define_insn_tail = &etxclass->define_insn_queue;
  etxclass->define_cond_exec_tail = &etxclass->define_cond_exec_queue;
  etxclass->other_tail = &etxclass->other_queue;
  etxclass->insn_elision = 1;
  etxclass->bitmap_obstack_init = FALSE;

  *rc = etxclass;
  /* -- ʱȭ    ---- */

  etxclass->max_include_len = 0;

  /* Common Tree  ʱȭѴ.  */
  build_common_etree_nodes (pool);

  return init_ed_reader (etxclass, ef);
}

/* The entry point for initializing the reader.  */

int
init_ed_reader (struct etx_classobj *rc, ec_file_t *ef)
{
  int c;
  char *lastsl;

  lastsl = strrchr (ef->filename, '/');
  if (lastsl != NULL)
    rc->base_dir = save_string (rc->pool, ef->filename, lastsl - ef->filename + 1 );

  rc->read_etx_filename = ef->filename;

  obstack_init (rc->etl_obstack);
  rc->errors = 0;
  rc->sequence_num = 0;

  /* Read the entire file.  */
  while (1)
    {
      etx desc;
      int lineno;

      c = read_etx_skip_spaces (rc, ef);
      if (c == EOF)
        break;

      ec_ungetc (c, ef);
      lineno = rc->read_etx_lineno;
      desc = read_etx (rc, ef);
      process_etx (rc, desc, lineno);
    }

  return rc->errors ? FATAL_EXIT_CODE : SUCCESS_EXIT_CODE;
}

/* The entry point for reading a single etx from an md file.  */

etx
ecd_read_etx (struct etx_classobj *rc, int *lineno, int *seqnr)
{
  struct queue_elem **queue, *elem;
  etx desc;

  /* Read all patterns from a given queue before moving on to the next.  */
  if (rc->define_attr_queue != NULL)
    queue = &rc->define_attr_queue;
  else if (rc->define_insn_queue != NULL)
    queue = &rc->define_insn_queue;
  else if (rc->other_queue != NULL)
    queue = &rc->other_queue;
  else
    return NULL_ETX;

  elem = *queue;
  *queue = elem->next;
  desc = elem->data;
  rc->read_etx_filename = elem->filename;
  *lineno = elem->lineno;
  *seqnr = rc->sequence_num;

  return desc;
}

void
ecd_finish_reader (struct etx_classobj *rc)
{
  apr_pool_destroy (rc->pool);

  apr_terminate ();
}
