/* Print ETL for GCC.
   Copyright (C) 1987, 1988, 1992, 1997, 1998, 1999, 2000, 2002, 2003
   Free Software Foundation, Inc.

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"

/* We don't want the etree code checking code for the access to the
   DECL_NAME to be included in the gen* programs.  */
#undef ENABLE_TREE_CHECKING

static FILE *outfile;

static int sawclose = 0;

static int indent;

static void print_etx (etx);

/* String printed at beginning of each ETL when it is dumped.
   This string is set to ASM_COMMENT_START when the ETL is dumped in
   the assembly output file.  */
const char *print_etx_head = "";

/* Nonzero means suppress output of instruction numbers and line number
   notes in debugging dumps.
   This must be defined here so that programs like gencodes can be linked.  */
int etx_flag_dump_unnumbered = 0;

/* Nonzero means use simplified format without flags, modes, etc.  */
int etx_flag_simple = 0;

/* Nonzero if we are dumping graphical description.  */
int dump_for_graph;

/* Holds an array of names indexed by insn_code_number.  */
static char **insn_name_ptr = 0;
static int insn_name_ptr_size = 0;

/* Define this so we can link with print-rtl.o to get debug_rtx function.  */
const char *
etx_get_insn_name (int code)
{
  if (code < insn_name_ptr_size)
    return insn_name_ptr[code];
  else
    return NULL;
}

/* Print IN_ETX onto OUTFILE.  This is the recursive part of printing.  */

static void
print_etx (etx in_etx)
{
  int i = 0;
  int j;
  const char *format_ptr;
  int is_insn;

  if (sawclose)
    {
      if (etx_flag_simple)
	fputc (' ', outfile);
      else
	fprintf (outfile, "\n%s%*s", print_etx_head, indent * 2, "");
      sawclose = 0;
    }

  if (in_etx == 0)
    {
      fputs ("(nil)", outfile);
      sawclose = 1;
      return;
    }
  else if (ETX_GET_CODE (in_etx) > NUM_ETX_CODE)
    {
       fprintf (outfile, "(??? bad code %d\n)", ETX_GET_CODE (in_etx));
       sawclose = 1;
       return;
    }

  is_insn = ETX_INSN_P (in_etx);

  /* When printing in VCG format we write INSNs, NOTE, LABEL, and BARRIER
     in separate nodes and therefore have to handle them special here.  */
  if (dump_for_graph && is_insn)
    {
      i = 3;
      indent = 0;
    }
  else
    {
      /* Print name of expression code.  */
      if (etx_flag_simple && ETX_GET_CODE (in_etx) == ETX_CONST_INT)
	fputc ('(', outfile);
      else
	fprintf (outfile, "(%s", GET_ETX_NAME (ETX_GET_CODE (in_etx)));

      if (! etx_flag_simple)
	{
	  if (ETX_FLAG (in_etx, in_struct))
	    fputs ("/s", outfile);

	  if (ETX_FLAG (in_etx, volatil))
	    fputs ("/v", outfile);

	  if (ETX_FLAG (in_etx, unchanging))
	    fputs ("/u", outfile);

	  if (ETX_FLAG (in_etx, integrated))
	    fputs ("/i", outfile);

	  if (ETX_FLAG (in_etx, frame_related))
	    fputs ("/f", outfile);

	  if (ETX_FLAG (in_etx, jump))
	    fputs ("/j", outfile);

	  if (ETX_FLAG (in_etx, call))
	    fputs ("/c", outfile);
	}
    }

  /* Get the format string and skip the first elements if we have handled
     them already.  */
  format_ptr = GET_ETX_FORMAT (ETX_GET_CODE (in_etx)) + i;
  for (; i < GET_ETX_LENGTH (ETX_GET_CODE (in_etx)); i++)
    switch (*format_ptr++)
      {
	const char *str;

      case 'T':
	str = ETX_XTMPL (in_etx, i);
	goto string;

      case 'S':
      case 's':
	str = ETX_XSTR (in_etx, i);
      string:

	if (str == 0)
	  fputs (dump_for_graph ? " \\\"\\\"" : " \"\"", outfile);
	else
	  {
	    if (dump_for_graph)
	      fprintf (outfile, " (\\\"%s\\\")", str);
	    else
	      fprintf (outfile, " (\"%s\")", str);
	  }
	sawclose = 1;
	break;

      case 'e':
	indent += 2;
	if (!sawclose)
	  fprintf (outfile, " ");
	print_etx (ETX_XEXP (in_etx, i));
	indent -= 2;
	break;

      case 'E':
      case 'V':
	indent += 2;
	if (sawclose)
	  {
	    fprintf (outfile, "\n%s%*s",
		     print_etx_head, indent * 2, "");
	    sawclose = 0;
	  }
	fputs (" [", outfile);
	if (NULL != ETX_XVEC (in_etx, i))
	  {
	    indent += 2;
	    if (ETX_XVECLEN (in_etx, i))
	      sawclose = 1;

	    for (j = 0; j < ETX_XVECLEN (in_etx, i); j++)
	      print_etx (ETX_XVECEXP (in_etx, i, j));

	    indent -= 2;
	  }
	if (sawclose)
	  fprintf (outfile, "\n%s%*s", print_etx_head, indent * 2, "");

	fputs ("]", outfile);
	sawclose = 1;
	indent -= 2;
	break;

      case 'w':
	if (! etx_flag_simple)
	  fprintf (outfile, " ");
	fprintf (outfile, HOST_WIDE_INT_PRINT_DEC, ETX_XWINT (in_etx, i));
	if (! etx_flag_simple)
	  fprintf (outfile, " [" HOST_WIDE_INT_PRINT_HEX "]",
		   ETX_XWINT (in_etx, i));
	break;

      case 'i':
	if (i == 4 && ETX_INSN_P (in_etx))
	  {
	  }
	else
	  {
	    int value = ETX_XINT (in_etx, i);
	    const char *name;

	      if (etx_flag_dump_unnumbered && is_insn)
	      fputc ('#', outfile);
	    else
	      fprintf (outfile, " %d", value);

	    if (is_insn && &ETX_INSN_CODE (in_etx) == &ETX_XINT (in_etx, i)
		&& ETX_XINT (in_etx, i) >= 0
		&& (name = etx_get_insn_name (ETX_XINT (in_etx, i))) != NULL)
	      fprintf (outfile, " {%s}", name);
	    sawclose = 0;
	  }
	break;

      case 'I':
      case 'F':
	debug_etree (ETX_XTREE (in_etx, i));
	break;

      case 'u':
	if (ETX_XEXP (in_etx, i) != NULL)
	  {
	    etx sub = ETX_XEXP (in_etx, i);

	    if (etx_flag_dump_unnumbered)
	      fputs (" #", outfile);
	    else
	      fprintf (outfile, " %d", ETX_INSN_UID (sub));
	  }
	else
	  fputs (" 0", outfile);
	sawclose = 0;
	break;

      case '*':
	fputs (" Unknown", outfile);
	sawclose = 0;
	break;

      default:
	fprintf (stderr,
		 "switch format wrong in etl.print_etx(). format was: %c.\n",
		 format_ptr[-1]);
	abort ();
      }

  switch (ETX_GET_CODE (in_etx))
    {
    default:
      break;
    }

  if (dump_for_graph && is_insn)
    sawclose = 0;
  else
    {
      fputc (')', outfile);
      sawclose = 1;
    }
}

/* Print an etx on the current line of FILE.  Initially indent IND
   characters.  */

void
print_inline_etx (FILE *outf, etx x, int ind)
{
  int oldsaw = sawclose;
  int oldindent = indent;

  sawclose = 0;
  indent = ind;
  outfile = outf;
  print_etx (x);
  sawclose = oldsaw;
  indent = oldindent;
}

/* Call this function from the debugger to see what X looks like.  */

void
debug_etx (etx x)
{
  outfile = stderr;
  sawclose = 0;
  print_etx (x);
  fprintf (stderr, "\n");
}

/* Count of etx's to print with debug_etx_list.
   This global exists because gdb user defined commands have no arguments.  */

int debug_etx_count = 0;	/* 0 is treated as equivalent to 1 */

/* Call this function to print list from X on.

   N is a count of the etx's to print. Positive values print from the specified
   etx on.  Negative values print a window around the etx.
   EG: -5 prints 2 etx's on either side (in addition to the specified etx).  */

void
debug_etx_list (etx x, int n)
{
  int i,count;
  etx insn;

  count = n == 0 ? 1 : n < 0 ? -n : n;

  /* If we are printing a window, back up to the start.  */

  if (n < 0)
    for (i = count / 2; i > 0; i--)
      {
	if (ETX_PREV_INSN (x) == 0)
	  break;
	x = ETX_PREV_INSN (x);
      }

  for (i = count, insn = x; i > 0 && insn != 0; i--, insn = ETX_NEXT_INSN (insn))
    {
      debug_etx (insn);
      fprintf (stderr, "\n");
    }
}

/* Call this function to print an etx list from START to END inclusive.  */

void
debug_etx_range (etx start, etx end)
{
  while (1)
    {
      debug_etx (start);
      fprintf (stderr, "\n");
      if (!start || start == end)
	break;
      start = ETX_NEXT_INSN (start);
    }
}

/* Call this function to search an etx list to find one with insn uid UID,
   and then call debug_etx_list to print it, using DEBUG_ETX_COUNT.
   The found insn is returned to enable further debugging analysis.  */

etx
debug_etx_find (etx x, int uid)
{
  while (x != 0 && ETX_INSN_UID (x) != uid)
    x = ETX_NEXT_INSN (x);
  if (x != 0)
    {
      debug_etx_list (x, debug_etx_count);
      return x;
    }
  else
    {
      fprintf (stderr, "insn uid %d not found\n", uid);
      return 0;
    }
}

/* External entry point for printing a chain of insns
   starting with ETX_FIRST onto file OUTF.
   A blank line separates insns.

   If ETX_FIRST is not an insn, then it alone is printed, with no newline.  */

void
print_etl (FILE *outf, etx etx_first)
{
  outfile = outf;
  sawclose = 0;

  if (etx_first == 0)
    {
      fputs (print_etx_head, outf);
      fputs ("(nil)\n", outf);
    }
  else
    switch (ETX_GET_CODE (etx_first))
      {
      default:
	fputs (print_etx_head, outfile);
	print_etx (etx_first);
      }
}

/* Like print_etx, except specify a file.  */
/* Return nonzero if we actually printed anything.  */

int
print_etl_single (FILE *outf, etx x)
{
  outfile = outf;
  sawclose = 0;
  if (! etx_flag_dump_unnumbered)
    {
      fputs (print_etx_head, outfile);
      print_etx (x);
      putc ('\n', outf);
      return 1;
    }
  return 0;
}


/* Like print_etl except without all the detail; for example,
   if ETX is a CONST_INT then print in decimal format.  */

void
print_simple_etl (FILE *outf, etx x)
{
  etx_flag_simple = 1;
  print_etl (outf, x);
  etx_flag_simple = 0;
}
