/* ETL reader for GCC.
   Copyright (C) 1987, 1988, 1991, 1994, 1997, 1998, 1999, 2000, 2001, 2002,
   2003
   Free Software Foundation, Inc.

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "hashtab.h"

static void fatal_with_file_and_line (struct etx_classobj *, ec_file_t *, const char *, ...)
  ATTRIBUTE_PRINTF_3 ATTRIBUTE_NORETURN;
static void fatal_expected_char (struct etx_classobj *, ec_file_t *, int, int) ATTRIBUTE_NORETURN;
static void read_name (struct etx_classobj *, char *, ec_file_t *);
static char *read_string (struct etx_classobj *, struct obstack *, ec_file_t *, int);
static char *read_quoted_string (struct etx_classobj *, struct obstack *, ec_file_t *);
static char *read_braced_string (struct etx_classobj *, struct obstack *, ec_file_t *);
static void read_escape (struct etx_classobj *, struct obstack *, ec_file_t *);
static hashval_t def_hash (const void *);
static int def_name_eq_p (const void *, const void *);
static void read_constants (struct etx_classobj *, ec_file_t *, char *tmp_char);
static void validate_const_int (struct etx_classobj *, ec_file_t *, const char *);
static void validate_const_float (struct etx_classobj *, ec_file_t *, const char *);

int ec_getc (ec_file_t *);
int ec_ungetc (int, ec_file_t *);
long ec_ftell (ec_file_t *);

int
ec_getc (ec_file_t *ef)
{
  int c;

  if ((ef->cur - ef->base) >= ef->size)
    return EOF;
  
  c = ef->cur[0] & 0xff;
  ef->cur++;

  return c;
}

int
ec_ungetc (int c, ec_file_t *ef)
{
  ef->cur--;
  ef->cur[0] = c & 0xff;

  return c;
}

long
ec_ftell (ec_file_t *ef)
{
  return (long) (ef->cur - ef->base);
}

/* Subroutines of read_etx.  */

static void
fatal_with_file_and_line (struct etx_classobj *rc, ec_file_t *ef, const char *msg, ...)
{
  char context[64];
  size_t i;
  int c;
  va_list ap;

  va_start (ap, msg);

  fprintf (stderr, "%s:%d: ", rc->read_etx_filename, rc->read_etx_lineno);
  vfprintf (stderr, msg, ap);
  putc ('\n', stderr);

  /* Gather some following context.  */
  for (i = 0; i < sizeof (context)-1; ++i)
    {
      c = ec_getc (ef);
      if (c == EOF)
	break;
      if (c == '\r' || c == '\n')
	break;
      context[i] = c;
    }
  context[i] = '\0';

  fprintf (stderr, "%s:%d: following context is `%s'\n",
	   rc->read_etx_filename, rc->read_etx_lineno, context);

  va_end (ap);
  exit (1);
}

/* Dump code after printing a message.  Used when read_etx finds
   invalid data.  */

static void
fatal_expected_char (struct etx_classobj *rc, ec_file_t *ef, int expected_c, int actual_c)
{
  fatal_with_file_and_line (rc, ef, "expected character `%c', found `%c'",
			    expected_c, actual_c);
}

/* Read chars from INFILE until a non-whitespace char
   and return that.  Comments, both Lisp style and C style,
   are treated as whitespace.
   Tools such as genflags use this function.  */

int
read_etx_skip_spaces (struct etx_classobj *rc, ec_file_t *ef)
{
  int c;

  while (1)
    {
      c = ec_getc (ef);
      switch (c)
	{
	case '\n':
	  rc->read_etx_lineno++;
	  break;

	case ' ': case '\t': case '\f': case '\r':
	  break;

	case ';':
	  do
	    c = ec_getc (ef);
	  while (c != '\n' && c != EOF);
	  rc->read_etx_lineno++;
	  break;

	case '/':
	  {
	    int prevc;
	    c = ec_getc (ef);
	    if (c != '*')
	      fatal_expected_char (rc, ef, '*', c);

	    prevc = 0;
	    while ((c = ec_getc (ef)) && c != EOF)
	      {
		if (c == '\n')
		   rc->read_etx_lineno++;
	        else if (prevc == '*' && c == '/')
		  break;
	        prevc = c;
	      }
	  }
	  break;

	default:
	  return c;
	}
    }
}

/* Read an etx code name into the buffer STR[].
   It is terminated by any of the punctuation chars of etx printed syntax.  */

static void
read_name (struct etx_classobj *rc, char *str, ec_file_t *ef)
{
  char *p;
  int c;

  c = read_etx_skip_spaces (rc, ef);

  p = str;
  while (1)
    {
      if (c == ' ' || c == '\n' || c == '\t' || c == '\f' || c == '\r')
	break;
      if (c == ':' || c == ')' || c == ']' || c == '"' || c == '/'
	  || c == '(' || c == '[')
	{
	  ec_ungetc (c, ef);
	  break;
	}
      *p++ = c;
      c = ec_getc (ef);
    }
  if (p == str)
    fatal_with_file_and_line (rc, ef, "missing name or number");
  if (c == '\n')
    rc->read_etx_lineno++;

  *p = 0;

  if (rc->md_constants)
    {
      /* Do constant expansion.  */
      struct ed_constant *def;

      p = str;
      do
	{
	  struct ed_constant tmp_def;

	  tmp_def.name = p;
	  def = htab_find (rc->md_constants, &tmp_def);
	  if (def)
	    p = def->value;
	} while (def);
      if (p != str)
	strcpy (str, p);
    }
}

/* Subroutine of the string readers.  Handles backslash escapes.
   Caller has read the backslash, but not placed it into the obstack.  */
static void
read_escape (struct etx_classobj *rc, struct obstack *ob, ec_file_t *ef)
{
  int c = ec_getc (ef);

  switch (c)
    {
      /* Backslash-newline is replaced by nothing, as in C.  */
    case '\n':
      rc->read_etx_lineno++;
      return;

      /* \" \' \\ are replaced by the second character.  */
    case '\\':
    case '"':
    case '\'':
      break;

      /* Standard C string escapes:
	 \a \b \f \n \r \t \v
	 \[0-7] \x
	 all are passed through to the output string unmolested.
	 In normal use these wind up in a string constant processed
	 by the C compiler, which will translate them appropriately.
	 We do not bother checking that \[0-7] are followed by up to
	 two octal digits, or that \x is followed by N hex digits.
	 \? \u \U are left out because they are not in traditional C.  */
    case 'a': case 'b': case 'f': case 'n': case 'r': case 't': case 'v':
    case '0': case '1': case '2': case '3': case '4': case '5': case '6':
    case '7': case 'x':
      obstack_1grow (ob, '\\');
      break;

      /* \; makes stuff for a C string constant containing
	 newline and tab.  */
    case ';':
      obstack_grow (ob, "\\n\\t", 4);
      return;

      /* pass anything else through, but issue a warning.  */
    default:
      fprintf (stderr, "%s:%d: warning: unrecognized escape \\%c\n",
	       rc->read_etx_filename, rc->read_etx_lineno, c);
      obstack_1grow (ob, '\\');
      break;
    }

  obstack_1grow (ob, c);
}


/* Read a double-quoted string onto the obstack.  Caller has scanned
   the leading quote.  */
static char *
read_quoted_string (struct etx_classobj *rc, struct obstack *ob, ec_file_t *ef)
{
  int c;

  while (1)
    {
      c = ec_getc (ef); /* Read the string  */
      if (c == '\n')
	rc->read_etx_lineno++;
      else if (c == '\\')
	{
	  read_escape (rc, ob, ef);
	  continue;
	}
      else if (c == '"')
	break;

      obstack_1grow (ob, c);
    }

  obstack_1grow (ob, 0);
  return obstack_finish (ob);
}

/* Read a braced string (a la Tcl) onto the obstack.  Caller has
   scanned the leading brace.  Note that unlike quoted strings,
   the outermost braces _are_ included in the string constant.  */
static char *
read_braced_string (struct etx_classobj *rc, struct obstack *ob, ec_file_t *ef)
{
  int c;
  int brace_depth = 1;  /* caller-processed */
  unsigned long starting_read_etx_lineno = rc->read_etx_lineno;

  obstack_1grow (ob, '{');
  while (brace_depth)
    {
      c = ec_getc (ef); /* Read the string  */

      if (c == '\n')
	rc->read_etx_lineno++;
      else if (c == '{')
	brace_depth++;
      else if (c == '}')
	brace_depth--;
      else if (c == '\\')
	{
	  read_escape (rc, ob, ef);
	  continue;
	}
      else if (c == EOF)
	fatal_with_file_and_line
	  (rc, ef, "missing closing } for opening brace on line %lu",
	   starting_read_etx_lineno);

      obstack_1grow (ob, c);
    }

  obstack_1grow (ob, 0);
  return obstack_finish (ob);
}

/* Read some kind of string constant.  This is the high-level routine
   used by read_etx.  It handles surrounding parentheses, leading star,
   and dispatch to the appropriate string constant reader.  */

static char *
read_string (struct etx_classobj *rc, struct obstack *ob, ec_file_t *ef, int star_if_braced)
{
  char *stringbuf;
  int saw_paren = 0;
  int c;

  c = read_etx_skip_spaces (rc, ef);
  if (c == '(')
    {
      saw_paren = 1;
      c = read_etx_skip_spaces (rc, ef);
    }

  if (c == '"')
    stringbuf = read_quoted_string (rc, ob, ef);
  else if (c == '{')
    {
      if (star_if_braced)
	obstack_1grow (ob, '*');
      stringbuf = read_braced_string (rc, ob, ef);
    }
  else
    fatal_with_file_and_line (rc, ef, "expected `\"' or `{', found `%c'", c);

  if (saw_paren)
    {
      c = read_etx_skip_spaces (rc, ef);
      if (c != ')')
	fatal_expected_char (rc, ef, ')', c);
    }

  return stringbuf;
}

/* Provide a version of a function to read a long long if the system does
   not provide one.  */
#if HOST_BITS_PER_WIDE_INT > HOST_BITS_PER_LONG && !defined(HAVE_ATOLL) && !defined(HAVE_ATOQ)
HOST_WIDE_INT atoll (const char *);

HOST_WIDE_INT
atoll (const char *p)
{
  int neg = 0;
  HOST_WIDE_INT tmp_wide;

  while (ISSPACE (*p))
    p++;
  if (*p == '-')
    neg = 1, p++;
  else if (*p == '+')
    p++;

  tmp_wide = 0;
  while (ISDIGIT (*p))
    {
      HOST_WIDE_INT new_wide = tmp_wide*10 + (*p - '0');
      if (new_wide < tmp_wide)
	{
	  /* Return INT_MAX equiv on overflow.  */
	  tmp_wide = (~(unsigned HOST_WIDE_INT) 0) >> 1;
	  break;
	}
      tmp_wide = new_wide;
      p++;
    }

  if (neg)
    tmp_wide = -tmp_wide;
  return tmp_wide;
}
#endif

/* Given a constant definition, return a hash code for its name.  */
static hashval_t
def_hash (const void *def)
{
  unsigned result, i;
  const char *string = ((const struct ed_constant *) def)->name;

  for (result = i = 0;*string++ != '\0'; i++)
    result += ((unsigned char) *string << (i % CHAR_BIT));
  return result;
}

/* Given two constant definitions, return true if they have the same name.  */
static int
def_name_eq_p (const void *def1, const void *def2)
{
  return ! strcmp (((const struct ed_constant *) def1)->name,
		   ((const struct ed_constant *) def2)->name);
}

/* INFILE is a FILE pointer to read text from.  TMP_CHAR is a buffer suitable
   to read a name or number into.  Process a define_constants directive,
   starting with the optional space after the "define_constants".  */
static void
read_constants (struct etx_classobj *rc, ec_file_t *ef, char *tmp_char)
{
  int c;
  htab_t defs;

  c = read_etx_skip_spaces (rc, ef);
  if (c != '[')
    fatal_expected_char (rc, ef, '[', c);
  defs = rc->md_constants;
  if (! defs)
    defs = htab_create (32, def_hash, def_name_eq_p, (htab_del) 0);
  /* Disable constant expansion during definition processing.  */
  rc->md_constants = 0;
  while ( (c = read_etx_skip_spaces (rc, ef)) != ']')
    {
      struct ed_constant *def;
      void **entry_ptr;

      if (c != '(')
	fatal_expected_char (rc, ef, '(', c);
      def = apr_pcalloc (rc->pool, sizeof (struct ed_constant));
      def->name = tmp_char;
      read_name (rc, tmp_char, ef);
      entry_ptr = htab_find_slot (defs, def, TRUE);
      if (! *entry_ptr)
	def->name = apr_pstrdup (rc->pool, tmp_char);
      c = read_etx_skip_spaces (rc, ef);
      ec_ungetc (c, ef);
      read_name (rc, tmp_char, ef);
      if (! *entry_ptr)
	{
	  def->value = apr_pstrdup (rc->pool, tmp_char);
	  *entry_ptr = def;
	}
      else
	{
	  def = *entry_ptr;
	  if (strcmp (def->value, tmp_char))
	    fatal_with_file_and_line (rc, ef,
				      "redefinition of %s, was %s, now %s",
				      def->name, def->value, tmp_char);
	}
      c = read_etx_skip_spaces (rc, ef);
      if (c != ')')
	fatal_expected_char (rc, ef, ')', c);
    }
  rc->md_constants = defs;
  c = read_etx_skip_spaces (rc, ef);
  if (c != ')')
    fatal_expected_char (rc, ef, ')', c);
}

/* For every constant definition, call CALLBACK with two arguments:
   a pointer a pointer to the constant definition and INFO.
   Stops when CALLBACK returns zero.  */
void
traverse_ed_constants (struct etx_classobj *rc, htab_trav callback, void *info)
{
  if (rc->md_constants)
    htab_traverse (rc->md_constants, callback, info);
}

static void
validate_const_int (struct etx_classobj *rc, ec_file_t *ef, const char *string)
{
  const char *cp;
  int valid = 1;

  cp = string;
  while (*cp && ISSPACE (*cp))
    cp++;
  if (*cp == '-' || *cp == '+')
    cp++;
  if (*cp == 0)
    valid = 0;
  for (; *cp; cp++)
    if (! ISDIGIT (*cp))
      valid = 0;
  if (!valid)
    fatal_with_file_and_line (rc, ef, "invalid decimal constant \"%s\"\n", string);
}

static void
validate_const_float (struct etx_classobj *rc, ec_file_t *ef, const char *string)
{
  const char *cp;
  int valid = 1;

  cp = string;
  while (*cp && ISSPACE (*cp))
    cp++;
  if (*cp == '-' || *cp == '+')
    cp++;
  if (*cp == 0)
    valid = 0;
  for (; *cp; cp++)
    if (! ISDIGIT (*cp) && *cp != '.')
      valid = 0;
  if (!valid)
    fatal_with_file_and_line (rc, ef, "invalid decimal constant \"%s\"\n", string);
}

/* Read an etx in printed representation from INFILE
   and return an actual etx in core constructed accordingly.
   read_etx is not used in the compiler proper, but rather in
   the utilities gen*.c that construct C code from machine descriptions.  */

etx
read_etx (struct etx_classobj *rc, ec_file_t *ef)
{
  int i;
  ETX_CODE tmp_code;
  const char *format_ptr;
  /* tmp_char is a buffer used for reading decimal integers
     and names of etx types and machine modes.
     Therefore, 256 must be enough.  */
  char tmp_char[256];
  etx return_etx;
  int c;
  int tmp_int;
  HOST_WIDE_INT tmp_wide;

  /* Linked list structure for making ETXs: */
  struct etx_list
    {
      struct etx_list *next;
      etx value;		/* Value of this node.  */
    };

  if (!rc->initialized) {
    obstack_init (rc->etl_obstack);
    rc->initialized = 1;
  }

again:
  c = read_etx_skip_spaces (rc, ef); /* Should be open paren.  */
  if (c != '(')
    fatal_expected_char (rc, ef, '(', c);

  read_name (rc, tmp_char, ef);

  tmp_code = ETX_UNKNOWN;

  if (! strcmp (tmp_char, "define_constants"))
    {
      read_constants (rc, ef, tmp_char);
      goto again;
    }
  for (i = 0; i < NUM_ETX_CODE; i++)
    if (! strcmp (tmp_char, GET_ETX_NAME (i)))
      {
	tmp_code = (ETX_CODE) i;	/* get value for name */
	break;
      }

  if (tmp_code == ETX_UNKNOWN)
    fatal_with_file_and_line (rc, ef, "unknown etx code `%s'", tmp_char);

  /* (NIL) stands for an expression that isn't there.  */
  if (tmp_code == ETX_NIL)
    {
      /* Discard the closeparen.  */
      while ((c = ec_getc (ef)) && c != ')')
	;

      return 0;
    }

  /* If we end up with an insn expression then we free this space below.  */
  return_etx = etx_alloc (rc->pool, tmp_code);
  format_ptr = GET_ETX_FORMAT (ETX_GET_CODE (return_etx));

  /* If what follows is `: mode ', read it and
     store the mode in the etx.  */

  i = read_etx_skip_spaces (rc, ef);
  ec_ungetc (i, ef);

  for (i = 0; i < GET_ETX_LENGTH (ETX_GET_CODE (return_etx)); i++)
    switch (*format_ptr++)
      {
	/* 0 means a field for internal use only.
	   Don't expect it to be present in the input.  */
      case '0':
	break;

      case 'e':
      case 'u':
	ETX_XEXP (return_etx, i) = read_etx (rc, ef);
	break;

      case 'V':
	/* 'V' is an optional vector: if a closeparen follows,
	   just store NULL for this element.  */
	c = read_etx_skip_spaces (rc, ef);
	ec_ungetc (c, ef);
	if (c == ')')
	  {
	    ETX_XVEC (return_etx, i) = 0;
	    break;
	  }
	/* Now process the vector.  */

      case 'E':
	{
	  /* Obstack to store scratch vector in.  */
	  struct obstack vector_stack;
	  int list_counter = 0;
	  etvec return_vec = NULL_ETVEC;

	  c = read_etx_skip_spaces (rc, ef);
	  if (c != '[')
	    fatal_expected_char (rc, ef, '[', c);

	  /* add expressions to a list, while keeping a count */
	  obstack_init (&vector_stack);
	  while ((c = read_etx_skip_spaces (rc, ef)) && c != ']')
	    {
	      ec_ungetc (c, ef);
	      list_counter++;
	      obstack_ptr_grow (&vector_stack, read_etx (rc, ef));
	    }
	  if (list_counter > 0)
	    {
	      return_vec = etvec_alloc (rc->pool, list_counter);
	      memcpy (&return_vec->elem[0], obstack_finish (&vector_stack),
		      list_counter * sizeof (etx));
	    }
	  ETX_XVEC (return_etx, i) = return_vec;
	  obstack_free (&vector_stack, NULL);
	  /* close bracket gotten */
	}
	break;

      case 'S':
	/* 'S' is an optional string: if a closeparen follows,
	   just store NULL for this element.  */
	c = read_etx_skip_spaces (rc, ef);
	ec_ungetc (c, ef);
	if (c == ')')
	  {
	    ETX_XSTR (return_etx, i) = 0;
	    break;
	  }

      case 'T':
      case 's':
	{
	  char *stringbuf;

	  /* The output template slot of a DEFINE_INSN,
	     DEFINE_INSN_AND_SPLIT, or DEFINE_PEEPHOLE automatically
	     gets a star inserted as its first character, if it is
	     written with a brace block instead of a string constant.  */
	  int star_if_braced = (format_ptr[-1] == 'T');

	  stringbuf = read_string (rc, rc->etl_obstack, ef, star_if_braced);

	  /* For insn patterns, we want to provide a default name
	     based on the file and line, like "*foo.md:12", if the
	     given name is blank.  These are only for define_insn and
	     define_insn_and_split, to aid debugging.  */
	  if (*stringbuf == '\0'
	      && i == 0)
	    {
	      char line_name[20];
	      const char *fn = (rc->read_etx_filename ? rc->read_etx_filename : "etx");
	      const char *slash;
	      for (slash = fn; *slash; slash ++)
		if (*slash == '/' || *slash == '\\' || *slash == ':')
		  fn = slash + 1;
	      obstack_1grow (rc->etl_obstack, '*');
	      obstack_grow (rc->etl_obstack, fn, strlen (fn));
	      sprintf (line_name, ":%d", rc->read_etx_lineno);
	      obstack_grow (rc->etl_obstack, line_name, strlen (line_name)+1);
	      stringbuf = (char *) obstack_finish (rc->etl_obstack);
	    }

	  if (star_if_braced)
	    ETX_XTMPL (return_etx, i) = stringbuf;
	  else
	    ETX_XSTR (return_etx, i) = stringbuf;
	}
	break;

      case 'w':
	read_name (rc, tmp_char, ef);
	validate_const_int (rc, ef, tmp_char);
#if HOST_BITS_PER_WIDE_INT == HOST_BITS_PER_INT
	tmp_wide = atoi (tmp_char);
#else
#if HOST_BITS_PER_WIDE_INT == HOST_BITS_PER_LONG
	tmp_wide = atol (tmp_char);
#else
	/* Prefer atoll over atoq, since the former is in the ISO C99 standard.
	   But prefer not to use our hand-rolled function above either.  */
#if defined(HAVE_ATOLL) || !defined(HAVE_ATOQ)
	tmp_wide = atoll (tmp_char);
#else
	tmp_wide = atoq (tmp_char);
#endif
#endif
#endif
	ETX_XWINT (return_etx, i) = tmp_wide;
	break;

      case 'i':
      case 'n':
	read_name (rc, tmp_char, ef);
	validate_const_int (rc, ef, tmp_char);
	tmp_int = atoi (tmp_char);
	ETX_XINT (return_etx, i) = tmp_int;
	break;

      case 'I':
	{
	  char *r;
	  int is_negate = 0;

	  read_name (rc, tmp_char, ef);
	  validate_const_int (rc, ef, tmp_char);
	  
	  r = tmp_char;
	  switch (r[0])
	    {
	    case '-':
	      is_negate = 1;
	      r++;
	      break;
	    default:
	      break;
	    }
	  while (ISSPACE (*r))
	    r++;
	  
	  ETX_XTREE (return_etx, i) = lex_number (rc->pool, r, strlen (r));
	  if (is_negate)
	    ETX_XTREE (return_etx, i) = etree_fold (rc->pool, etree_build_minus (rc->pool, ETX_XTREE (return_etx, i)));
	  break;
	}
      case 'F':
	{
	  char *r;
	  int is_negate = 0;

	  read_name (rc, tmp_char, ef);
	  validate_const_float (rc, ef, tmp_char);
	  
	  r = tmp_char;
	  switch (r[0])
	    {
	    case '-':
	      is_negate = 1;
	      r++;
	      break;
	    default:
	      break;
	    }
	  while (ISSPACE (*r))
	    r++;
	  
	  ETX_XTREE (return_etx, i) = lex_number (rc->pool, r, strlen (r));
	  if (is_negate)
	    ETX_XTREE (return_etx, i) = etree_fold (rc->pool, etree_build_minus (rc->pool, ETX_XTREE (return_etx, i)));
	  break;
	}

      default:
	fprintf (stderr,
		 "switch format wrong in etl.read_etx(). format was: %c.\n",
		 format_ptr[-1]);
	fprintf (stderr, "\tfile position: %ld\n", ec_ftell (ef));
	abort ();
      }

  c = read_etx_skip_spaces (rc, ef);
  if (c != ')')
    fatal_expected_char (rc, ef, ')', c);

  return return_etx;
}
